#include <form.h>
#include <mask.h>
#include <printapp.h>
#include <utility.h>

#include "soggetti.h"
#include "contsan.h"
#include "sezioni.h"
#include <comuni.h>

#include "at2.h"
#include "at2100a.h"
#include "atlib.h"

#define ALIAS_LCPDOM	100
#define ALIAS_TCS			200
#define ALIAS_GAZ			300   // gruppi aziendale

enum ts { undefined=0, sintetico=1, completo=2, etichette=3, lettere=4 };

// definizione form per etichette
class TEti_sospesi_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TEti_sospesi_form(): TForm() {};
  TEti_sospesi_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TEti_sospesi_form() {};
};

class TStampaSospesi : public TPrintapp
{
  TRelation*					_rel;
  TMask*      				_msk;
  TLocalisamfile*			_contsan;
  TEti_sospesi_form*	_form_eti;
  TEti_sospesi_form*	_form_let;
  TAssoc_array		    _categorie;
  TParagraph_string   _cognome_nome, _dencom, _motivo;
  TDate 							_data_stampa;
  TDate								_dataini, _datafin, _sodataini, _sodatafin;
	ts 	      					_tipostampa;
	TString16						_codsez, _codsot, _motivo1, _motivo2, _motivo3, _motivo4, _motivo5, _tiposo, _prosstipo;
  TString16				  	_gruppoazie, _gruppoold;
	int									_etlarghezza, _etcolonne, _contatore, _totale;
	bool								_motivi, _pergruppo;

  static bool filter_func_sospesi(const TRelation* rel);
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page(int file, int counter);
  virtual print_action postprocess_print(int file, int counter);

public:
  void crea_intestazione();
  void filtra_sezioni();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  void header_gruppo(const TString16 gruppo);
  void footer_sezione();
  void fine_stampa();
  TMask& app_mask() { return *_msk; }
  TStampaSospesi() : _data_stampa(TODAY), _cognome_nome("",25), _dencom("",30), _motivo("",5)  {}
};

HIDDEN inline TStampaSospesi& app() { return (TStampaSospesi&) main_app(); }

TCursor* TEti_sospesi_form::cursor() const { return app().current_cursor(); }

TRelation* TEti_sospesi_form::relation() const { return cursor()->relation(); }

void TStampaSospesi::filtra_sezioni()
{
 	const TString16 sezini = _msk->get(F_SEZINI);
 	const TString16 sotini = _msk->get(F_SOTINI);
 	const TString16 sezfin = _msk->get(F_SEZFIN);
 	const TString16 sotfin = _msk->get(F_SOTFIN);
  TRectype da(LF_SOGGETTI);
  TRectype a (LF_SOGGETTI);
  if (sezini.not_empty())
		da.put(SOG_CODSEZ, sezini);
  if (sotini.not_empty())
		da.put(SOG_CODSOT, sotini);
  if (sezfin.not_empty())
		a.put(SOG_CODSEZ, sezfin);
  if (sotfin.not_empty())
		a.put(SOG_CODSOT, sotfin);
	current_cursor()->setregion(da, a);
}	

void TStampaSospesi::set_page(int file, int cnt)
{
	// costruzione etichette
	switch (_tipostampa)
	{
		case etichette:
		{
			TPrint_section& corpo = _form_eti->get_body();
			corpo.update();
  		for (word i = 0; i < corpo.height(); i++)
  		{
  			TPrintrow& riga = corpo.row(i);
  			set_row(i+1,riga);
			}  			
			force_setpage(TRUE);	// serve perch alla prossima etichetta rifaccia la setpage
														// altrimenti stampa sempre la stessa etichetta
		}
		break;
		case lettere:
		{
			TPrint_section& corpo = _form_let->get_body();
			corpo.update();
  		for (word i = 0; i < corpo.height(); i++)
  		{
  			TPrintrow& riga = corpo.row(i);
  			set_row(i+1,riga);
			}  			
			force_setpage(TRUE);	// serve perch alla prossima etichetta rifaccia la setpage
														// altrimenti stampa sempre la stessa etichetta
		}
		break;
		case sintetico:
		{
			set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
			set_row(2,"@0g@8,rs", FLD(LF_SOGGETTI,SOG_TESSAVIS));
			set_row(1,"@9g@S", FLD(LF_SOGGETTI,SOG_CATDON));
			set_row(1,"@12g#a", &_cognome_nome);
			set_row(1,"@38g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
			set_row(2,"@38g@S", FLD(LF_SOGGETTI,SOG_GRUPPOAB0));
			set_row(2,"@45g@S", FLD(LF_SOGGETTI,SOG_RHANTID));
			set_row(1,"@49g@ld", FLD(LF_SOGGETTI,SOG_DATASTATO));
			set_row(1,"@60g@S", FLD(LF_SOGGETTI,SOG_STATO));
			set_row(2,"@49g@ld", FLD(LF_SOGGETTI,SOG_DATA_PROS));
			set_row(2,"@60g@S", FLD(LF_SOGGETTI,SOG_PROS_STATO));
			set_row(1,"@63g@ld", FLD(LF_SOGGETTI,SOG_DATAULTID));
			set_row(1,"@74g@S", FLD(LF_SOGGETTI,SOG_TIPOULTID));
			set_row(1,"@77g@pn", FLD(LF_SOGGETTI,SOG_INTSI,"###"));
			set_row(1,"@81g@pn", FLD(LF_SOGGETTI,SOG_INTAF,"###"));
			set_row(1,"@88g@ld", FLD(LF_SOGGETTI,SOG_DATAULTDON));
			set_row(1,"@99g@S", FLD(LF_SOGGETTI,SOG_TIPOULTDON));
			set_row(3,"");
		}
		break;
		case completo:
		{
			set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
			set_row(2,"@0g@8,rs", FLD(LF_SOGGETTI,SOG_TESSAVIS));
			set_row(3,"@0g@8,rs", FLD(LF_SOGGETTI,SOG_CATDON));

			set_row(1,"@9g@S",FLD(LF_SOGGETTI,SOG_COGNOME));
			set_row(2,"@9g@S",FLD(LF_SOGGETTI,SOG_NOME));
			set_row(3,"@9g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));

			set_row(1,"@35g@S", FLD(LF_SOGGETTI,SOG_DOM_INDIR));
			set_row(2,"@35g#a", &_dencom);
	
			set_row(1,"@66g@S", FLD(LF_SOGGETTI,SOG_TELABI));
			set_row(2,"@66g@S", FLD(LF_SOGGETTI,SOG_TELLAV));
			set_row(3,"@66g@S", FLD(LF_SOGGETTI,SOG_TELALT));

			set_row(1,"@82gGr @S", FLD(LF_SOGGETTI,SOG_GRUPPOAB0));
			set_row(2,"@82gRh @S", FLD(LF_SOGGETTI,SOG_RHANTID));
			set_row(3,"@82gFe @S", FLD(LF_SOGGETTI,SOG_FENOTIPORH));
			set_row(4,"@82gKe @S", FLD(LF_SOGGETTI,SOG_KELL));
			set_row(5,"@82gDu @S", FLD(LF_SOGGETTI,SOG_DU));

			set_row(1,"@94g@S", FLD(LF_SOGGETTI,SOG_STATO));
			set_row(2,"@94g@ld", FLD(LF_SOGGETTI,SOG_DATASTATO));
			set_row(3,"@94g@S", FLD(LF_SOGGETTI,SOG_PROS_STATO));
			set_row(4,"@94g@ld", FLD(LF_SOGGETTI,SOG_DATA_PROS));
			set_row(5,"@94g#a", &_motivo);

			set_row(1,"@105g@ld", FLD(LF_SOGGETTI,SOG_DATAULTID));
			set_row(2,"@105g@S", FLD(LF_SOGGETTI,SOG_TIPOULTID));
			set_row(4,"@112g@pn", FLD(LF_SOGGETTI,SOG_INTSI,"###"));
			set_row(5,"@112g@pn", FLD(LF_SOGGETTI,SOG_INTAF,"###"));
			set_row(1,"@116g@ld", FLD(LF_SOGGETTI,SOG_DATAULTDON));
			set_row(2,"@116g@S", FLD(LF_SOGGETTI,SOG_TIPOULTDON));
			set_row(4,"@123g@pn", FLD(LF_SOGGETTI,SOG_TOTDON,"###"));
			set_row(6,"");
		}
		break;
	}					
}

bool TStampaSospesi::filter_func_sospesi(const TRelation* rel)
{   
	bool filtrato = TRUE;
	TRectype recsog = rel->lfile().curr();
	// filtro per categorie
	TAssoc_array& categorie = app()._categorie;
	if (categorie.items() != 0)
	{
		const TString16 cat = recsog.get(SOG_CATDON);
		filtrato = categorie.is_key((const char*) cat);
	}	
	if ((filtrato) && (app()._sodataini.ok()))
	{
		const TDate datastato = recsog.get_date(SOG_DATASTATO);
		filtrato = (datastato >= app()._sodataini);
	}
	if ((filtrato) && (app()._sodatafin.ok()))
	{
		const TDate datastato = recsog.get_date(SOG_DATASTATO);
		filtrato = (datastato <= app()._sodatafin);
	}
	if ((filtrato) && (app()._dataini.ok()))
	{
		const TDate datapross = recsog.get_date(SOG_DATA_PROS);
		filtrato = (datapross >= app()._dataini);
	}
	if ((filtrato) && (app()._datafin.ok()))
	{
		const TDate datapross = recsog.get_date(SOG_DATA_PROS);
		filtrato = (datapross <= app()._datafin);
	}
	if ((filtrato) && app()._motivi)
	{
		TLocalisamfile contsan(LF_CONTSAN);
		contsan.setkey(2);
		contsan.put(CON_DATACON, recsog.get(SOG_DATASTATO));
		contsan.put(CON_CODICE, recsog.get(SOG_CODICE));
		if (contsan.read() == NOERR)
		{                                            
			const TString16 motivo = contsan.get(CON_MOTIVO);
			if ((app()._motivo1==motivo)
			 || (app()._motivo2==motivo && app()._motivo2.not_empty()) || (app()._motivo3==motivo && app()._motivo3.not_empty()) || (app()._motivo4==motivo && app()._motivo4.not_empty()) || (app()._motivo5==motivo) && app()._motivo5.not_empty())
				filtrato = TRUE;
			else 
				filtrato = FALSE;				
		}				
		else
			filtrato = FALSE;				
	}
  return filtrato;
}

bool TStampaSospesi::preprocess_page(int file, int counter)
{
	if ((_tipostampa==sintetico) || (_tipostampa==completo))
	{
		TRectype& recsog = current_cursor()->curr();
		TString80 nome = recsog.get(SOG_COGNOME);
		nome << ' ';
		nome << recsog.get(SOG_NOME);
		_cognome_nome = nome;
		TString256 localita = "";
		localita << recsog.get(SOG_DOM_CAP);
		localita << ' ';
		if (recsog.get(SOG_DOM_CODLOC).not_empty())
			localita << current_cursor()->curr(-ALIAS_LCPDOM).get("S0");
		else
			localita << current_cursor()->curr(LF_COMUNI).get(COM_DENCOM);
		localita << ' ' << current_cursor()->curr(LF_COMUNI).get(COM_PROVCOM);
		localita.trim();
		_dencom = localita;                
		TString16 motivo = "";
		TLocalisamfile contsan(LF_CONTSAN);
		contsan.setkey(2);
		contsan.put(CON_DATACON, recsog.get(SOG_DATASTATO));
		contsan.put(CON_CODICE, recsog.get(SOG_CODICE));
		if (contsan.read() == NOERR)
			motivo = contsan.get(CON_MOTIVO);
		_motivo = motivo;
		// salto pagina se cambio sezione
		if (_pergruppo)
		{
			TString16 grupponew = current_cursor()->curr(LF_SOGGETTI).get(SOG_GRUPPOAZIE);		
			if (grupponew != _gruppoold )
			{
				if (_gruppoold != "**")
					footer_sezione();
				_contatore = 0;				
				_gruppoold = grupponew;
				header_gruppo(grupponew);
			}
		}
		else
		{
			const TString16 codsez = recsog.get(SOG_CODSEZ);		
			const TString16 codsot = recsog.get(SOG_CODSOT);		
			if ((_codsez!=codsez)||(_codsot!=codsot))
			{
				if (_codsez != "**")
					footer_sezione();
				_contatore = 0;				
				_codsez = codsez;
				_codsot = codsot;
				header_sezione(codsez, codsot);
			}
		}	
	}		
	if (_tipostampa==etichette)		
		if (printer().rows_left() < _form_eti->get_body().height())
			printer().formfeed();
	if (_tipostampa==lettere)		
		if (printer().rows_left() < _form_let->get_body().height())
			printer().formfeed();
	_contatore++;			
	_totale++;
  return TRUE;
}

print_action TStampaSospesi::postprocess_print(int file, int counter)
{ 
	if (_contatore > 0)
		footer_sezione();
	fine_stampa();	
	return NEXT_PAGE;
}

void TStampaSospesi::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(132);
	intestazione = "Sezione: ";
	intestazione << codsez;
	if (codsot.not_empty())
	{
		intestazione << '/';
		intestazione << codsot;
	}		
	intestazione << ' ';
	intestazione << densez;
	if (densot.not_empty())
	{
		intestazione << '/';
		intestazione << densot;
	}		
	if (_tipostampa==sintetico)
		intestazione.center_just(100);
	else	
		intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

void TStampaSospesi::footer_sezione()
{
	// stampa totale soggetti appartenenti alla sezione
  if (_tipostampa==sintetico || _tipostampa==completo)
	{
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		set_footer(3,"TOTALE SOGGETTI STAMPATI %d", _contatore);
		printer().formfeed(); 
		reset_footer();
	}	
}
  
void TStampaSospesi::header_gruppo(const TString16 gruppo)
{
	TString intestazione(132);
	intestazione = "GRUPPO AZIENDALE ";
	intestazione << gruppo;
	intestazione << " - ";
	intestazione << current_cursor()->curr(-ALIAS_GAZ).get("S0");
	if (_tipostampa==sintetico)
		intestazione.center_just(100);
	else	
		intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

void TStampaSospesi::fine_stampa()
{
	// stampa totale soggetti a fine stampa
  if (_tipostampa==sintetico || _tipostampa==completo)
	{
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		if (_totale > 0 && _totale != _contatore)
		{
			set_footer(3,"TOTALE SOGGETTI STAMPATI %d", _totale);
			printer().formfeed();		
		}			
		reset_footer(); 
	}	
}

bool TStampaSospesi::set_print(int m)
{       
  TPrinter& p = printer();
	p.read_configuration();
	_tipostampa = undefined;
  KEY tasto;
  tasto = _msk->run();
  switch (tasto)
  {
  case F_SINTETICO:
  {
  	_tipostampa = sintetico;
  	_codsez = _codsot = _gruppoold = "**";
  	_contatore = 0;
  	_totale = 0;
	}
  break;
  case F_COMPLETO:
  {
  	_tipostampa = completo;
  	_codsez = _codsot = _gruppoold = "**";
  	_contatore = 0;
  	_totale = 0;
 	}  	
  break;
  case F_ETICHETTE:
  	_tipostampa = (configura_stampante(p, "AT_ETICHETTE", "etichette")) ? etichette : undefined;
  break;
  case F_LETTERE:
  	_tipostampa = (configura_stampante(p, "AT_LETTERE", "lettere")) ? lettere : undefined;
  break;
  }	
	if (_tipostampa != undefined)
  {
    reset_files(); 
    add_file(LF_SOGGETTI);
    // filtro per sezioni selezionate
	  filtra_sezioni();
    _motivo1 = _msk->get(F_MOTIVO1);
    _motivo2 = _msk->get(F_MOTIVO2);
    _motivo3 = _msk->get(F_MOTIVO3);
    _motivo4 = _msk->get(F_MOTIVO4);
    _motivo5 = _msk->get(F_MOTIVO5);
    _motivi = (_motivo1.not_empty() || _motivo2.not_empty() || _motivo3.not_empty() || _motivo4.not_empty() || _motivo5.not_empty());
  	_pergruppo = _msk->get_bool(F_PERGRUPPO);
	  TString256 filtro = "";
	  // filtro per gruppo aziendale
  	if (_pergruppo)
  	{ 	
  		_gruppoazie = _msk->get(F_GRUPPOAZIE);
  		if (_gruppoazie.not_empty())
	  		filtro.format("(90->GRUPPOAZIE == \"%s\")",(const char*)_gruppoazie);
			else	  		
	  		filtro.format("(90->GRUPPOAZIE != \"\")");
	  	filtro << " && ";	
		}	  		
		// filtro per tipo sospensione                                     
    _tiposo = _msk->get(F_TIPO);
    if (_tiposo.not_empty())
    	filtro << '(' << SOG_STATO << "== \"" << _tiposo << "\")";
 		else      
 			filtro << "((TCS->S6 == \"S\") || (TCS->S6 == \"B\"))";     
		_prosstipo = _msk->get(F_PROSSTIPO);
		if (_prosstipo.not_empty())
			filtro << " && (" << SOG_PROS_STATO << "== \"" << _prosstipo << "\")";
    current_cursor()->setfilter(filtro, TRUE);
	  // filtro per categorie
	  _categorie.destroy();
  	const TString16 catpri = _msk->get(F_CAT1);
  	const TString16 catsec = _msk->get(F_CAT2);
  	const TString16 catter = _msk->get(F_CAT3);
  	const TString16 catqua = _msk->get(F_CAT4);
  	const TString16 catqui = _msk->get(F_CAT5);
  	const TString16 catses = _msk->get(F_CAT6);
	  if (catpri.not_empty())
  		_categorie.add((const char*) catpri);
  	if (catsec.not_empty())
  		_categorie.add((const char*) catsec);
  	if (catter.not_empty())
  		_categorie.add((const char*) catter);
  	if (catqua.not_empty())
  		_categorie.add((const char*) catqua);
  	if (catqui.not_empty())
  		_categorie.add((const char*) catqui);
  	if (catses.not_empty())
  		_categorie.add((const char*) catses);
		_dataini = _msk->get_date(F_DATAINI);  		
		_datafin = _msk->get_date(F_DATAFIN);  		
		_sodataini = _msk->get_date(F_SODATAINI);  		
		_sodatafin = _msk->get_date(F_SODATAFIN);  		
		TString80 chiave = "";
		if (_pergruppo && _gruppoazie.empty())  		
		{
			chiave << "90->GRUPPOAZIE|UPPER(90->COGNOME)|UPPER(90->NOME)";
			((TSorted_cursor*)current_cursor())->change_order(chiave);
		}
		current_cursor()->set_filterfunction(filter_func_sospesi);
		reset_print();
		printer().footerlen(0);		
		crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TStampaSospesi::crea_intestazione()
{
  reset_header();
  if (_tipostampa==sintetico || _tipostampa==completo)
  {
  	TString sep(132);
		if (_tipostampa==sintetico)
	  	sep = "ELENCO SINTETICO SOSPESI ";
	  else	
	  	sep = "ELENCO COMPLETO SOSPESI ";
  	if (_tiposo.not_empty())
  		sep << _tiposo;
  	if (_sodataini.ok())
  		sep << " DAL " << _sodataini;	
  	if (_sodatafin.ok())
  		sep << " AL " << _sodatafin;	
		if (_tipostampa==sintetico)
			sep.center_just(100);
		else	
			sep.center_just();
 		set_header(2, "@0g%s", (const char*) sep);
  	const TString16 data_stampa = _data_stampa.string();
  	set_header(2,"@0g%10s", (const char*) data_stampa);
  	sep = "";
  	sep << "Pag. @#";
		if (_tipostampa==sintetico)  	
  		set_header(2, "@90g%s", (const char*) sep);  
  	else	
  		set_header(2, "@120g%s", (const char*) sep);  
  	if (_prosstipo.not_empty() || _dataini.ok() || _datafin.ok())
  	{
	  	sep = "CON PROSSIMO TIPO ";
	  	if (_prosstipo.not_empty())
	  		sep << _prosstipo;
	  	if (_dataini.ok())
	  		sep << " DAL " << _dataini;	
	  	if (_datafin.ok())
	  		sep << " AL " << _datafin;	
			if (_tipostampa==sintetico)
				sep.center_just(100);
			else	
				sep.center_just();
	 		set_header(3, "@0g%s", (const char*) sep);
	 	}	
  	sep = "";
  	if (_tipostampa==sintetico)
  	{
	  	sep.fill('-',100);
  		set_header(4, (const char *) sep);
  		set_header(5,"@0gCodice@9gC.@12gCognome e nome@38gNato il@49gSospensione@63gUltima idon.@77gIntervalli@88gUltima donaz.");
	  	set_header(6,"@0gTessera@38gGr.AB0 Rh@49gPr.tipo@78gSI  AF");
  		set_header(7,"@0g--------@9g--@12g-------------------------@38g----------@49g-------------@63g-------------@77g----------@88g---------- --");
		}
		else  			
		{
	  	sep.fill('-');
  		set_header(4, (const char *) sep);
  		set_header(5,"@0gCodice@9gCognome@35gIndirizzo@66gTel.abitazione@82gTipiz@94gSospensione@105gUltima id.@116gUlt.donaz.");
	  	set_header(6,"@0gTessera@9gNome@66gTel.lavoro@82gzazione@94gData@105gTipo@116gTipo");
	  	set_header(7,"@0gCateg.@9gData nascita@66gTel.altro@94gPr.tipo");
	  	set_header(8,"@94gData pros.@109gInt.SI@118gTot.don.");
	  	set_header(9,"@94gMotivo@109gInt.AF");
  		set_header(10,"@0g--------@9g-------------------------@35g------------------------------@66g--------------@82g------------@94g----------@105g----------@116g----------");
		}
  	printer().footerlen(3);
	}  	
}

bool TStampaSospesi::user_create()
{
  _rel = new TRelation(LF_SOGGETTI);
  _rel->add("TCS", "CODTAB==STATO",1,0,ALIAS_TCS);
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,0,ALIAS_LCPDOM);
  _rel->add("GAZ", "CODTAB==GRUPPOAZIE",1,0,ALIAS_GAZ);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM");
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT");
  _contsan = new TLocalisamfile(LF_CONTSAN);
	add_cursor(new TSorted_cursor(_rel,"90->CODSEZ|90->CODSOT|90->COGNOME|90->NOME","",3));
  _msk = new TMask("at2100a");
  TConfig config(CONFIG_STUDIO);
  TString16 etformato = config.get("EtFormato");
  _etlarghezza = config.get_int("EtLarghezza");
  _etcolonne = config.get_int("EtColonne");
	_form_eti = new TEti_sospesi_form(etformato);
	_form_let = new TEti_sospesi_form("ATLETTER");
  return TRUE;
}

bool TStampaSospesi::user_destroy()
{
	delete _form_let;
  delete _form_eti;
  delete _msk;
  delete _contsan;
  delete _rel;
  return TRUE;
}

int at2100(int argc, char* argv[])
{
  TStampaSospesi a;
  a.run(argc, argv, "Soggetti sospesi");
  return 0;
}
