#include <form.h>
#include <mask.h>
#include <printapp.h>

#include "soggetti.h"
#include "sezioni.h"
#include <comuni.h>

#include "at2.h"
#include "at2200a.h"
#include "atlib.h"

#define ALIAS_CTD			700		// categoria donatori
#define ALIAS_CTN1		701		// categoria non donatori 1
#define ALIAS_CTN2		702		// categoria non donatori 2

#define ALIAS_TCS 		200		// tipi/esiti controlli sanitari
#define ALIAS_LDN 		400		// punti di prelievo
#define ALIAS_LDNDON	401		// punti di prelievo

#define ALIAS_LCPDOM	100		// localita' postale di domicilio
#define ALIAS_COMDOM 	501		// comune di domicilio
#define ALIAS_COMNAS 	500		// comune di nascita

#define ALIAS_BNZ			600		// benemerenze
#define ALIAS_GAZ			300   // gruppi aziendale

#define ETI_COGNOME 3			// numero campo cognome e nome in etichetta (deve essere sempre cosi') per famiglie

enum ts { undefined = 0, elenco = 1, etichette = 2 };

// definizione form per etichette
class TEti_pereta_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TEti_pereta_form(): TForm() {};
  TEti_pereta_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TEti_pereta_form() {};
};

class TStampaPerEta : public TPrintapp
{
  static bool filter_func_pereta(const TRelation* rel);

  TRelation*        _rel;
  TMask*            _msk;
  TEti_pereta_form*	_form_eti;
  TAssoc_array			_categorie;
  TParagraph_string _cognome_nome, _dencom;
  TDate 						_data_stampa;
  TDate							_dataini, _datafin;
  ts								_tipostampa;
  TString16					_codsez, _codsot;
	int								_etlarghezza, _etcolonne, _contatore, _totale;  
  bool							_stampa80;
	
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page (int file, int counter);
  virtual print_action postprocess_print(int file, int counter);

public:
  void crea_intestazione();
  void filtra_sezioni();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  void footer_sezione();
  void fine_stampa();
  TMask& app_mask() { return *_msk; }

  TStampaPerEta() : _data_stampa(TODAY), _cognome_nome("",25), _dencom("",50) {}
};

HIDDEN inline TStampaPerEta& app() { return (TStampaPerEta&) main_app(); }

TCursor* TEti_pereta_form::cursor() const { return app().current_cursor(); }

TRelation* TEti_pereta_form::relation() const { return cursor()->relation(); }

void TStampaPerEta::filtra_sezioni()
{
 	const TString16 sezini = _msk->get(F_SEZINI);
 	const TString16 sotini = _msk->get(F_SOTINI);
 	const TString16 sezfin = _msk->get(F_SEZFIN);
 	const TString16 sotfin = _msk->get(F_SOTFIN);
  TRectype da(LF_SOGGETTI);
  TRectype a(LF_SOGGETTI);   
	if (sezini.not_empty())  
		da.put(SOG_CODSEZ, sezini);
  if (sotini.not_empty())
		da.put(SOG_CODSOT, sotini);
  if (sezfin.not_empty())
		a.put(SOG_CODSEZ, sezfin);
  if (sotfin.not_empty())
		a.put(SOG_CODSOT, sotfin);
	current_cursor()->setregion(da, a);
}	

void TStampaPerEta::set_page(int file, int cnt)
{
	switch (_tipostampa)
	{
		case etichette:
		{
			TPrint_section& corpo = _form_eti->get_body();
			for (int r=1;r<=_etcolonne;r++)
			{                     
				if (current_cursor()->pos()<current_cursor()->items())
				{ 
					force_setpage(TRUE);
					corpo.update();
  				for (word i = 0; i < corpo.height(); i++)
  				{
  					TPrintrow& riga = corpo.row(i);
  					TString256 riga1 = riga.row();
						riga1.cut(_etlarghezza);
  					int colonna = ((r-1)*_etlarghezza);
  					TString16 formato;
  					formato << '@' << colonna << "g";
  					riga1.insert(formato,0);
  					set_row(i+1,riga1);
					}
					if (r < _etcolonne)  			
						++(*current_cursor());
				}					
			}				
		}
		break;
		case elenco:
		{
			if (_stampa80)
			{
				set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
				set_row(2,"@6g@S", FLD(LF_SOGGETTI,SOG_CATDON));
				set_row(3,"@2g@S", FLD(LF_SOGGETTI,SOG_TESSAVIS));
				
				set_row(1,"@12g#a", &_cognome_nome);
				set_row(2,"@12g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
				set_row(3,"@12g@S", FLD(LF_SOGGETTI,SOG_DOM_INDIR));
				set_row(4,"@12g#a", &_dencom);

				set_row(1,"@62g@S", FLD(LF_SOGGETTI,SOG_TELABI));
				set_row(2,"@62g@S", FLD(LF_SOGGETTI,SOG_TELLAV));
				set_row(3,"@62g@S", FLD(LF_SOGGETTI,SOG_TELALT));
				
				set_row(5,"");
			} 
			else
			{		
				set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
				set_row(2,"@0g@8,rs", FLD(LF_SOGGETTI,SOG_TESSAVIS));
				set_row(1,"@9g@S", FLD(LF_SOGGETTI,SOG_CATDON));
				set_row(1,"@12g#a", &_cognome_nome);
				set_row(1,"@38g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
				set_row(1,"@49g@S", FLD(LF_SOGGETTI,SOG_DOM_INDIR));
				set_row(1,"@100g@S", FLD(LF_SOGGETTI,SOG_TELABI));
				set_row(1,"@116g@S", FLD(LF_SOGGETTI,SOG_TELLAV));
				set_row(2,"@49g#a", &_dencom);
				set_row(2,"@116g@S", FLD(LF_SOGGETTI,SOG_TELALT));
			}				
		}
		break;					
	}	
}

bool TStampaPerEta::filter_func_pereta(const TRelation* rel)
{   
	bool filtrato = TRUE;
	TLocalisamfile& sog = rel->lfile();
	// filtro per categorie
	TAssoc_array& categorie = app()._categorie;
	if (categorie.items() != 0)
	{
		const TString16 cat = sog.get(SOG_CATDON);
		filtrato = categorie.is_key((const char*) cat);
	}	
	// filtro per data di nascita
	if (filtrato)		
	{
		const TDate datanasc = sog.get(SOG_DATANASC);
  	if (app()._dataini.ok())
  		filtrato = datanasc >= app()._dataini;
  	if (app()._datafin.ok())
  		filtrato = (filtrato) && (datanasc <= app()._datafin);
  }  	
  return filtrato;
}

void TStampaPerEta::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(132);
	intestazione = "Sezione: ";
	intestazione << codsez;
	intestazione << '/';
	intestazione << codsot;
	intestazione << ' ';
	intestazione << densez;
	if (densot.not_empty())
	{
		intestazione << '/';
		intestazione << densot;
	}		
	if (_stampa80)
		intestazione.center_just(80);
	else		
		intestazione.center_just(132);		
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

bool TStampaPerEta::preprocess_page(int file, int counter)
{
	if (_tipostampa == elenco)
	{
		TRectype recsog = current_cursor()->curr();
		TString80 nome = recsog.get(SOG_COGNOME);
		nome << ' ';
		nome << recsog.get(SOG_NOME);
		_cognome_nome = nome;
		TString256 localita = "";
		localita = recsog.get(SOG_DOM_CAP);			
		localita << ' ';
		if (recsog.get(SOG_DOM_CODLOC).not_empty())
		{
			localita << current_cursor()->curr(-ALIAS_LCPDOM).get("S0");
			localita << " - ";		
		}
		localita << current_cursor()->curr(LF_COMUNI).get(COM_DENCOM);
		localita.trim();
		_dencom = localita;
		// salto pagina se cambio sezione
		const TString16 codsez = recsog.get(SOG_CODSEZ);		
		const TString16 codsot = recsog.get(SOG_CODSOT);		
		if ((_codsez!=codsez)||(_codsot!=codsot))
		{
			if (_codsez != "**")
				footer_sezione();
			_contatore = 0;				
			_codsez = codsez;
			_codsot = codsot;
			header_sezione(codsez, codsot);
		}
		if (printer().rows_left()<5)
			printer().formfeed();
	}		
	if (_tipostampa==etichette)		
		if (printer().rows_left() < _form_eti->get_body().height())
			printer().formfeed();
	_contatore++;			
	_totale++;
  return TRUE;
}

print_action TStampaPerEta::postprocess_print(int file, int counter)
{ 
	if (_contatore > 0)
		footer_sezione();
	fine_stampa();	
	return NEXT_PAGE;
}

void TStampaPerEta::footer_sezione()
{
	// stampa totale soggetti appartenenti alla sezione
  if (_tipostampa==elenco)
	{
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		set_footer(3,"TOTALE SOGGETTI STAMPATI %d", _contatore);
		printer().formfeed(); 
		reset_footer();
	}	
}
                          
void TStampaPerEta::fine_stampa()
{
	// stampa totale soggetti a fine stampa
  if (_tipostampa==elenco)
	{
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		if (_totale > 0 && _totale != _contatore)
		{
			set_footer(3,"TOTALE SOGGETTI STAMPATI %d", _totale);
			printer().formfeed();		
		}			
		reset_footer(); 
	}	
}

bool TStampaPerEta::set_print(int m)
{
  TPrinter& p = printer();
	p.read_configuration();
	_tipostampa = undefined;
  KEY tasto;
  tasto = _msk->run();
  switch (tasto)
  {
  case F_ELENCO:
  {
  	_tipostampa = elenco;
  	_codsez = "**";
  	_codsot = "**";
  	_contatore = 0;
  	_totale = 0;
  }	
  break;
  case F_ETICHETTE:
  	_tipostampa = (configura_stampante(p, "AT_ETICHETTE", "etichette")) ? etichette : undefined;
	break;
  }	
	if (_tipostampa != undefined)
  {
    reset_files(); 
    add_file(LF_SOGGETTI);
		filtra_sezioni();
    _dataini = _msk->get(F_DATAINI);
    _datafin = _msk->get(F_DATAFIN);
    _categorie.destroy();
	  const TString16 catpri = _msk->get(F_CAT1);
	  const TString16 catsec = _msk->get(F_CAT2);
	  const TString16 catter = _msk->get(F_CAT3);
	  const TString16 catqua = _msk->get(F_CAT4);
	  const TString16 catqui = _msk->get(F_CAT5);
	  const TString16 catses = _msk->get(F_CAT6);
	  if (catpri.not_empty())
	  	_categorie.add((const char*) catpri);
	  if (catsec.not_empty())
	  	_categorie.add((const char*) catsec);
	  if (catter.not_empty())
	  	_categorie.add((const char*) catter);
	  if (catqua.not_empty())
	  	_categorie.add((const char*) catqua);
	  if (catqui.not_empty())
	  	_categorie.add((const char*) catqui);
	  if (catses.not_empty())
	  	_categorie.add((const char*) catses);
		current_cursor()->set_filterfunction (filter_func_pereta);
		reset_print();
		printer().footerlen(0);		
    crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TStampaPerEta::crea_intestazione()
{
  reset_header();
  if (_tipostampa == elenco)
  {
  	TString sep(132);
  	sep = "ELENCO PER DATA DI NASCITA";
  	TDate data = _msk->get(F_DATAINI);
  	if (data.ok())
  	{
  		sep << " dal ";
  		sep << data.string();
		}
  	data = _msk->get(F_DATAFIN);
  	if (data.ok())
  	{
  		sep << " al ";
  		sep << data.string();
		}         
		if (_stampa80)
			sep.center_just(80);
		else		
			sep.center_just(132);		
  	set_header(2, "@0g%s", (const char*) sep);
  	TString16 data_stampa = _data_stampa.string();
  	set_header(2,"@0g%10s", (const char*) data_stampa);
  	sep = "";
  	sep << "Pag. @#";
  	if (_stampa80)
  		set_header(2, "@73g%s", (const char*) sep);  
		else  		
			set_header(2, "@110g%s", (const char*) sep);  		
  	sep = "";
  	sep.fill('-');
  	set_header(3, (const char *) sep);
		if (_stampa80)
		{
			_cognome_nome.set_width(51);
	  	set_header(4,"@0gCodice@9gCognome e nome@62gTelefono abit.");
  		set_header(5,"@0gCat.@9gData di nascita@62gTelefono lavoro");
  		set_header(6,"@0gTessera@9gIndirizzo@62gTelefono altro");
  		set_header(7,"@0g--------@9g--------------------------------------------------@62g---------------");
		}
		else
		{
			_cognome_nome.set_width(25);
  		set_header(4,"@0gCodice@9gC.@12gCognome e nome@38gNato il@49gIndirizzo@100gTelefono abit.@116gTelefono lavoro");
  		set_header(5,"@0gTessera@49gCAP/Localit/Comune/Prov.@116gTelefono altro");
  		set_header(6,"@0g--------@9g--@12g-------------------------@38g----------@49g--------------------------------------------------@100g---------------@116g---------------");
		}  	
  	printer().footerlen(3);
	}  	
}

bool TStampaPerEta::user_create()
{
  _rel = new TRelation(LF_SOGGETTI);
  _rel->add("TCS", "CODTAB==STATO",1,0,ALIAS_TCS);
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,0,ALIAS_LCPDOM);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM",1,0,ALIAS_COMDOM);
  _rel->add(LF_MEDICI, "CODMED==CODMED");
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT");
  add_cursor(new TCursor(_rel, "", 3));
  TConfig config(CONFIG_STUDIO);
  TString16 etformato = config.get("EtFormato");
  _etlarghezza = config.get_int("EtLarghezza");
  _etcolonne = config.get_int("EtColonne");
	_form_eti = new TEti_pereta_form(etformato);
  _msk = new TMask("at2200a");
  _stampa80 = config.get_bool("Stampa80");
  return TRUE;
}

bool TStampaPerEta::user_destroy()
{
  delete _form_eti;
  delete _msk;
  delete _rel;
  return TRUE;
}

int at2200(int argc, char* argv[])
{
  TStampaPerEta a;
  a.run(argc, argv, "Soggetti per data nascita");
  return 0;
}
