#include <form.h>
#include <mask.h>
#include <printapp.h>
#include <recarray.h>

#include "soggetti.h"
#include "donaz.h"
#include "benem.h"
#include "sezioni.h"
#include <comuni.h>

#include "at2.h"
#include "at2400a.h"

#define ALIAS_CTD			700		// categoria donatori
#define ALIAS_CTN1		701		// categoria non donatori 1
#define ALIAS_CTN2		702		// categoria non donatori 2

#define ALIAS_TCS 		200		// tipi/esiti controlli sanitari
#define ALIAS_LDN 		400		// punti di prelievo

#define ALIAS_LCPDOM	100		// localita' postale di domicilio
#define ALIAS_COMDOM 	501		// comune di domicilio
#define ALIAS_COMNAS 	500		// comune di nascita

#define ALIAS_BNZ			600		// benemerenze

enum ts { undefined = 0, elenco = 1, schede = 2 };

// definizione form 
class TModificati_form : public TForm
{
public:
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TModificati_form(): TForm() {};
  TModificati_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TModificati_form() {};
};

class TStampaModificati : public TPrintapp
{
  static bool filter_func_modificati(const TRelation* rel);

  TRelation*        		_rel;
  TMask*            		_msk;
	TLocalisamfile* 			_donaz;
	TLocalisamfile* 			_benem;
	TRecord_array*				_sdonazioni;
	TRecord_array*  			_sbenemerenze;
  TModificati_form*			_form_sch;
  TAssoc_array					_categorie;
  TParagraph_string 		_cognome_nome, _dencom;
  TDate 								_data_stampa, _dataini, _datafin;
  ts										_tipostampa;
  TString16							_codsez, _codsot, _utente;
  TString80							_eledon;
  TString80							_eleben1,_eleben2;
  int										_lenpage, _schxpag, _totale, _contatore;
  TString16 						_schformato;
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page (int file, int counter);
  virtual print_action postprocess_print(int file, int counter);

public:
  void crea_intestazione();
  void filtra_sezioni();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  void footer_sezione();
  void fine_stampa();
  TMask& app_mask() { return *_msk; }

  TStampaModificati() : _data_stampa(TODAY), _cognome_nome("",25), _dencom("",50) {}
};

HIDDEN inline TStampaModificati& app() { return (TStampaModificati&) main_app(); }

TCursor* TModificati_form::cursor() const { return app().current_cursor(); }

TRelation* TModificati_form::relation() const { return cursor()->relation(); }

print_action TStampaModificati::postprocess_print(int file, int counter)
{ 
	if ((_tipostampa == schede) && (_schxpag > 1))
		printer().formlen(_lenpage);
	else
	{
		if (_contatore > 0)
			footer_sezione();
		fine_stampa();	
	}	
	return NEXT_PAGE;
}

void TStampaModificati::filtra_sezioni()
{
 	const TString16 sezini = _msk->get(F_SEZINI);
 	const TString16 sotini = _msk->get(F_SOTINI);
 	const TString16 sezfin = _msk->get(F_SEZFIN);
 	const TString16 sotfin = _msk->get(F_SOTFIN);
  TRectype da(LF_SOGGETTI);
  TRectype a(LF_SOGGETTI);   
  if (sezini.not_empty())
		da.put(SOG_CODSEZ, sezini);
  if (sotini.not_empty())
		da.put(SOG_CODSOT, sotini);
  if (sezfin.not_empty())
		a.put(SOG_CODSEZ, sezfin);
  if (sotfin.not_empty())
		a.put(SOG_CODSOT, sotfin);
	current_cursor()->setregion(da, a);
}	

void TStampaModificati::footer_sezione()
{
	// stampa totale soggetti appartenenti alla sezione
  if (_tipostampa==elenco)
	{
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		set_footer(3,"TOTALE SOGGETTI STAMPATI %d", _contatore);
		printer().formfeed(); 
		reset_footer();
	}	
}
                          
void TStampaModificati::fine_stampa()
{
	// stampa totale soggetti a fine stampa
  if (_tipostampa==elenco)
	{
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		if (_totale > 0 && _totale != _contatore)
		{
			set_footer(3,"TOTALE SOGGETTI STAMPATI %d", _totale);
			printer().formfeed();		
		}			
		reset_footer(); 
	}	
}

void TStampaModificati::set_page(int file, int cnt)
{
	switch (_tipostampa)
	{
		case schede:
		{
			TPrint_section& corpo = _form_sch->get_body();
			corpo.update_and_print(printmode_normal, TRUE);
			force_setpage(TRUE);
		}
		break;
		
		case elenco:
		{
			set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
			set_row(2,"@0g@8,rs", FLD(LF_SOGGETTI,SOG_TESSAVIS));
			set_row(1,"@9g@S", FLD(LF_SOGGETTI,SOG_CATDON));
			set_row(1,"@12g#a", &_cognome_nome);
			set_row(1,"@38g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
			set_row(1,"@49g@S", FLD(LF_SOGGETTI,SOG_DOM_INDIR));
			set_row(1,"@100g@S", FLD(LF_SOGGETTI,SOG_TELABI));
			set_row(1,"@116g@S", FLD(LF_SOGGETTI,SOG_TELLAV));
			set_row(2,"@49g#a", &_dencom);
			set_row(2,"@116g@S", FLD(LF_SOGGETTI,SOG_TELALT));
			set_row(3,"");				
		}
		break;					
	}	
}

bool TStampaModificati::filter_func_modificati(const TRelation* rel)
{   
	bool filtrato = TRUE;
	TLocalisamfile& sog = rel->lfile(); 
	//filtro per categorie
	TAssoc_array& categorie = app()._categorie;
	if (categorie.items() != 0)
	{
		const TString16 cat = sog.get(SOG_CATDON);
		filtrato = categorie.is_key((const char*) cat);
	}	
	// filtro per data aggiornamento
	if (filtrato && app()._dataini.ok())
  {
  	const TDate dataultagg = sog.get(SOG_DATAULTAGG);
  	filtrato = (dataultagg >= app()._dataini);
	}  	
	if (filtrato && app()._datafin.ok())
  {
  	const TDate dataultagg = sog.get(SOG_DATAULTAGG);
  	filtrato = (dataultagg <= app()._datafin);
	}  	
	if (filtrato && app()._utente.not_empty())
  {
  	const TString16 utenultagg = sog.get(SOG_UTENULTAGG);
  	filtrato = (utenultagg == app()._utente);
	}  	
	
  return filtrato;
}

bool TStampaModificati::preprocess_page(int file, int counter)
{
	TRectype& recsog = current_cursor()->curr();
	if (_tipostampa == elenco)
	{
		TString80 nome = recsog.get(SOG_COGNOME);
		nome << " ";
		nome << recsog.get(SOG_NOME);
		_cognome_nome = nome;
		TString256 localita = "";
		localita = recsog.get(SOG_DOM_CAP);			
		localita << ' ';
		if (recsog.get(SOG_DOM_CODLOC).not_empty())
		{
			localita << current_cursor()->curr(-ALIAS_LCPDOM).get("S0");
			localita << " - ";		
		}
		localita << current_cursor()->curr(LF_COMUNI).get(COM_DENCOM);
		localita.trim();
		_dencom = localita;
		// salto pagina se cambio sezione
		const TString16 codsez = recsog.get(SOG_CODSEZ);		
		const TString16 codsot = recsog.get(SOG_CODSOT);		
		if ((_codsez!=codsez)||(_codsot!=codsot))
		{
			if (_codsez != "**")
				footer_sezione();
			_contatore = 0;				
			_codsez = codsez;
			_codsot = codsot;
			header_sezione(codsez, codsot);
		}
		_contatore++;			
		_totale++;
	}		
	else
	{
		const long codice = recsog.get_long(SOG_CODICE);
		TRectype* key = new TRectype(LF_DONAZ);
		key->put(DON_CODICE, codice);
		int err = _sdonazioni->read(key);
		_eledon = "";
		if (err == NOERR)
		{ 
			for (int r=_sdonazioni->rows(); r>_sdonazioni->rows()-3 && r>0; r--)
			{
				const TRectype& riga = _sdonazioni->row(r);
				const TDate datadon = riga.get(DON_DATADON);
				_eledon << datadon.string();
				_eledon << " ";
				_eledon << riga.get(DON_TIPODON);
				_eledon << " ";
				_eledon << riga.get(DON_LUOGODON);
			}				
		}
	}
  return TRUE;
}

void TStampaModificati::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(132);
	intestazione = "Sezione: ";
	intestazione << codsez;
	intestazione << "/";
	intestazione << codsot;
	intestazione << " ";
	intestazione << densez;
	if (densot.not_empty())
	{
		intestazione << "/";
		intestazione << densot;
	}		
	intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

bool TStampaModificati::set_print(int)
{
	_tipostampa = undefined;
  KEY tasto;
  tasto = _msk->run();
  switch (tasto)
  {
	  case F_ELENCO:
	  {
	  	_tipostampa = elenco;
	  	_codsez = "**";
	  	_codsez = "**";
	  	_totale = 0;
	  	_contatore = 0;
	  }	
	  break;
	  case F_SCHEDE:
	  {
	  	_tipostampa = schede;	
	  	if (_schxpag > 1)
	  		printer().formlen(_lenpage/_schxpag);
	  }		
	  break;
  }	
	if (_tipostampa != undefined)
  {
    reset_files(); 
    add_file(LF_SOGGETTI);
    _dataini = _msk->get_date(F_DATAINI);
    _datafin = _msk->get_date(F_DATAFIN);
    _utente = _msk->get(F_UTENTE);
		filtra_sezioni();
		// filtro per categorie		
    _categorie.destroy();		
  	const TString16 catpri = _msk->get(F_CAT1);
  	const TString16 catsec = _msk->get(F_CAT2);
  	const TString16 catter = _msk->get(F_CAT3);
  	const TString16 catqua = _msk->get(F_CAT4);
  	const TString16 catqui = _msk->get(F_CAT5);
  	const TString16 catses = _msk->get(F_CAT6);
  	if (catpri.not_empty())
  		_categorie.add((const char*) catpri);
  	if (catsec.not_empty())
  		_categorie.add((const char*) catsec);
  	if (catter.not_empty())
  		_categorie.add((const char*) catter);
  	if (catqua.not_empty())
  		_categorie.add((const char*) catqua);
  	if (catqui.not_empty())
  		_categorie.add((const char*) catqui);
  	if (catses.not_empty())
  		_categorie.add((const char*) catses);
		current_cursor()->set_filterfunction(filter_func_modificati);
		reset_print();   
		printer().footerlen(0);
    crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TStampaModificati::crea_intestazione()
{
  reset_header();
  if (_tipostampa == elenco)
  {
  	TString sep(132);
  	sep = "ELENCO MODIFICATI";
  	if (_dataini.ok())
  	{
  		sep << " DAL ";
  		sep << _dataini.string();
		}
  	if (_datafin.ok())
  	{
  		sep << " AL ";
  		sep << _datafin.string();
		}
  	if (_utente.not_empty())
  	{
  		sep << " UTENTE ";
  		sep << _utente;
		}
		
		sep.center_just();
  	set_header(2, "@0g%s", (const char*) sep);
  	TString16 data_stampa = _data_stampa.string();
  	set_header(2,"@0g%10s", (const char*) data_stampa);
  	sep = "";
  	sep << "Pag. @#";
  	set_header(2, "@120g%s", (const char*) sep);  
  	sep = "";
  	sep.fill('-');
  	set_header(3, (const char *) sep);
  	set_header(4,"@0gCodice@9gC.@12gCognome e nome@38gNato il@49gIndirizzo@100gTelefono abit.@116gTelefono lavoro");
  	set_header(5,"@0gTessera@49gCAP/Localit/Comune/Prov.@116gTelefono altro");
  	set_header(6,"@0g--------@9g--@12g-------------------------@38g----------@49g--------------------------------------------------@100g---------------@116g---------------");
		printer().footerlen(3);
	}  	
}

bool TStampaModificati::user_create()
{
  _rel = new TRelation(LF_SOGGETTI);
  _rel->add(LF_DONAZ, "CODICE==CODICE", 1);
  _rel->add(LF_BENEM, "CODICE==CODICE", 1);
  _rel->add("BNZ", "CODTAB==TIPOBEN", 1, LF_BENEM, ALIAS_BNZ);
  _rel->add("CTD", "CODTAB==CATDON",1,0,ALIAS_CTD);
  _rel->add("CTN", "CODTAB==CATNOND1",1,0,ALIAS_CTN1);
  _rel->add("CTN", "CODTAB==CATNOND2",1,0,ALIAS_CTN2);
  _rel->add("TCS", "CODTAB==STATO",1,0,ALIAS_TCS);
  _rel->add("LDN", "CODTAB==PUNTORACC",1,0,ALIAS_LDN);
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,0,ALIAS_LCPDOM);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM",1,0,ALIAS_COMDOM);
  _rel->add(LF_COMUNI, "COM==COMNASC",1,0,ALIAS_COMNAS);
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT");
  _rel->add(LF_MEDICI, "CODMED==CODMED");
	_donaz 		= new TLocalisamfile(LF_DONAZ);
	_benem 		=	new TLocalisamfile(LF_BENEM);
	_sdonazioni 	= new TRecord_array(LF_DONAZ,DON_PROGDON);
	_sbenemerenze = new TRecord_array(LF_BENEM,BEN_PROGBEN);	
  add_cursor(new TCursor(_rel, "", 3));
  _msk = new TMask("at2400a");
  TConfig config(CONFIG_STUDIO);
  TString16 schformato = config.get("SchFormato");
  _schxpag = config.get_int("SchXPag");
	_form_sch = new TModificati_form(schformato);
	_lenpage = printer().formlen();
  return TRUE;
}

bool TStampaModificati::user_destroy()
{
  delete _form_sch;
  delete _msk;
  delete _sbenemerenze;
  delete _sdonazioni;
  delete _benem;
  delete _donaz;
  delete _rel;     
  return TRUE;
}

int at2400(int argc, char* argv[])
{
  TStampaModificati a;
  a.run(argc, argv, "Soggetti modificati");
  return 0;
}
