#include <form.h>
#include <mask.h>
#include <printapp.h>
#include <utility.h>

#include "soggetti.h"
#include "sezioni.h"

#include "at2.h"
#include "at2500a.h"
#include "atlib.h"

#define ALIAS_LCP	100
#define ALIAS_TCS	200

enum ts { undefined = 0, elenco = 1, etichette = 2 };

// definizione form per etichette
class TEti_idonei_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TEti_idonei_form(): TForm() {};
  TEti_idonei_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TEti_idonei_form() {};
};

class TStampaIdonei : public TPrintapp
{
  TRelation*				_rel;
  TMask*      			_msk;
  TEti_idonei_form*	_form_eti;
  TAssoc_array			_categorie;
  TString16 				_tipoidon;
  TString16					_idon1, _idon2, _idon3, _idon4;
  TParagraph_string _cognome_nome;
  TDate 						_data_stampa;
	ts 	      				_tipostampa;
	TString16					_codsez, _codsot;
  int								_etlarghezza, _etcolonne, _stampa, _numidon, _contatore, _totale;

  static bool filter_func_idonei(const TRelation* rel);
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page(int file, int counter);
  virtual print_action postprocess_print(int file, int counter);

public:
  void crea_intestazione();
  void filtra_sezioni();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  void footer_sezione();
  void fine_stampa();
  TMask& app_mask() { return *_msk; }
  TStampaIdonei() : _data_stampa(TODAY), _cognome_nome("",25) {}
};

HIDDEN inline TStampaIdonei& app() { return (TStampaIdonei&) main_app(); }

TCursor* TEti_idonei_form::cursor() const { return app().current_cursor(); }

TRelation* TEti_idonei_form::relation() const { return cursor()->relation(); }

void TStampaIdonei::fine_stampa()
{
	// stampa totale soggetti a fine stampa
  if (_tipostampa==elenco)
	{
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		if (_totale > 0 && _totale != _contatore)
		{
			set_footer(3,"TOTALE SOGGETTI STAMPATI %d", _totale);
			printer().formfeed();		
		}			
		reset_footer(); 
	}	
}

void TStampaIdonei::footer_sezione()
{
	// stampa totale soggetti appartenenti alla sezione
  if (_tipostampa==elenco)
	{
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		set_footer(3,"TOTALE SOGGETTI STAMPATI %d", _contatore);
		printer().formfeed(); 
		reset_footer();
	}	
}

void TStampaIdonei::filtra_sezioni()
{
 	const TString16 sezini = _msk->get(F_SEZINI);
 	const TString16 sotini = _msk->get(F_SOTINI);
 	const TString16 sezfin = _msk->get(F_SEZFIN);
 	const TString16 sotfin = _msk->get(F_SOTFIN);
  TRectype da(LF_SOGGETTI);
  TRectype a(LF_SOGGETTI);   
  if (sezini.not_empty())
		da.put(SOG_CODSEZ, sezini);
  if (sotini.not_empty())
		da.put(SOG_CODSOT, sotini);
  if (sezfin.not_empty())
		a.put(SOG_CODSEZ, sezfin);
  if (sotfin.not_empty())
		a.put(SOG_CODSOT, sotfin);
	current_cursor()->setregion(da, a);
}	

void TStampaIdonei::set_page(int file, int cnt)
{
	// costruzione etichette
	switch (_tipostampa)
	{
		case etichette:
		{ 
			TPrint_section& corpo = _form_eti->get_body();
			for (int r=1;r<=_etcolonne;r++)
			{                     
				if (current_cursor()->pos()<current_cursor()->items())
				{ 
					force_setpage(TRUE);
					corpo.update();
  				for (word i = 0; i < corpo.height(); i++)
  				{
  					TPrintrow& riga = corpo.row(i);
  					TString256 riga1 = riga.row();
						riga1.cut(_etlarghezza);
  					int colonna = ((r-1)*_etlarghezza);
  					TString16 formato;
  					formato << '@' << colonna << "g";
  					riga1.insert(formato,0);
  					set_row(i+1,riga1);
					}
					if (r < _etcolonne)  			
						++(*current_cursor());
				}					
			}				
		}
		break;
		case elenco:
		{
			set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
			set_row(2,"@0g@8,rs", FLD(LF_SOGGETTI,SOG_TESSAVIS));
			set_row(1,"@9g@S", FLD(LF_SOGGETTI,SOG_CATDON));
			set_row(1,"@12g#a", &_cognome_nome);
			set_row(1,"@38g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
			set_row(2,"@38g@S", FLD(LF_SOGGETTI,SOG_GRUPPOAB0));
			set_row(2,"@45g@S", FLD(LF_SOGGETTI,SOG_RHANTID));
			set_row(1,"@49g@ld", FLD(LF_SOGGETTI,SOG_DATASTATO));	// potremmo mettere anche datastato
			set_row(1,"@60g@S", FLD(LF_SOGGETTI,SOG_IDON1));
			set_row(1,"@63g@S", FLD(LF_SOGGETTI,SOG_IDON2));
			set_row(2,"@60g@S", FLD(LF_SOGGETTI,SOG_IDON3));
			set_row(2,"@63g@S", FLD(LF_SOGGETTI,SOG_IDON4));
			set_row(2,"@52g@S", FLD(LF_SOGGETTI,SOG_STATO));
			set_row(1,"@66g@pn", FLD(LF_SOGGETTI,SOG_INTSI,"###"));
			set_row(2,"@66g@pn", FLD(LF_SOGGETTI,SOG_INTAF,"###"));
			set_row(1,"@70g@ld", FLD(LF_SOGGETTI,SOG_DATAULTDON));
			set_row(2,"@73g@S", FLD(LF_SOGGETTI,SOG_TIPOULTDON));
			set_row(3,"");
		}
		break;
	}					
}

bool TStampaIdonei::filter_func_idonei(const TRelation * rel)
{   
	bool filtrato = TRUE;
	TLocalisamfile& sog = rel->lfile(); 
	// filtro per categorie
	TAssoc_array& categorie = app()._categorie;
	if (categorie.items() != 0)
	{
		const TString16 catsog = sog.get(SOG_CATDON);
		filtrato = categorie.is_key((const char*) catsog);
	}	
	// filtro per idoneit
	if (filtrato)
	{   
		TAssoc_array idoneita;
		TString16 idonsog = sog.get(SOG_IDON1);
  	if (idonsog.not_empty())
  		idoneita.add((const char*) idonsog);
		idonsog = sog.get(SOG_IDON2);
  	if (idonsog.not_empty())
  		idoneita.add((const char*) idonsog);
		idonsog = sog.get(SOG_IDON3);
  	if (idonsog.not_empty())
  		idoneita.add((const char*) idonsog);
		idonsog = sog.get(SOG_IDON4);
  	if (idonsog.not_empty())
  		idoneita.add((const char*) idonsog);
		if (idoneita.items() != 0)
		{                             
			if (app()._stampa < 2)
			{                            
	  		if (app()._idon1.not_empty())
					filtrato = idoneita.is_key((const char*) app()._idon1);
				if (app()._idon2.not_empty())
					filtrato = filtrato && idoneita.is_key((const char*) app()._idon2);
				if (app()._idon3.not_empty())
					filtrato = filtrato && idoneita.is_key((const char*) app()._idon3);
				if (app()._idon4.not_empty())
					filtrato = filtrato && idoneita.is_key((const char*) app()._idon4);
				if (app()._stampa == 1)
					filtrato = (filtrato && (app()._numidon == idoneita.items()));
			}
			else
			{
	  		if (app()._idon1.not_empty())
					filtrato = idoneita.is_key((const char*) app()._idon1);
				if (app()._idon2.not_empty())
					filtrato = filtrato || idoneita.is_key((const char*) app()._idon2);
				if (app()._idon3.not_empty())
					filtrato = filtrato || idoneita.is_key((const char*) app()._idon3);
				if (app()._idon4.not_empty())
					filtrato = filtrato || idoneita.is_key((const char*) app()._idon4);
			}		
		}				
		else
			if (!(app()._idon1.empty() && app()._idon2.empty() && app()._idon3.empty() && app()._idon4.empty()))
				filtrato = FALSE;
	}
  return filtrato;
}

bool TStampaIdonei::preprocess_page(int file, int counter)
{
	if (_tipostampa == elenco)
	{
		TRectype& recsog = current_cursor()->curr();
		TString80 nome = recsog.get(SOG_COGNOME);
		nome << " ";
		nome << recsog.get(SOG_NOME);
		_cognome_nome = nome;
		// salto pagina se cambio sezione
		const TString16 codsez = recsog.get(SOG_CODSEZ);		
		const TString16 codsot = recsog.get(SOG_CODSOT);		
		if ((_codsez!=codsez)||(_codsot!=codsot))
		{
			if (_codsez != "**")
				footer_sezione();
			_contatore = 0;	
			_codsez = codsez;
			_codsot = codsot;
			header_sezione(codsez, codsot);
		}
	}		
	if (_tipostampa==etichette)		
		if (printer().rows_left() < _form_eti->get_body().height())
			printer().formfeed();
	_contatore++;			
	_totale++;
  return TRUE;
}

print_action TStampaIdonei::postprocess_print(int file, int counter)
{ 
	if (_contatore > 0)
		footer_sezione();
	fine_stampa();	
	return NEXT_PAGE;
}


void TStampaIdonei::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(80);
	intestazione = "Sezione: ";
	intestazione << codsez;
	intestazione << "/";
	intestazione << codsot;
	intestazione << " ";
	intestazione << densez;
	if (densot.not_empty())
	{
		intestazione << "/";
		intestazione << densot;
	}		
	intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

bool TStampaIdonei::set_print(int m)
{ 
  TPrinter& p = printer();
	p.read_configuration();
	_tipostampa = undefined;
  KEY tasto;
  tasto = _msk->run();
  switch (tasto)
  {
  case F_ELENCO:
  {
  	_tipostampa = elenco;
  	_codsez = "**";
  	_codsot = "**";
  	_contatore = 0;
  	_totale = 0;
  }	
  break;
  case F_ETICHETTE:
  	_tipostampa = (configura_stampante(p, "AT_ETICHETTE", "etichette")) ? etichette : undefined;
  break;
  }	
	if (_tipostampa != undefined)
  {
    reset_files(); 
    add_file(LF_SOGGETTI);
    // filtro per sezioni selezionati
	  filtra_sezioni();
		// filtro per tipo idoneit
    _tipoidon = _msk->get(F_TIPO);
    if (_tipoidon.not_empty())
    	current_cursor()->setfilter(format("STATO == \"%s\"",(const char*)_tipoidon));
 		else      
      current_cursor()->setfilter("(TCS->S6 == \"I\") || (TCS->S6 == \"F\")", TRUE);
		// filtro per categorie
    _categorie.destroy();		
  	const TString16 catpri = _msk->get(F_CAT1);
  	const TString16 catsec = _msk->get(F_CAT2);
  	const TString16 catter = _msk->get(F_CAT3);
  	const TString16 catqua = _msk->get(F_CAT4);
  	const TString16 catqui = _msk->get(F_CAT5);
  	const TString16 catses = _msk->get(F_CAT6);
	  if (catpri.not_empty())
  		_categorie.add((const char*) catpri);
  	if (catsec.not_empty())
  		_categorie.add((const char*) catsec);
  	if (catter.not_empty())
  		_categorie.add((const char*) catter);
  	if (catqua.not_empty())
  		_categorie.add((const char*) catqua);
  	if (catqui.not_empty())
  		_categorie.add((const char*) catqui);
  	if (catses.not_empty())
  		_categorie.add((const char*) catses);
		// filtro per idoneita
		_idon1 = _msk->get(F_IDON1);  		
		_idon2 = _msk->get(F_IDON2);  		
		_idon3 = _msk->get(F_IDON3);  		
		_idon4 = _msk->get(F_IDON4);  		
		_stampa = _msk->get_int(F_STAMPA);
		_numidon = 0;
		if (_idon1.not_empty())
			_numidon++;
		if (_idon2.not_empty())
			_numidon++;
		if (_idon3.not_empty())
			_numidon++;
		if (_idon4.not_empty())
			_numidon++;
		current_cursor()->set_filterfunction(filter_func_idonei);
		reset_print();
		crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TStampaIdonei::crea_intestazione()
{
  reset_header();
  if (_tipostampa == elenco)
  {
  	TString sep(80);
  	sep = "ELENCO IDONEI";
  	if (_tipoidon.not_empty())
  	{
  		sep << ": tipo ";
  		sep << _tipoidon;
  		const TString80 dtipo = _msk->get(F_D_TIPO);
  		sep << " ";
  		sep << dtipo;
		}
  	if (_idon1.not_empty())
  	{
  		sep << " per ";
  		sep << _idon1;
		}
  	if (_idon2.not_empty())
  	{
  		sep << " ";
  		sep << _idon2;
		}
  	if (_idon3.not_empty())
  	{
  		sep << " ";
  		sep << _idon3;
		}
  	if (_idon4.not_empty())
  	{
  		sep << " ";
  		sep << _idon4;
		}       
		switch (_stampa)
		{
		case 0:
			sep << " - Tutte queste idon.";
		break;	
		case 1:
			sep << " - Esclusivamente queste idon.";
		break;	
		case 2:
			sep << " - Almeno una idon.";
		break;	
		}
		sep.center_just();
  	set_header(2, "@0g%s", (const char*) sep);
  	const TString16 data_stampa = _data_stampa.string();
  	set_header(2,"@0g%10s", (const char*) data_stampa);
  	sep = "";
  	sep << "Pag. @#";
  	set_header(2, "@70g%s", (const char*) sep);  
  	sep = "";
  	sep.fill('-');
  	set_header(3, (const char *) sep);
  	set_header(4,"@0gCodice@9gC.@12gCognome e nome@38gNato il@49gData stato@60gIdon.@66g SI@70gUltima don.");
  	set_header(5,"@0gTessera@38gGruppo Rh@49g   Stato@66g AF");  
  	set_header(6,"@0g--------@9g--@12g-------------------------@38g----------@49g----------@60g-- --@66g---@70g----------");
  	printer().footerlen(3);
	}  	
}

bool TStampaIdonei::user_create()
{
  _rel = new TRelation(LF_SOGGETTI);
  _rel->add("TCS", "CODTAB==STATO",1,0,ALIAS_TCS);
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,0,ALIAS_LCP);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM");
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT");
  add_cursor(new TCursor(_rel, "", 3));
  _msk = new TMask("at2500a");

  TConfig config(CONFIG_STUDIO);
  TString16 etformato = config.get("EtFormato");
  _etlarghezza = config.get_int("EtLarghezza");
  _etcolonne = config.get_int("EtColonne");
	_form_eti = new TEti_idonei_form(etformato);
  return TRUE;
}

bool TStampaIdonei::user_destroy()
{
  delete _form_eti;
  delete _msk;
  delete _rel;
  return TRUE;
}

int at2500(int argc, char* argv[])
{
  TStampaIdonei a;
  a.run(argc, argv, "Soggetti idonei");
  return 0;
}
