#include <form.h>
#include <mask.h>
#include <printapp.h>
#include <utility.h>

#include "soggetti.h"
#include "sezioni.h"

#include "at2.h"
#include "at2600a.h"
#include "atlib.h"

#define ALIAS_LCP	100		// localita' postali
#define ALIAS_CTD	700		// categoria donatori

#define ISCRITTI 'I'
enum ts { undefined = 0, elenco = 1, etichette = 2 };

// definizione form per etichette
class TEti_iscritti_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TEti_iscritti_form(): TForm() {};
  TEti_iscritti_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TEti_iscritti_form() {};
};

class TStampaIscritti : public TPrintapp
{
  TRelation*					_rel;
  TMask*      				_msk;
  TEti_iscritti_form*	_form_eti;
  TAssoc_array				_categorie;
  TParagraph_string 	_cognome_nome;
  TDate 							_data_stampa;
	ts 	      					_tipostampa;
  int 								_contatore, _totale;
	TString16						_codsez, _codsot;
	char								_tipo_iscdim;	// iscritti o dimessi
	TDate								_dataini, _datafin;
  int									_etlarghezza, _etcolonne;
  bool								_attuale;
  int									_numdon;

  static bool filter_func_iscritti(const TRelation* rel);
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page(int file, int counter);
  virtual print_action postprocess_print(int file, int counter);

public:
  void crea_intestazione();
  void filtra_sezioni();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  void footer_sezione();
  void fine_stampa();
  TMask& app_mask() { return *_msk; }
  TStampaIscritti() : _data_stampa(TODAY), _cognome_nome("",25) {}
};

HIDDEN inline TStampaIscritti& app() { return (TStampaIscritti&) main_app(); }

TCursor* TEti_iscritti_form::cursor() const { return app().current_cursor(); }

TRelation* TEti_iscritti_form::relation() const { return cursor()->relation(); }

void TStampaIscritti::filtra_sezioni()
{
 	const TString16 sezini = _msk->get(F_SEZINI);
 	const TString16 sotini = _msk->get(F_SOTINI);
 	const TString16 sezfin = _msk->get(F_SEZFIN);
 	const TString16 sotfin = _msk->get(F_SOTFIN);
  TRectype da(LF_SOGGETTI);
  TRectype a(LF_SOGGETTI);   
  if (sezini.not_empty())
		da.put(SOG_CODSEZ, sezini);
  if (sotini.not_empty())
		da.put(SOG_CODSOT, sotini);
  if (sezfin.not_empty())
		a.put(SOG_CODSEZ, sezfin);
  if (sotfin.not_empty())
		a.put(SOG_CODSOT, sotfin);
	current_cursor()->setregion(da, a);
}	

void TStampaIscritti::set_page(int file, int cnt)
{
	// costruzione etichette
	switch (_tipostampa)
	{
		case etichette:
		{ 
			TPrint_section& corpo = _form_eti->get_body();
			for (int r=1;r<=_etcolonne;r++)
			{                     
				if (current_cursor()->pos()<current_cursor()->items())
				{ 
					force_setpage(TRUE);
					corpo.update();
  				for (word i = 0; i < corpo.height(); i++)
  				{
  					TPrintrow& riga = corpo.row(i);
  					TString256 riga1 = riga.row();
						riga1.cut(_etlarghezza);
  					int colonna = ((r-1)*_etlarghezza);
  					TString16 formato;
  					formato << '@' << colonna << "g";
  					riga1.insert(formato,0);
  					set_row(i+1,riga1);
					}
					if (r < _etcolonne)  			
						++(*current_cursor());
				}					
			}				
		}
		break;
		case elenco:
		{
			set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
			set_row(1,"@9g@S", FLD(LF_SOGGETTI,SOG_CATDON));
			set_row(1,"@12g#a", &_cognome_nome);
			set_row(1,"@38g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
			set_row(1,"@49g@ld", FLD(LF_SOGGETTI,SOG_DATAISC));
			set_row(1,"@60g@ld", FLD(LF_SOGGETTI,SOG_DATADIM));
			set_row(1,"@71g@8,rs", FLD(LF_SOGGETTI,SOG_TESSAVIS));
		}
		break;
	}					
}

bool TStampaIscritti::filter_func_iscritti(const TRelation * rel)
{   
	bool filtrato = TRUE;
	TLocalisamfile& sog = rel->lfile();
	// filtro per categorie 
	TAssoc_array& categorie = app()._categorie;
	if (categorie.items() != 0)
	{
		const TString16 cat = sog.get(SOG_CATDON);
		filtrato = categorie.is_key((const char*) cat);
		if ((!filtrato) && (!app()._attuale))
		{
			const TString16 catcoll = rel->lfile(-ALIAS_CTD).get("S6");
			filtrato = categorie.is_key((const char*) catcoll);
		}
	}	
	// filtro per date iscrizione/dimissione
	if (filtrato)
	{
		if (app()._tipo_iscdim == ISCRITTI)
		{
			const TDate dataisc = sog.get_date(SOG_DATAISC);
			filtrato = (dataisc >= app()._dataini && dataisc <= app()._datafin);
		}
		else
		{
			const TDate datadim = sog.get(SOG_DATADIM);
			filtrato = (datadim >= app()._dataini && datadim <= app()._datafin);
		}				
	}
	if (app()._numdon > 0 && filtrato)
	{
		const int totdon = sog.get_int(SOG_TOTDON);
		filtrato = (totdon >= app()._numdon);
	}	
  return filtrato;
}

bool TStampaIscritti::preprocess_page(int file, int counter)
{
	if (_tipostampa == elenco)
	{
		TRectype& recsog = current_cursor()->curr();
		TString80 nome = recsog.get(SOG_COGNOME);
		nome << " ";
		nome << recsog.get(SOG_NOME);
		_cognome_nome = nome;
		// salto pagina se cambio sezione
		const TString16 codsez = recsog.get(SOG_CODSEZ);		
		const TString16 codsot = recsog.get(SOG_CODSOT);		
		if ((_codsez!=codsez)||(_codsot!=codsot))
		{
			if (_codsez != "**")
				footer_sezione();
			_contatore = 0;				
			_codsez = codsez;
			_codsot = codsot;
			header_sezione(codsez, codsot);
		}
	}		
	if (_tipostampa==etichette)		
		if (printer().rows_left() < _form_eti->get_body().height())
			printer().formfeed();
	_contatore++;
	_totale++;			
  return TRUE;
}

print_action TStampaIscritti::postprocess_print(int file, int counter)
{ 
	if (_contatore > 0)
			footer_sezione();
	fine_stampa();	
	return NEXT_PAGE;
}

void TStampaIscritti::footer_sezione()
{
	// stampa totale soggetti appartenenti alla sezione
	if (_tipostampa==elenco)
	{
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		set_footer(3,"TOTALE SOGGETTI STAMPATI %d", _contatore);
		printer().formfeed(); 
		reset_footer();
	}	
}

void TStampaIscritti::fine_stampa()
{
	// stampa totale soggetti a fine stampa
  if (_tipostampa==elenco)
	{
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		if (_totale > 0 && _totale != _contatore)
		{
			set_footer(3,"TOTALE SOGGETTI STAMPATI %d", _totale);
			printer().formfeed();		
		}			
		reset_footer(); 
	}	
}

void TStampaIscritti::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(80);
	intestazione = "Sezione: ";
	intestazione << codsez;
	intestazione << "/";
	intestazione << codsot;
	intestazione << " ";
	intestazione << densez;
	if (densot.not_empty())
	{
		intestazione << "/";
		intestazione << densot;
	}		
	intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

bool TStampaIscritti::set_print(int m)
{ 
  TPrinter& p = printer();
	p.read_configuration();
	_tipostampa = undefined;
  KEY tasto;
  tasto = _msk->run();
  switch (tasto)
  {
  case F_ELENCO:
  {
  	_tipostampa = elenco;
  	_codsez = "**";
  	_codsot = "**";
  	_contatore = 0;
  	_totale = 0;
  }	
  break;
  case F_ETICHETTE:
  	_tipostampa = (configura_stampante(p, "AT_ETICHETTE", "etichette")) ? etichette : undefined;
  break;
  }	
	if (_tipostampa != undefined)
  {
    reset_files(); 
    add_file(LF_SOGGETTI);
    // filtro per sezioni
	  filtra_sezioni();
    // filtro per categorie
    _categorie.destroy();    
		const TString16 catpri = _msk->get(F_CAT1);
	  const TString16 catsec = _msk->get(F_CAT2);
	  const TString16 catter = _msk->get(F_CAT3);
	  const TString16 catqua = _msk->get(F_CAT4);
	  const TString16 catqui = _msk->get(F_CAT5);
	  const TString16 catses = _msk->get(F_CAT6);
	  if (catpri.not_empty())
	  	_categorie.add((const char*) catpri);
	  if (catsec.not_empty())
	  	_categorie.add((const char*) catsec);
	  if (catter.not_empty())
	  	_categorie.add((const char*) catter);
	  if (catqua.not_empty())
	  	_categorie.add((const char*) catqua);
	  if (catqui.not_empty())
	  	_categorie.add((const char*) catqui);
	  if (catses.not_empty())
	  	_categorie.add((const char*) catses);
		// filtro per iscritti/dimessi e date	  	
		_tipo_iscdim 	= _msk->get(F_TIPO)[0];
		_dataini 			= _msk->get(F_DATAINI);    
		_datafin 			= _msk->get(F_DATAFIN);    
		_attuale      = _msk->get_bool(F_ATTUALE);    
		_numdon 			= _msk->get_int(F_NUMDON);    
		current_cursor()->set_filterfunction(filter_func_iscritti, TRUE);
		reset_print();
		printer().footerlen(0);		
		crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TStampaIscritti::crea_intestazione()
{
  reset_header();
  if (_tipostampa == elenco)
  {
  	TString sep(80);
  	sep = "";
  	if (_tipo_iscdim == ISCRITTI)
  		sep << "ISCRITTI";
		else  		
  		sep << "DIMESSI";
  	if (_dataini.ok())
  	{
  		sep << " DAL ";
  		sep << _dataini.string();
		}
  	if (_datafin.ok())
  	{
  		sep << " AL ";
  		sep << _datafin.string();
		}
		sep.center_just();
  	set_header(2, "@0g%s", (const char*) sep);
  	const TString16 data_stampa = _data_stampa.string();
  	set_header(2,"@0g%10s", (const char*) data_stampa);
  	sep = "";
  	sep << "Pag. @#";
  	set_header(2, "@70g%s", (const char*) sep);  
  	sep = "";
  	sep.fill('-');
  	set_header(3, (const char *) sep);
  	set_header(4,"@0gCodice@9gC.@12gCognome e nome@38gNato il@49gData isc.@60gData dim.@71gTessera");
  	set_header(5,"@0g--------@9g--@12g-------------------------@38g----------@49g----------@60g----------@71g-------");
  	printer().footerlen(3);
	}  	
}

bool TStampaIscritti::user_create()
{
  _rel = new TRelation(LF_SOGGETTI);
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,0,ALIAS_LCP);
  _rel->add("CTD", "CODTAB==CATDON",1,0,ALIAS_CTD);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM");
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT");
  add_cursor(new TCursor(_rel, "", 3));
  _msk = new TMask("at2600a");
  TConfig config(CONFIG_STUDIO);
  TString16 etformato = config.get("EtFormato");
  _etlarghezza = config.get_int("EtLarghezza");
  _etcolonne = config.get_int("EtColonne");
	_form_eti = new TEti_iscritti_form(etformato);
  return TRUE;
}

bool TStampaIscritti::user_destroy()
{
  delete _form_eti;
  delete _msk;
  delete _rel;
  return TRUE;
}

int at2600(int argc, char* argv[])
{
  TStampaIscritti a;
  a.run(argc, argv, "Iscritti/dimessi");
  return 0;
}
