#include <form.h>
#include <mask.h>
#include <printapp.h>
#include <recarray.h>
#include <utility.h>

#include "donaz.h"
#include "soggetti.h"
#include "sezioni.h"
#include <comuni.h>

#include "at2.h"
#include "at2900a.h"
#include "atlib.h"

#define ALIAS_LCPDOM	100
#define ALIAS_TCS	200
#define ALIAS_COMDOM 	501		// comune di domicilio

enum ts { undefined = 0, elenco = 1, etichette = 2, stampasufile = 3 };

// definizione form per etichette
class TEti_freq_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TEti_freq_form(): TForm() {};
  TEti_freq_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TEti_freq_form() {};
};

class TFrequenza : public TPrintapp
{
  static bool filter_func_freq(const TRelation* rel);

  TRelation*        _rel;
  TMask*            _msk;
 	TLocalisamfile*		_donaz;
	TExternisamfile*  _file;
	TRecord_array*		_sdonazioni;
  TEti_freq_form*	  _form_eti;
  TAssoc_array			_categorie;
  int 							_cur;
  TParagraph_string _cognome_nome, _dencom, _numdonsog, _dataultdon;
  TDate 						_data_stampa;
  TDate							_dataini,_datafin;
  int								_numdon;
  TString16					_tipodon1, _tipodon2, _tipodon3, _tipodon4, _tipodon5, _tipodon;
  ts								_tipostampa;
  TString16					_codsez, _codsot;
  int								_etlarghezza, _etcolonne;  
  bool							_stampa80, _ctrltotale;
  int 							_contatore, _totale;
  char							_condizione, _sesso;
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page (int file, int counter);
  virtual print_action postprocess_print(int file, int counter);

public:
  void crea_intestazione();
  void filtra_sezioni();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  void footer_sezione();
  void fine_stampa();
  TMask& app_mask() { return *_msk; }
  TFrequenza() : _data_stampa(TODAY), _cognome_nome("",25), _dencom("",50), _numdonsog("",5), _dataultdon("",10) {}
};
                                             
HIDDEN inline TFrequenza& app() { return (TFrequenza&) main_app(); }

TCursor* TEti_freq_form::cursor() const { return app().current_cursor(); }

TRelation* TEti_freq_form::relation() const { return cursor()->relation(); }

void TFrequenza::filtra_sezioni()
{
 	const TString16 sezini = _msk->get(F_SEZINI);
 	const TString16 sotini = _msk->get(F_SOTINI);
 	const TString16 sezfin = _msk->get(F_SEZFIN);
 	const TString16 sotfin = _msk->get(F_SOTFIN);
  TRectype da(LF_SOGGETTI);
  TRectype a(LF_SOGGETTI);   
	if (sezini.not_empty())  
		da.put(SOG_CODSEZ, sezini);
  if (sotini.not_empty())
		da.put(SOG_CODSOT, sotini);
  if (sezfin.not_empty())
		a.put(SOG_CODSEZ, sezfin);
  if (sotfin.not_empty())
		a.put(SOG_CODSOT, sotfin);
	current_cursor()->setregion(da, a);
}	

void TFrequenza::set_page(int file, int cnt)
{
	switch (_tipostampa)
	{
		case etichette:
		{ 
			TPrint_section& corpo = _form_eti->get_body();
			for (int r=1;r<=_etcolonne;r++)
			{                     
				if (current_cursor()->pos()<current_cursor()->items())
				{ 
					force_setpage(TRUE);
					corpo.update();
  				for (word i = 0; i < corpo.height(); i++)
  				{
  					TPrintrow& riga = corpo.row(i);
  					TString256 riga1 = riga.row();
						riga1.cut(_etlarghezza);
  					int colonna = ((r-1)*_etlarghezza);
  					TString16 formato;
  					formato << '@' << colonna << "g";
  					riga1.insert(formato,0);
  					set_row(i+1,riga1);
					}
					if (r < _etcolonne)  			
						++(*current_cursor());
				}					
			}				
		}
		break;
		case elenco:
		{
			if (_stampa80)
			{                     
				set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
				set_row(2,"@0g@8,rs", FLD(LF_SOGGETTI,SOG_TESSAVIS));
				set_row(3,"@6g@S", FLD(LF_SOGGETTI,SOG_CATDON));
				set_row(1,"@9g#a", &_cognome_nome);
				set_row(3,"@9g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
				set_row(1,"@44g@S", FLD(LF_SOGGETTI,SOG_TELABI));
				set_row(2,"@44g@S", FLD(LF_SOGGETTI,SOG_TELLAV));
				set_row(3,"@44g@S", FLD(LF_SOGGETTI,SOG_TELALT));
				set_row(1,"@60g#a", &_numdonsog);
				set_row(2,"@60g#a", &_dataultdon);
				set_row(4,"");
			}			
			else
			{
				set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
				set_row(2,"@0g@8,rs", FLD(LF_SOGGETTI,SOG_TESSAVIS));
				set_row(1,"@9g@S", FLD(LF_SOGGETTI,SOG_CATDON));
				set_row(1,"@12g#a", &_cognome_nome);
				set_row(1,"@38g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
				set_row(1,"@49g@S", FLD(LF_SOGGETTI,SOG_DOM_INDIR));
				set_row(2,"@49g#a", &_dencom);
				set_row(1,"@100g@S", FLD(LF_SOGGETTI,SOG_TELABI));
				set_row(2,"@100g@S", FLD(LF_SOGGETTI,SOG_TELLAV));
				set_row(3,"@100g@S", FLD(LF_SOGGETTI,SOG_TELALT));
				set_row(1,"@116g#a", &_numdonsog);
				set_row(2,"@116g#a", &_dataultdon);
			}				
		}
		break;					
	}	
}

bool TFrequenza::filter_func_freq(const TRelation* rel)
{   
	bool filtrato = TRUE;
	TLocalisamfile& sog = rel->lfile();
	// filtro per sesso
	if (app()._sesso  != 'I')
		filtrato = (sog.get(SOG_SESSO)[0] == app()._sesso);
	// filtro per categorie
	if (filtrato)
	{
		TAssoc_array& categorie = app()._categorie;
		if (categorie.items() != 0)
		{
			const TString16 cat = sog.get(SOG_CATDON);
			filtrato = categorie.is_key((const char*) cat);
		}	
	}	
	// filtro per numero donazioni nel periodo
	if (filtrato)		
	{
		int numdon = 0;
		filtrato = FALSE;
		if (app()._ctrltotale)
			numdon = sog.get_int(SOG_TOTDON);
		else
		{
			const long codice = sog.get_long(SOG_CODICE);
			TRectype* key = new TRectype(LF_DONAZ);
			key->put(DON_CODICE, codice);
			int err = app()._sdonazioni->read(key);
			if (err == NOERR)
			{
				for (int r=1; r<=app()._sdonazioni->rows(); r++)
				{
					const TRectype& riga = app()._sdonazioni->row(r);
					TDate datadon = riga.get_date(DON_DATADON);
					if ((datadon>=app()._dataini || !app()._dataini.ok()) && (datadon<=app()._datafin || !app()._datafin.ok()))
					{
						TString16 tipodon = riga.get(DON_TIPODON);
					  if ((app()._tipodon.empty()) || (app()._tipodon1 == tipodon) || (app()._tipodon2 == tipodon) || (app()._tipodon3 == tipodon) || (app()._tipodon4 == tipodon) || (app()._tipodon5 == tipodon))
							numdon++;
					}						
				}         
			}	      
		}	
		switch (app()._condizione)
		{
			case 'A':
				filtrato = (numdon >= app()._numdon);
			break;	
			case 'E':
				filtrato = (numdon == app()._numdon);
			break;	
			case 'D':
				filtrato = (numdon > 1 && numdon <= app()._numdon);
			break;	
		}
  }  	
  return filtrato;
}

void TFrequenza::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(132);
	intestazione = "Sezione: ";
	intestazione << codsez;
	intestazione << "/";
	intestazione << codsot;
	intestazione << " ";
	intestazione << densez;
	if (densot.not_empty())
	{
		intestazione << "/";
		intestazione << densot;
	}		
	if (_stampa80)
		intestazione.center_just(80);
	else		
		intestazione.center_just(132);		
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

bool TFrequenza::preprocess_page(int file, int counter)
{
  TRectype recsog = current_cursor()->curr();
	if (_tipostampa == elenco)
	{
		int numdon = 0;
		TDate dataultdon(NULLDATE);
		if (_ctrltotale)
		{
			numdon = recsog.get_int(SOG_TOTDON);
			dataultdon = recsog.get_date(SOG_DATAULTDON);
		}	
		else
		{	
			const long codice = recsog.get_long(SOG_CODICE);
			TRectype* key = new TRectype(LF_DONAZ);
			key->put(DON_CODICE, codice);
			int err = _sdonazioni->read(key);
			if (err == NOERR)
			{
				for (int r=1; r<=_sdonazioni->rows(); r++)
				{
					const TRectype& riga = _sdonazioni->row(r);
					TDate datadon = riga.get_date(DON_DATADON);
					if ((datadon>=_dataini) && (datadon<=_datafin))
					{
						TString16 tipodon = riga.get(DON_TIPODON);
					  if ((app()._tipodon.empty()) || (app()._tipodon1 == tipodon) || (app()._tipodon2 == tipodon) || (app()._tipodon3 == tipodon) || (app()._tipodon4 == tipodon) || (app()._tipodon5 == tipodon))  
					  {
							numdon++;
							dataultdon = datadon;
						}						
					}					
				}         
			}	
		}
		_numdonsog = format("%3d",numdon);
		_dataultdon = dataultdon.string();
		TString80 nome = recsog.get(SOG_COGNOME);
		nome << " ";
		nome << recsog.get(SOG_NOME);
		_cognome_nome = nome;
		TString256 localita = "";
		localita = recsog.get(SOG_DOM_CAP);			
		localita << ' ';
		if (recsog.get(SOG_DOM_CODLOC).not_empty())
		{
			localita << current_cursor()->curr(-ALIAS_LCPDOM).get("S0");
			localita << " - ";		
		}
		localita << current_cursor()->curr(LF_COMUNI).get(COM_DENCOM);
		localita.trim();
		_dencom = localita;
		// salto pagina se cambio sezione
		const TString16 codsez = recsog.get(SOG_CODSEZ);		
		const TString16 codsot = recsog.get(SOG_CODSOT);		
		if ((_codsez!=codsez)||(_codsot!=codsot))
		{
			if (_codsez != "**")
				footer_sezione();
			_contatore = 0;				
			_codsez = codsez;
			_codsot = codsot;
			header_sezione(codsez, codsot);
		}
		if (printer().rows_left() < 4)
			printer().formfeed();
	}		
	if (_tipostampa==etichette)		
		if (printer().rows_left() < _form_eti->get_body().height())
			printer().formfeed();
	if (_tipostampa==stampasufile)
	{
		_file->zero();
		_file->put("CODICE",recsog.get(SOG_CODICE)); 
		_file->put("COGNOME",recsog.get(SOG_COGNOME)); 
		_file->put("NOME",recsog.get(SOG_NOME)); 
		_file->put("SPOSATA",recsog.get(SOG_COGNOME_SP)); 
		_file->put("INDIRIZZO",recsog.get(SOG_DOM_INDIR)); 
		_file->put("PRESSO",recsog.get(SOG_DOM_DISTRI)); 
		TString80 localita = "";
		localita << recsog.get(SOG_DOM_CAP);
		localita << ' ';
		if (recsog.get(SOG_DOM_CODLOC).not_empty())
			localita << current_cursor()->curr(-ALIAS_LCPDOM).get("S0");
		else
			localita << current_cursor()->curr(LF_COMUNI).get(COM_DENCOM);
		localita << ' ' << current_cursor()->curr(LF_COMUNI).get(COM_PROVCOM);
		localita.trim();
		_file->put("LOCALITA",localita); 
		_file->write();	
		do_events();	
	}	
	_contatore++;
	_totale++;			
  return TRUE;
}

print_action TFrequenza::postprocess_print(int file, int counter)
{ 
	if (_contatore > 0)
			footer_sezione();
	fine_stampa();		
	return NEXT_PAGE;
}

void TFrequenza::footer_sezione()
{
	// stampa totale soggetti appartenenti alla sezione
	if (_tipostampa==elenco)
	{
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		set_footer(3,"TOTALE SOGGETTI STAMPATI %d", _contatore);
		printer().formfeed(); 
		reset_footer();
	}	
}

void TFrequenza::fine_stampa()
{
	// stampa totale soggetti a fine stampa
  if (_tipostampa==elenco)
	{
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		if (_totale > 0 && _totale != _contatore)
		{
			set_footer(3,"TOTALE SOGGETTI STAMPATI %d", _totale);
			printer().formfeed();		
		}			
		reset_footer(); 
	}	
}

bool TFrequenza::set_print(int m)
{
  TPrinter& p = printer();
	p.read_configuration();
	_tipostampa = undefined;
  KEY tasto;
  tasto = _msk->run();
  switch (tasto)
  {
  case F_ELENCO:
  {
  	_tipostampa = elenco;
  	_codsez = "**";
  	_codsot = "**";
  	_contatore = 0;
  	_totale = 0;
  }	
 	break;
  case F_ETICHETTE:
  	_tipostampa = (configura_stampante(p, "AT_ETICHETTE", "etichette")) ? etichette : undefined;
 	break;
  case F_FILE:        
  {
  	_tipostampa=stampasufile;
  	if (yesno_box("Azzerare il contenuto del file prima dell'elaborazione?"))
			_file->zap();
  }	
  break;
  }	
	if (_tipostampa != undefined)
  {
    reset_files(); 
    add_file(LF_SOGGETTI);
		filtra_sezioni();
		_condizione = _msk->get(F_CONDIZIONE)[0];
		_numdon  = _msk->get_int(F_NUMDON);
    _dataini = _msk->get(F_DATAINI);
    _datafin = _msk->get(F_DATAFIN);
    _tipodon1 = _msk->get(F_TIPODON1);
    _tipodon2 = _msk->get(F_TIPODON2);
    _tipodon3 = _msk->get(F_TIPODON3);
    _tipodon4 = _msk->get(F_TIPODON4);
    _tipodon5 = _msk->get(F_TIPODON5);
    _tipodon = _tipodon1;
    _tipodon << _tipodon2 << _tipodon3 << _tipodon4 << _tipodon5;
    if ((_tipodon.empty()) && (!_dataini.ok()) && (!_datafin.ok()))
    	_ctrltotale = TRUE;
    else
    	_ctrltotale = FALSE;	
    _categorie.destroy();    
	  const TString16 catpri = _msk->get(F_CAT1);
	  const TString16 catsec = _msk->get(F_CAT2);
	  const TString16 catter = _msk->get(F_CAT3);
	  const TString16 catqua = _msk->get(F_CAT4);
	  const TString16 catqui = _msk->get(F_CAT5);
	  const TString16 catses = _msk->get(F_CAT6);
	  if (catpri.not_empty())
	  	_categorie.add((const char*) catpri);
	  if (catsec.not_empty())
	  	_categorie.add((const char*) catsec);
	  if (catter.not_empty())
	  	_categorie.add((const char*) catter);
	  if (catqua.not_empty())
	  	_categorie.add((const char*) catqua);
	  if (catqui.not_empty())
	  	_categorie.add((const char*) catqui);
	  if (catses.not_empty())
	  	_categorie.add((const char*) catses);
	  _sesso = _msk->get(F_SESSO)[0];
	  const char ordina = _msk->get(F_ORDINA)[0];
		TString256 chiave = "";
		switch (ordina)
		{
			case 'A':
				chiave << "90->CODSEZ|90->CODSOT|UPPER(90->COGNOME)|UPPER(90->NOME)";
			break;
			case 'C':
				chiave << "90->DOM_CAP|UPPER(90->COGNOME)|UPPER(90->NOME)";
			break;
		}
		((TSorted_cursor*) current_cursor())->change_order(chiave);
		current_cursor()->set_filterfunction (filter_func_freq);
		reset_print();
		printer().footerlen(0);		
    crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TFrequenza::crea_intestazione()
{
  reset_header();
  if (_tipostampa == elenco)
  {
  	TString sep(132);
  	sep = "SOGGETTI CHE HANNO FATTO ";
		switch (app()._condizione)
		{
			case 'A':
				sep << "ALMENO ";
			break;	
			case 'E':
				sep << "ESATTAMENTE ";
			break;	
			case 'D':
				sep << "DA 1 A ";
			break;	
		}
  	sep << _numdon;
  	sep << " DONAZIONI";
		if (_tipodon1.not_empty())
			sep << ' ' << _tipodon1;
		if (_tipodon2.not_empty())
			sep << ' ' << _tipodon2;
		if (_tipodon3.not_empty())
			sep << ' ' << _tipodon3;
		if (_tipodon4.not_empty())
			sep << ' ' << _tipodon4;
		if (_tipodon5.not_empty())
			sep << ' ' << _tipodon5;
  	if (_dataini.ok())
			sep << " DAL " << _dataini.string();
		if (_datafin.ok())	
			sep << " AL " << _datafin.string();  	
		if (_stampa80)
			sep.center_just(80);  		
		else			
			sep.center_just(132);
 		set_header(3, "@0g%s", (const char*) sep);
  	TString16 data_stampa = _data_stampa.string();
  	set_header(2,"@0g%10s", (const char*) data_stampa);
  	sep = "";
  	sep << "Pag. @#";
  	if (_stampa80)
  		set_header(2, "@73g%s", (const char*) sep);  
		else  		
			set_header(2, "@110g%s", (const char*) sep);  		
  	set_header(4, "");
		if (_stampa80)
		{  	
  		set_header(5,"@0gCodice@9gCognome e nome@44gTelefono abit.@60gNum.don.");
  		set_header(6,"@0gTessera@44gTelefono lavoro@60gUlt.don.");
  		set_header(7,"@0gCateg.@9gNato il@44gTelefono altro");
  		set_header(8,"@0g--------@9g----------------------------------@44g--------------@60g----------");
			_cognome_nome.set_width(35);
  	}
  	else
  	{
  		set_header(5,"@0gCodice@9gC.@12gCognome e nome@38gNato il@49gIndirizzo@100gTelefono abit.@116gNum.don.");
  		set_header(6,"@0gTessera@49gCAP/Localit/Comune/Prov.@100gTelefono lavoro@116gUlt.don.");
  		set_header(7,"@100gTelefono altro");
  		set_header(8,"@0g--------@9g--@12g-------------------------@38g----------@49g--------------------------------------------------@100g---------------@116g----------");
			_cognome_nome.set_width(25);
		}  		
	 	printer().footerlen(3);
	}  	
}

bool TFrequenza::user_create()
{
  _rel = new TRelation(LF_SOGGETTI);
  _rel->add("TCS", "CODTAB==STATO",1,0,ALIAS_TCS);
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,0,ALIAS_LCPDOM);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM");
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT");
  add_cursor(new TSorted_cursor(_rel, "", "", 3));
  _msk = new TMask("at2900a");
	_file = new TExternisamfile("$file");	 		  
	_donaz = new TLocalisamfile(LF_DONAZ);
	_sdonazioni = new TRecord_array(LF_DONAZ,DON_PROGDON);
  TConfig config(CONFIG_STUDIO);
  TString16 etformato = config.get("EtFormato");
  _etlarghezza = config.get_int("EtLarghezza");
  _etcolonne = config.get_int("EtColonne");
	_form_eti = new TEti_freq_form(etformato);
  _stampa80 = config.get_bool("Stampa80");
  return TRUE;
}

bool TFrequenza::user_destroy()
{
  delete _form_eti;
  delete _sdonazioni;
  delete _donaz;
  delete _file;
  delete _msk;
  delete _rel;
  return TRUE;
}

int at2900(int argc, char* argv[])
{
  TFrequenza a;
  a.run(argc, argv, "Soggetti per frequenza donazione");
  return 0;
}
