#include <form.h>
#include <mask.h>
#include <printapp.h>
#include <utility.h>

#include "soggetti.h"
#include "donaz.h"
#include "sezioni.h"  
#include <comuni.h>

#include "at4.h"
#include "at4100a.h"
#include "at4100b.h"
#include "at4100c.h"
#include "atlib.h"

#define ALIAS_CTD			700		// categoria donatori
#define ALIAS_CTN1		701		// categoria non donatori 1
#define ALIAS_CTN2		702		// categoria non donatori 2

#define ALIAS_TCS 		200		// tipi/esiti controlli sanitari
#define ALIAS_LDN 		400		// punti di prelievo

#define ALIAS_LCPDOM	100		// localita' postale di domicilio
#define ALIAS_COMDOM 	501		// comune di domicilio
#define ALIAS_COMNAS 	500		// comune di nascita

#define ALIAS_BNZ			600		// benemerenze
#define ALIAS_GAZ			300   // gruppi aziendale

#define IDON_SI		"SI"
#define IDON_AF		"AF"

enum ts { undefined = 0, elenco = 1, completo = 2, etichette = 3, cartoline = 4, personale = 5 };

// definizione form per etichette
class TScadenze_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
	TPrint_section& get_head() { return section('H'); } ;
	TPrint_section& get_foot() { return section('F'); } ;
  TScadenze_form(): TForm() {};
  TScadenze_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TScadenze_form() {};
};

class TStampaScadenze : public TPrintapp
{
  TRelation*				_rel;
  TMask*      			_msk;
	TLocalisamfile* 	_donaz;
  TScadenze_form*		_form_eti;
  TScadenze_form*   _form_car;
  TScadenze_form*   _form_per;
  TAssoc_array			_categorie;
  TParagraph_string _cognome_nome, _dencom, _giopredon;
  TDate 						_data_stampa;
  TDate							_dataini, _datafin, _dataultid;
  int								_intminconv;
	bool							_usomodo, _usasez, _registra, _gpd_esclusivo, _proc_esclusivo;
	TString16					_giorni, _punto;
	TString16					_procdon, _modo, _ab01, _ab02, _ab03, _rh;
	ts 	      				_tipostampa;
	TString16					_codsez, _codsot, _orario, _datacart;
	TString80					_invitoper, _data, _presso1, _presso2, _presso3, _presso4;
	TString80					_note, _intest1, _intest2, _intest3, _intest4;
	int								_etlarghezza, _etcolonne;
  int								_contatore, _totfinestampa;

  static bool filter_func_scadenze(const TRelation* rel);
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page(int file, int counter);
  virtual print_action postprocess_print(int file, int counter);
  ts dati_cartoline();

public:
  void crea_intestazione();
  void filtra_sezioni();
  void footer_sezione();
  void fine_stampa();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  void dati_sezione(const TString16 codsez, const TString16 codsot);
  TMask& app_mask() { return *_msk; }
  TStampaScadenze() : _data_stampa(TODAY), _cognome_nome("",25), _dencom("",25), _giopredon("",10) {}
};

HIDDEN inline TStampaScadenze& app() { return (TStampaScadenze&) main_app(); }

TCursor* TScadenze_form::cursor() const { return app().current_cursor(); }

TRelation* TScadenze_form::relation() const { return cursor()->relation(); }

ts TStampaScadenze::dati_cartoline()
{
	TMask msk("at4100b");
 	if (msk.run() == K_ENTER) 
 	{ 
 		_invitoper = msk.get(F_INVITOPER);
 		_data			 = msk.get(F_DATA);
 		_presso1   = msk.get(F_PRESSO1);
 		_presso2   = msk.get(F_PRESSO2);
 		_presso3   = msk.get(F_PRESSO3);
 		_presso4   = msk.get(F_PRESSO4);
 		_datacart  = msk.get(F_DATACART);
 		_orario  	 = msk.get(F_ORARIO);
 		_note   	 = msk.get(F_NOTE);
 		_usasez		 = msk.get_bool(F_USASEZ);
 		return cartoline;
 	}
 	else
 		return undefined;
}

void TStampaScadenze::fine_stampa()
{
	// stampa totale soggetti a fine stampa
	
	reset_footer();
 	printer().footerlen(20);
	TString sep(80);
	sep.fill('-');
	set_footer(1, (const char *) sep);
	set_footer(2,"TOTALE SOGGETTI IN SCADENZA %d", _totfinestampa);
	printer().formfeed();
	reset_footer();
}

print_action TStampaScadenze::postprocess_print(int file, int counter)
{ 
	if (_contatore > 0)
		footer_sezione();
	if (_totfinestampa > 0 && _contatore != _totfinestampa)
		fine_stampa();				
	return NEXT_PAGE;
}

void TStampaScadenze::filtra_sezioni()
{
 	const TString16 sezini = _msk->get(F_SEZINI);
 	const TString16 sotini = _msk->get(F_SOTINI);
 	const TString16 sezfin = _msk->get(F_SEZFIN);
 	const TString16 sotfin = _msk->get(F_SOTFIN);
  TRectype da(LF_SOGGETTI);
  TRectype a(LF_SOGGETTI);   
  if (sezini.not_empty())
		da.put(SOG_CODSEZ, sezini);
  if (sotini.not_empty())
		da.put(SOG_CODSOT, sotini);
  if (sezfin.not_empty())
		a.put(SOG_CODSEZ, sezfin);
  if (sotfin.not_empty())
		a.put(SOG_CODSOT, sotfin);
	current_cursor()->setregion(da, a);
}	

/*
void TStampaScadenze::posizione()
{
	TPrint_section& corpo = _form_car->get_body();
	const TString16 codsez = current_cursor()->curr().get(SOG_CODSEZ);		
	const TString16 codsot = current_cursor()->curr().get(SOG_CODSOT);		
	dati_sezione(codsez, codsot);
	corpo.reset();
  TForm_item& intest1 = corpo.find_field(CAR_INTEST1);
  intest1.set(_intest1);
  TForm_item& intest2 = corpo.find_field(CAR_INTEST2);
  intest2.set(_intest2);
  TForm_item& intest3 = corpo.find_field(CAR_INTEST3);
  intest3.set(_intest3);
  TForm_item& intest4 = corpo.find_field(CAR_INTEST4);
  intest4.set(_intest4);
  TForm_item& invito = corpo.find_field(CAR_INVITOPER);
	invito.set(_invitoper);
  TForm_item& data = corpo.find_field(CAR_DATA);
	data.set(_data);
  TForm_item& presso1 = corpo.find_field(CAR_PRESSO1);
	presso1.set(_presso1);
  TForm_item& presso2 = corpo.find_field(CAR_PRESSO2);
	presso2.set(_presso2);
  TForm_item& presso3 = corpo.find_field(CAR_PRESSO3);
	presso3.set(_presso3);
  TForm_item& presso4 = corpo.find_field(CAR_PRESSO4);
	presso4.set(_presso4);
  TForm_item& orario = corpo.find_field(CAR_ORARIO);
	orario.set(_orario);
  TForm_item& datacart = corpo.find_field(CAR_DATACART);
	datacart.set(_datacart);
  TForm_item& note = corpo.find_field(CAR_NOTE);
	note.set(_note);
	corpo.update();
	bool ripeti = TRUE;
	while (ripeti)
	{
		for (word i = 0; i < corpo.height(); i++)
		{
			TPrintrow& riga = corpo.row(i);
			set_row(i+1,riga);
			printer().print(riga);
		}
		ripeti = !(yesno_box("Il modulo e' posizionato correttamente?"));
	}		
}
*/  
  
void TStampaScadenze::set_page(int file, int cnt)
{
	// costruzione etichette
	switch (_tipostampa)
	{
		case etichette:
		{
			TPrint_section& corpo = _form_eti->get_body();
			for (int r=1;r<=_etcolonne;r++)
			{                     
				if (current_cursor()->pos()<current_cursor()->items())
				{ 
					force_setpage(TRUE);
					corpo.update();
  				for (word i = 0; i < corpo.height(); i++)
  				{
  					TPrintrow& riga = corpo.row(i);
  					TString256 riga1 = riga.row();
						riga1.cut(_etlarghezza);
  					int colonna = ((r-1)*_etlarghezza);
  					TString16 formato;
  					formato << '@' << colonna << "g";
  					riga1.insert(formato,0);
  					set_row(i+1,riga1);
					}
					if (r < _etcolonne)  			
						++(*current_cursor());
				}					
			}				
		}
		break;
		case cartoline:
		{
			TPrint_section& corpo = _form_car->get_body();
			const TString16 codsez = current_cursor()->curr().get(SOG_CODSEZ);		
			const TString16 codsot = current_cursor()->curr().get(SOG_CODSOT);		
			if ((_codsez!=codsez)||(_codsot!=codsot))
			{
				_codsez = codsez;
				_codsot = codsot;
				dati_sezione(codsez, codsot);
				corpo.reset();
			  TForm_item& intest1 = corpo.find_field(CAR_INTEST1);
			  intest1.set(_intest1);
			  TForm_item& intest2 = corpo.find_field(CAR_INTEST2);
			  intest2.set(_intest2);
			  TForm_item& intest3 = corpo.find_field(CAR_INTEST3);
			  intest3.set(_intest3);
			  TForm_item& intest4 = corpo.find_field(CAR_INTEST4);
			  intest4.set(_intest4);
			  TForm_item& invito = corpo.find_field(CAR_INVITOPER);
	  		invito.set(_invitoper);
			  TForm_item& data = corpo.find_field(CAR_DATA);
	  		data.set(_data);
			  TForm_item& presso1 = corpo.find_field(CAR_PRESSO1);
	  		presso1.set(_presso1);
			  TForm_item& presso2 = corpo.find_field(CAR_PRESSO2);
	  		presso2.set(_presso2);
			  TForm_item& presso3 = corpo.find_field(CAR_PRESSO3);
	  		presso3.set(_presso3);
			  TForm_item& presso4 = corpo.find_field(CAR_PRESSO4);
	  		presso4.set(_presso4);
			  TForm_item& orario = corpo.find_field(CAR_ORARIO);
	  		orario.set(_orario);
			  TForm_item& datacart = corpo.find_field(CAR_DATACART);
	  		datacart.set(_datacart);
			  TForm_item& note = corpo.find_field(CAR_NOTE);
	  		note.set(_note);
			}
			corpo.update();
  		for (word i = 0; i < corpo.height(); i++)
  		{
  			TPrintrow& riga = corpo.row(i);
  			set_row(i+1,riga);
			}  			
			force_setpage(TRUE);	// serve perch alla prossima etichetta rifaccia la setpage
														// altrimenti stampa sempre la stessa etichetta
		}
		break;
		case elenco:
		{
			set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"#########"));
			set_row(1,"@10g@S", FLD(LF_SOGGETTI,SOG_CATDON));
			set_row(1,"@13g@S", FLD(LF_SOGGETTI,SOG_COGNOME));
			set_row(1,"@39g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
			set_row(1,"@50g@ld", FLD(LF_SOGGETTI,SOG_DATAPROSSI));
			set_row(1,"@61g@ld", FLD(LF_SOGGETTI,SOG_DATAPROSAF));
			set_row(1,"@72g@ld", FLD(LF_SOGGETTI,SOG_DATAULTDON));
			set_row(1,"@83g@ld", FLD(LF_SOGGETTI,SOG_DATAULTID));
			set_row(1,"@94g@S", FLD(LF_SOGGETTI,SOG_IDON1));
			set_row(1,"@97g@S", FLD(LF_SOGGETTI,SOG_IDON2));
			set_row(1,"@100g@S", FLD(LF_SOGGETTI,SOG_GRUPPOAB0));
			set_row(1,"@105g@S", FLD(LF_SOGGETTI,SOG_TELABI));
			set_row(1,"@120g#a", &_giopredon);
			
			set_row(2,"@0g@8,rs", FLD(LF_SOGGETTI,SOG_TESSAVIS));
			set_row(2,"@13g@S", FLD(LF_SOGGETTI,SOG_NOME));
			set_row(2,"@53g@pn", FLD(LF_SOGGETTI,SOG_INTSI,"###"));
			set_row(2,"@64g@pn", FLD(LF_SOGGETTI,SOG_INTAF,"###"));
			set_row(2,"@76g@S", FLD(LF_SOGGETTI,SOG_TIPOULTDON));
			set_row(2,"@87g@S", FLD(LF_SOGGETTI,SOG_TIPOULTID));
			set_row(2,"@94g@S", FLD(LF_SOGGETTI,SOG_IDON3));
			set_row(2,"@97g@S", FLD(LF_SOGGETTI,SOG_IDON4));
			set_row(2,"@100g@S", FLD(LF_SOGGETTI,SOG_RHANTID));
			set_row(2,"@105g@S", FLD(LF_SOGGETTI,SOG_TELLAV));
			set_row(3,"@105g@S", FLD(LF_SOGGETTI,SOG_TELALT));
			//set_row(3,"");			
		}
		break;
		case completo:
		{
			set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"#########"));
			set_row(2,"@0g@8,rs", FLD(LF_SOGGETTI,SOG_TESSAVIS));

			set_row(1,"@10g#a", &_cognome_nome);
			set_row(3,"@10g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));

			set_row(1,"@36g@S", FLD(LF_SOGGETTI,SOG_DOM_INDIR));
			set_row(2,"@36g#a", &_dencom);
			
			set_row(1,"@62g@S", FLD(LF_SOGGETTI,SOG_TELABI));
			set_row(2,"@62g@S", FLD(LF_SOGGETTI,SOG_TELLAV));
			set_row(3,"@62g@S", FLD(LF_SOGGETTI,SOG_TELALT));
			
			
			set_row(1,"@77g@ld", FLD(LF_SOGGETTI,SOG_DATAPROSSI));
			set_row(2,"@77g@ld", FLD(LF_SOGGETTI,SOG_DATAPROSAF));
			
			set_row(1,"@88g@ld", FLD(LF_SOGGETTI,SOG_DATAULTSI));
			set_row(2,"@88g@ld", FLD(LF_SOGGETTI,SOG_DATAULTAF));
			
			set_row(1,"@99g@ld", FLD(LF_SOGGETTI,SOG_DATAULTID));
			set_row(2,"@99g#a", &_giopredon);

			set_row(3,"@77g@S", FLD(LF_SOGGETTI,SOG_GRUPPOAB0));
			set_row(3,"@81g@S", FLD(LF_SOGGETTI,SOG_RHANTID));
			set_row(3,"@85g@S", FLD(LF_SOGGETTI,SOG_FENOTIPORH));
			set_row(3,"@92g@S", FLD(LF_SOGGETTI,SOG_KELL));
			set_row(3,"@96g@S", FLD(LF_SOGGETTI,SOG_DU));
			set_row(3,"@100g@S", FLD(LF_SOGGETTI,SOG_ANTICORPI));
			set_row(3,"@102g@S", FLD(LF_SOGGETTI,SOG_CMV));
			set_row(4,"");
		}
		break;
		case personale:
		{
			TPrint_section& corpo = _form_per->get_body();
			force_setpage(TRUE);
			corpo.update();
			TString riga1;
			for (word i = 0; i < corpo.height(); i++)
			{
				TPrintrow& riga = corpo.row(i);
				riga1 = riga.row();
				set_row(i+1,riga1);
			}
			
			//force_setpage(TRUE);
			//corpo.update_and_print(TRUE);
		}
		break;
	}					
}

bool TStampaScadenze::filter_func_scadenze(const TRelation* rel)
{   
	bool filtrato = TRUE;
	TLocalisamfile& sog = rel->lfile(); 
	// filtro per categorie
	TAssoc_array& categorie = app()._categorie;
	if (categorie.items() != 0)
	{
		const TString16 cat = sog.get(SOG_CATDON);
		filtrato = categorie.is_key((const char*) cat);
	}	
	// filtro per modalita' di convocazione
	if (filtrato && app()._usomodo)
	{
		const TString16 modconv = sog.get(SOG_MODCONV);
		filtrato = (modconv == app()._modo);
	}
	// filtro per data ultima idoneita'
	if (filtrato && app()._dataultid.ok())
	{
		const TDate dataultid = sog.get(SOG_DATAULTID);
		filtrato = (dataultid >= app()._dataultid);
	}
	// filtro per tipizzazione
	if (filtrato && (app()._ab01.not_empty() || app()._ab02.not_empty() || app()._ab03.not_empty()))
	{
		const TString16 ab0 = sog.get(SOG_GRUPPOAB0);
		filtrato = ((app()._ab01.not_empty() && ab0 == app()._ab01) || (app()._ab02.not_empty() && ab0 == app()._ab02) || (app()._ab03.not_empty() && ab0 == app()._ab03));
	}		
	if (filtrato && app()._rh.not_empty())
	{
		const TString16 rh	= sog.get(SOG_RHANTID);
		filtrato = (rh == app()._rh);
	}			
	// filtro per data di prossima donazione
	if (filtrato)
	{   
		TDate datapross(NULLDATE);
		TDate datanulla(NULLDATE);
		if (app()._proc_esclusivo)
		{
			if (app()._procdon == IDON_SI)
			{
				datapross = sog.get(SOG_DATAPROSAF);					
				filtrato = (datapross == datanulla);
			}	
			else
			{
				datapross = sog.get(SOG_DATAPROSSI);					
				filtrato = (datapross == datanulla);
			}	
		}
		if (filtrato)
		{
			if (app()._procdon == IDON_SI)
				datapross = sog.get(SOG_DATAPROSSI);					
			else
				datapross = sog.get(SOG_DATAPROSAF);
			if (datapross.ok())			
			{
		  	if (app()._dataini.ok())
		  		filtrato = datapross >= app()._dataini;
				if (filtrato)
		  		if (app()._datafin.ok())
		  		{
						if (sog.get_int(SOG_TOTDON) == 0)			
							datapross+=app()._intminconv;	  		
		  			filtrato = datapross <= app()._datafin;
					}	  			
			}
			else
				filtrato = FALSE;	  			
		}		
	}
	// filtro per giorno preferito
	const TString16& giorni = app()._giorni;
	if (filtrato && giorni != "NNNNNNN")
	{
		bool almenoungiorno = FALSE;
		const TString16 giopredon = sog.get(SOG_GIOPREDON);
		if ((giopredon.blank()) && (!app()._gpd_esclusivo))
			almenoungiorno = TRUE;
		else	
			for (int i = 0; i < giorni.len(); i++)
			{
				if (giorni[i] == 'X')
					almenoungiorno = (almenoungiorno || giopredon[i] == 'X');
			}				
		filtrato = almenoungiorno;
	}
	if (filtrato && !app()._punto.blank())
	{
		TDate dataultima(NULLDATE);
		if (app()._procdon == IDON_SI)
			dataultima = sog.get(SOG_DATAULTSI);					
		else
			dataultima = sog.get(SOG_DATAULTAF);
		if (dataultima.ok())
		{	
			TLocalisamfile donaz(LF_DONAZ);
			donaz.setkey(2);
			donaz.zero();
			donaz.put(DON_DATADON, dataultima);
			donaz.put(DON_CODICE, sog.get(SOG_CODICE));
			if (donaz.read() == NOERR)
				filtrato = (donaz.get(DON_LUOGODON) == app()._punto);
			else
				filtrato = (sog.get(SOG_PUNTORACC) == app()._punto);				
		}
		else
			filtrato = (sog.get(SOG_PUNTORACC) == app()._punto);				
	}
  return filtrato;
}

bool TStampaScadenze::preprocess_page(int file, int counter)
{
	TRectype& recsog = current_cursor()->curr();
	if (_tipostampa==elenco || _tipostampa==completo || _tipostampa==personale)
	{
		TRectype& recsog = current_cursor()->curr();
		const TString16 codsez = recsog.get(SOG_CODSEZ);		
		const TString16 codsot = recsog.get(SOG_CODSOT);		
		TString80 nome = recsog.get(SOG_COGNOME);
		nome << " ";
		nome << recsog.get(SOG_NOME);
		nome << " ";
		nome << recsog.get(SOG_COGNOME_SP);
		_cognome_nome = nome;
		TString256 localita = "";
		localita = recsog.get(SOG_DOM_CAP);			
		localita << ' ';
		if (recsog.get(SOG_DOM_CODLOC).not_empty())
		{
			localita << current_cursor()->curr(-ALIAS_LCPDOM).get("S0");
			localita << " - ";		
		}
		localita << current_cursor()->curr(LF_COMUNI).get(COM_DENCOM);
		localita.trim();
		_dencom = localita;
		// giorni preferiti                             
		_giopredon = "LMMGVSD";
		const TString16 giopredon = recsog.get(SOG_GIOPREDON);
		for (int i = 0; i < 7; i++)
		{
			if (giopredon[i] != 'X')
				_giopredon[i] = ' ';
		}				
		// salto pagina se cambio sezione
		if ((_codsez!=codsez)||(_codsot!=codsot))
		{
			if (_codsez != "**")
				footer_sezione();
			_codsez = codsez;
			_codsot = codsot;
			header_sezione(codsez, codsot);
			_contatore = 0;
		}			
		else                        
			if ((printer().rows_left()<4 && _tipostampa==elenco) || (printer().rows_left()<3 && _tipostampa==completo))
				printer().formfeed();
		_contatore++;
		_totfinestampa++;
	}
	if (_registra)
	{
		recsog.put(SOG_DATACONV, _data_stampa);
		recsog.put(SOG_NUMCONV, 0);
		current_cursor()->file().rewrite();
	}
	if (_tipostampa==etichette)		
		if (printer().rows_left() < _form_eti->get_body().height())
			printer().formfeed();
	if (_tipostampa==personale)		
		if (printer().rows_left() < _form_per->get_body().height())
			printer().formfeed();
  return TRUE;
}

void TStampaScadenze::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(132);
	intestazione = "Sezione: ";
	intestazione << codsez;
	intestazione << "/";
	intestazione << codsot;
	intestazione << " ";
	intestazione << densez;
	if (densot.not_empty())
	{
		intestazione << "/";
		intestazione << densot;
	}		
	intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

void TStampaScadenze::footer_sezione()
{
	// stampa totale soggetti a fine pagina
	reset_footer();
	TString sep(80);
	sep.fill('-');
	set_footer(1, (const char *) sep);
	set_footer(2,"SOGGETTI IN SCADENZA PER SEZIONE %d", _contatore);
	printer().formfeed(); 
	reset_footer();
}



void TStampaScadenze::dati_sezione(const TString16 codsez, const TString16 codsot)
{
	_intest1 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ1);
	_intest2 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ2);
	_intest3 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ3);
	_intest4 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ4);
	if (_usasez)
	{
		if (_presso1.blank())
		{
			_presso1 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE1);
			_presso2 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE2);
			_presso3 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE3);
			_presso4 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE4);
		}			
		if (_orario.blank())
			_orario = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVORA);
		if (_note.blank())
			_note = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVNOTE);
	}
	return;
}

bool TStampaScadenze::set_print(int m)
{ 
  TPrinter& p = printer();
	p.read_configuration();
	_tipostampa = undefined;
  KEY tasto;
  tasto = _msk->run();
  switch (tasto)
  {
  case F_ELENCO:
  	_tipostampa = elenco;
  break;
  case F_COMPLETO:
  	_tipostampa = completo;
  break;
  case F_PERSONALE:
  	_tipostampa=personale;
 	break;
  case F_ETICHETTE:
  	_tipostampa = (configura_stampante(p, "AT_ETICHETTE", "etichette")) ? etichette : undefined;
	break;  	
  case F_CARTOLINE:
  	_tipostampa = (configura_stampante(p, "AT_CARTOLINE", "cartoline")) ? dati_cartoline() : undefined;
  break;
  }	
	if (_tipostampa != undefined)
  {
  	_codsez = "**";
  	_codsot = "**";
    reset_files(); 
    add_file(LF_SOGGETTI);
    _dataini 		= _msk->get(F_DATAINI);
    _datafin 		= _msk->get(F_DATAFIN);
    _intminconv = _msk->get_int(F_INTMINCONV);
    _dataultid 	= _msk->get(F_DATAULTID);
    _procdon 		= _msk->get(F_PROCDON1);
    _proc_esclusivo = _msk->get_bool(F_PROC_ESCLUSIVO);
    _usomodo 		= _msk->get_bool(F_USOMODO);
    _modo 	 		= _msk->get(F_MODO);
    _ab01		 		= _msk->get(F_AB01);
    _ab02		 		= _msk->get(F_AB02);
    _ab03		 		= _msk->get(F_AB03);
    _rh			 		= _msk->get(F_RHANTID);
    _punto	 		= _msk->get(F_PUNTO);
    _registra   = _msk->get_bool(F_REGISTRA);
		_giorni 		= "NNNNNNN";    
    if (_msk->get_bool(F_GPD_LUN))
    	_giorni[0] = 'X';
    if (_msk->get_bool(F_GPD_MAR))
    	_giorni[1] = 'X';
    if (_msk->get_bool(F_GPD_MER))
    	_giorni[2] = 'X';
    if (_msk->get_bool(F_GPD_GIO))
    	_giorni[3] = 'X';
    if (_msk->get_bool(F_GPD_VEN))
    	_giorni[4] = 'X';
    if (_msk->get_bool(F_GPD_SAB))
    	_giorni[5] = 'X';
    if (_msk->get_bool(F_GPD_DOM))
    	_giorni[6] = 'X';
		_gpd_esclusivo   = _msk->get_bool(F_GPD_ESCLUSIVO);    	
    // filtro per sezioni selezionati
	  filtra_sezioni();
		// filtro per categorie	
		_categorie.destroy();	
  	const TString16 catpri = _msk->get(F_CAT1);
  	const TString16 catsec = _msk->get(F_CAT2);
  	const TString16 catter = _msk->get(F_CAT3);
  	const TString16 catqua = _msk->get(F_CAT4);
  	const TString16 catqui = _msk->get(F_CAT5);
  	const TString16 catses = _msk->get(F_CAT6);
  	if (catpri.not_empty())
  		_categorie.add((const char*) catpri);
  	if (catsec.not_empty())
  		_categorie.add((const char*) catsec);
  	if (catter.not_empty())
  		_categorie.add((const char*) catter);
  	if (catqua.not_empty())
  		_categorie.add((const char*) catqua);
  	if (catqui.not_empty())
  		_categorie.add((const char*) catqui);
  	if (catses.not_empty())
  		_categorie.add((const char*) catses);
		// filtro per idoneo e non escluso                                     
		//if (_usomodo)
    	//current_cursor()->setfilter(format("(TCS->S6 == \"I\" || TCS->S6 == \"F\") && (ESCLUSO == \"\") && (MODCONV == \"%s\")",(const char*)_modo), TRUE);
		//else
    	current_cursor()->setfilter("((TCS->S6 == \"I\" ) || (TCS->S6 == \"F\")) && (ESCLUSO == \"\")", TRUE);
		// filtro per categorie e altro
		current_cursor()->set_filterfunction(filter_func_scadenze);
 		_contatore = 0;
 		_totfinestampa = 0;
		reset_print();
		printer().footerlen(0);
		crea_intestazione();
    return TRUE;
  }   
  else
    return FALSE;
}

void TStampaScadenze::crea_intestazione()
{
  reset_header();
  if (_tipostampa == elenco || _tipostampa == completo)
  {
  	TString sep(132);
  	sep = "SCADENZE DI DONAZIONE ";
  	TString16 procdon = _msk->get(F_PROCDON1);
 		sep << procdon ;
		TDate data = _msk->get(F_DATAINI);
		if (data.ok())
		{
 			sep << " dal ";
  		sep << data.string();
		}
		data = _msk->get(F_DATAFIN);
		if (data.ok())
		{
 			sep << " al ";
  		sep << data.string();
		}
		if (_ab01.not_empty())
		{
			sep << " Gruppi ";
			sep << _ab01;
		}
		if (_ab02.not_empty())
		{
			sep << ", ";
			sep << _ab02;
		}
		if (_ab03.not_empty())
		{
			sep << ", ";
			sep << _ab03;
		}
		if (_rh.not_empty())
		{
			sep << " - Rh ";
			sep << _rh;
		}
		sep.center_just();
  	set_header(2, "@0g%s", (const char*) sep);
  	TString16 data_stampa = _data_stampa.string();
  	set_header(2,"@0g%10s", (const char*) data_stampa);
  	sep = "";
  	sep << "Pag. @#";
  	set_header(2, "@120g%s", (const char*) sep);  
  	sep = "";
  	sep.fill('-');
  	set_header(3, (const char *) sep);
  	if (_tipostampa == elenco)
  	{
		  set_header(5,"@0gCodice@10gC.@13gCognome@39gNato il@50gData pr.SI@61gData pr.AF@72gData/Tipo@83gData/Tipo@94gIdon.@100gAB0@105gTel. abitaz.@120gGio.pre.");
		  set_header(6,"@0gTessera@13gNome@50gInterv. SI@61gInterv. AF@72gult. don.@83gult. idon.@94gper@100gRh@105gTel. lavoro");  
		  set_header(7,"@105gTel. altro");  
		  set_header(8,"@0g---------@10g--@13g-------------------------@39g----------@50g----------@61g----------@72g----------@83g----------@94g-----@100g----@105g--------------@120g--------");
		}
		if (_tipostampa == completo)
		{
		  set_header(5,"@0gCodice@10gCognome e Nome@36gIndirizzo@62gTelefoni@77gData pr.SI@88gData ult.SI@99g Ult.idon.");
		  set_header(6,"@0gTessera@10gData nasc.@77gData pr.AF@88gData ult.AF@99g Gio.pre.");
		  set_header(7,"@77gGr. Rh. Fen.   Kell Du A CMV");
		  set_header(8,"");
		}
	 	printer().footerlen(3);
	}  	
	if (_tipostampa==personale)
	{
		TPrint_section& testata = _form_per->get_head();
		testata.update();
		for (word i=0; i<testata.height(); i++)
		{
			TPrintrow& riga = testata.row(i);
			TString256 riga1 = riga.row();
			set_header(i+1,riga1);
		}
		TPrint_section& piede = _form_per->get_foot();
		piede.update();
		for (i=0; i<piede.height(); i++)
		{
			TPrintrow& riga = piede.row(i);
			TString256 riga1 = riga.row();
			set_footer(i+1,riga1);
		}
	}
}

bool TStampaScadenze::user_create()
{
  _rel = new TRelation(LF_SOGGETTI);
  _rel->add("TCS", "CODTAB==STATO",1,0,ALIAS_TCS);
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,0,ALIAS_LCPDOM);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM",1,0,ALIAS_COMDOM);
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT");
	_donaz 		= new TLocalisamfile(LF_DONAZ);
  add_cursor(new TCursor(_rel, "", 3));
  _msk = new TMask("at4100a");
  TConfig config(CONFIG_STUDIO);
  TString16 etformato = config.get("EtFormato");
  _etlarghezza = config.get_int("EtLarghezza");
  _etcolonne = config.get_int("EtColonne");
	_form_eti = new TScadenze_form(etformato);
  _form_car = new TScadenze_form("ATCARTO1");
  TString16 performato = config.get("ElencoFormato");
	_form_per = new TScadenze_form(performato);
  return TRUE;
}

bool TStampaScadenze::user_destroy()
{                        
	delete _form_per;
  delete _form_car;
  delete _form_eti;
  delete _msk;
  delete _donaz;
  delete _rel;
  return TRUE;
}

int at4100(int argc, char* argv[])
{
  TStampaScadenze a;
  a.run(argc, argv, "Scadenze di donazione");
  return 0;
}
