#include <form.h>
#include <mask.h>
#include <printapp.h>

#include "soggetti.h"
#include "sezioni.h"
#include <comuni.h>

#include "at4.h"
#include "at4200a.h"
#include "atlib.h"

#define ALIAS_LCP	100
#define ALIAS_TCS	200
#define IDON_SI "SI"

enum ts { undefined = 0, elenco = 1, etichette = 2 };

// definizione form per etichette
class TUrgenze_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TUrgenze_form(): TForm() {};
  TUrgenze_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TUrgenze_form() {};
};

class TStampaUrgenze : public TPrintapp
{
  static bool filter_func_urgenze(const TRelation* rel);

  TRelation*        _rel;
  TMask*            _msk;
  TUrgenze_form*		_form_eti;
  TAssoc_array			_categorie;
  TString16					_tipodon;
  long 							_giorni;
  TParagraph_string _cognome_nome, _dencom;
  TDate 						_data_stampa, _datalimite;
  ts								_tipostampa;
  TString16					_codsez, _codsot;
	int								_etlarghezza, _etcolonne;
	bool							_unica;
	int								_cur;
	int								_contatore, _totale;
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page (int file, int counter);
  virtual print_action postprocess_print(int file, int counter);

public:
  void crea_intestazione();
  void filtra_sezioni();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  void footer_sezione();
  void fine_stampa();
  TMask& app_mask() { return *_msk; }

  TStampaUrgenze() : _data_stampa(TODAY), _cognome_nome("",25), _dencom("",50) {}
};

HIDDEN inline TStampaUrgenze& app() { return (TStampaUrgenze&) main_app(); }

TCursor* TUrgenze_form::cursor() const { return app().current_cursor(); }

TRelation* TUrgenze_form::relation() const { return cursor()->relation(); }

void TStampaUrgenze::filtra_sezioni()
{
 	const TString16 sezini = _msk->get(F_SEZINI);
 	const TString16 sotini = _msk->get(F_SOTINI);
 	const TString16 sezfin = _msk->get(F_SEZFIN);
 	const TString16 sotfin = _msk->get(F_SOTFIN);
  TRectype da(LF_SOGGETTI);
  TRectype a(LF_SOGGETTI);   
  if (sezini.not_empty())
		da.put(SOG_CODSEZ, sezini);
  if (sotini.not_empty())
		da.put(SOG_CODSOT, sotini);
  if (sezfin.not_empty())
		a.put(SOG_CODSEZ, sezfin);
  if (sotfin.not_empty())
		a.put(SOG_CODSOT, sotfin);
	current_cursor()->setregion(da, a);
}	

void TStampaUrgenze::set_page(int file, int cnt)
{
	switch (_tipostampa)
	{                
		case etichette:
		{
			TPrint_section& corpo = _form_eti->get_body();
			for (int r=1;r<=_etcolonne;r++)
			{                     
				if (current_cursor()->pos()<current_cursor()->items())
				{ 
					force_setpage(TRUE);
					corpo.update();
  				for (word i = 0; i < corpo.height(); i++)
  				{
  					TPrintrow& riga = corpo.row(i);
  					TString256 riga1 = riga.row();
						riga1.cut(_etlarghezza);
  					int colonna = ((r-1)*_etlarghezza);
  					TString16 formato;
  					formato << '@' << colonna << "g";
  					riga1.insert(formato,0);
  					set_row(i+1,riga1);
					}
					if (r < _etcolonne)  			
						++(*current_cursor());
				}					
			}				
		}
		break;
		case elenco:
		{
			set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
			set_row(2,"@0g@8,rs", FLD(LF_SOGGETTI,SOG_TESSAVIS));
			
			set_row(1,"@9g@S", FLD(LF_SOGGETTI,SOG_CATDON));
			set_row(1,"@12g#a", &_cognome_nome);

			set_row(1,"@38g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
			set_row(2,"@38g@S", FLD(LF_SOGGETTI,SOG_CODSEZ));
			set_row(2,"@41g@S", FLD(LF_SOGGETTI,SOG_CODSOT));

			set_row(1,"@49g@S", FLD(LF_SOGGETTI,SOG_GRUPPOAB0));
			set_row(1,"@53g@S", FLD(LF_SOGGETTI,SOG_RHANTID));
			set_row(1,"@57g@S", FLD(LF_SOGGETTI,SOG_FENOTIPORH));
			set_row(1,"@64g@S", FLD(LF_SOGGETTI,SOG_KELL));
			set_row(1,"@68g@S", FLD(LF_SOGGETTI,SOG_DU));
			set_row(1,"@72g@ld", FLD(LF_SOGGETTI,SOG_DATAULTDON));
			set_row(2,"@76g@S", FLD(LF_SOGGETTI,SOG_TIPOULTDON));
			set_row(1,"@86g@S", FLD(LF_SOGGETTI,SOG_TELABI));
			set_row(2,"@86g@S", FLD(LF_SOGGETTI,SOG_TELLAV));
			set_row(1,"@101g@S", FLD(LF_SOGGETTI,SOG_TELALT));
			set_row(1,"@116g@ld", FLD(LF_SOGGETTI,SOG_DATAULTID));
			set_row(2,"@119g@pn", FLD(LF_SOGGETTI,SOG_INTSI,"###"));
			set_row(2,"@123g@pn", FLD(LF_SOGGETTI,SOG_INTAF,"###"));
			set_row(3,"");			
		}
		break;					
	}	
}

bool TStampaUrgenze::filter_func_urgenze(const TRelation* rel)
{   
	bool filtrato = TRUE;
	TLocalisamfile& sog = rel->lfile(); 
	//filtro per categorie
	TAssoc_array& categorie = app()._categorie;
	if (categorie.items() != 0)
	{
		const TString16 cat = sog.get(SOG_CATDON);
		filtrato = categorie.is_key((const char*) cat);
	}	
	// filtro per idoneita al tipo donazione
	if (filtrato)
	{     
		filtrato = FALSE;
		TString16 idon = sog.get(SOG_IDON1);
		if (idon == app()._tipodon)
			filtrato = TRUE;
		else
		{
			TString16 idon = sog.get(SOG_IDON2);
			if (idon == app()._tipodon)
				filtrato = TRUE;
			else
			{              
				TString16 idon = sog.get(SOG_IDON3);
				if (idon == app()._tipodon)
					filtrato = TRUE;
				else
				{                                 
					TString16 idon = sog.get(SOG_IDON4);
					if (idon == app()._tipodon)
						filtrato = TRUE;
				}			
			}			
		}			
	}
	// filtro su data donazione
	if (filtrato)
	{                  
		const TDate datanulla(NULLDATE);	
		const TDate dataultdon = sog.get(SOG_DATAULTDON);
		if (app()._datalimite != datanulla && dataultdon != datanulla)
			filtrato = (dataultdon >= app()._datalimite);
    if (filtrato)
    {
			const long giorni_sez = rel->lfile(LF_SEZIONI).get_long(SEZ_INTMINCONV);
			long giorni;
			if (giorni_sez == 0)
				giorni = app()._giorni;
			else
				giorni = (giorni_sez > app()._giorni) ? giorni_sez : app()._giorni;
			TDate data(TODAY);
			if (giorni==0)
			{ 
				TDate datapros;
				if (app()._tipodon == IDON_SI)
					datapros = sog.get(SOG_DATAPROSSI);
				else
					datapros = sog.get(SOG_DATAPROSAF);
				filtrato = (datapros <= data);			
			}				
			else
			{
				data = data - (const long) giorni;
	  		filtrato = (dataultdon <= data);
			}  		
		}	
  }  	
  return filtrato;
}

bool TStampaUrgenze::preprocess_page(int file, int counter)
{
	if (_tipostampa == elenco)
	{
		TRectype& recsog = current_cursor()->curr();
		TString80 nome = recsog.get(SOG_COGNOME);
		nome << " ";
		nome << recsog.get(SOG_NOME);
		_cognome_nome = nome;
		TString256 localita = "";
		localita = recsog.get(SOG_DOM_CAP);			
		localita << ' ';
		if (recsog.get(SOG_DOM_CODLOC).not_empty())
		{
			localita << current_cursor()->curr(-ALIAS_LCP).get("S0");
			localita << " - ";		
		}
		localita << current_cursor()->curr(LF_COMUNI).get(COM_DENCOM);
		localita.trim();
		_dencom = localita;
		// salto pagina se cambio sezione
		const TString16 codsez = recsog.get(SOG_CODSEZ);		
		const TString16 codsot = recsog.get(SOG_CODSOT);		
		if ((!_unica) && ((_codsez!=codsez)||(_codsot!=codsot)))
		{
			if (_codsez != "**")
				footer_sezione();
			_codsez = codsez;
			_codsot = codsot;
			_contatore = 0;
			header_sezione(codsez, codsot);
		}
		else
			if (printer().rows_left() < 2)
				printer().formfeed();
	}		
	if (_tipostampa==etichette)		
		if (printer().rows_left() < _form_eti->get_body().height())
			printer().formfeed();
	_contatore++;
	_totale++;		
  return TRUE;
}

void TStampaUrgenze::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(132);
	intestazione = "Sezione: ";
	intestazione << codsez;
	intestazione << "/";
	intestazione << codsot;
	intestazione << " ";
	intestazione << densez;
	if (densot.not_empty())
	{
		intestazione << "/";
		intestazione << densot;
	}		
	intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

print_action TStampaUrgenze::postprocess_print(int file, int counter)
{ 
	if (_contatore > 0)
		footer_sezione();
	fine_stampa();	
	return NEXT_PAGE;
}

void TStampaUrgenze::footer_sezione()
{
	// stampa totale soggetti appartenenti alla sezione
  if (_tipostampa==elenco)
	{
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		set_footer(3,"TOTALE SOGGETTI STAMPATI %d", _contatore);
		printer().formfeed(); 
		reset_footer();
	}	
}
                          
void TStampaUrgenze::fine_stampa()
{
	// stampa totale soggetti a fine stampa
  if (_tipostampa==elenco)
	{
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		if (_totale > 0 && _totale != _contatore)
		{
			set_footer(3,"TOTALE SOGGETTI STAMPATI %d", _totale);
			printer().formfeed();		
		}			
		reset_footer(); 
	}	
}

bool TStampaUrgenze::set_print(int)
{
  TPrinter& p = printer();
	p.read_configuration();
	_tipostampa = undefined;
  KEY tasto;
  tasto = _msk->run();
  switch (tasto)
  {
  case F_ELENCO:
  {
  	_tipostampa = elenco;
  	_codsez = "**";
  	_codsot = "**";
  	_contatore = 0;
  	_totale = 0;
  }	
  break;
  case F_ETICHETTE:
  	_tipostampa = (configura_stampante(p, "AT_ETICHETTE", "etichette")) ? etichette : undefined;
	break;  	
  }	
	if (_tipostampa != undefined)
  {
    reset_files(); 
    add_file(LF_SOGGETTI);
		// filtro per categorie		
    _categorie.destroy();		
  	const TString16 catpri = _msk->get(F_CAT1);
  	const TString16 catsec = _msk->get(F_CAT2);
  	const TString16 catter = _msk->get(F_CAT3);
  	const TString16 catqua = _msk->get(F_CAT4);
  	const TString16 catqui = _msk->get(F_CAT5);
  	const TString16 catses = _msk->get(F_CAT6);
  	if (catpri.not_empty())
  		_categorie.add((const char*) catpri);
  	if (catsec.not_empty())
  		_categorie.add((const char*) catsec);
  	if (catter.not_empty())
  		_categorie.add((const char*) catter);
  	if (catqua.not_empty())
  		_categorie.add((const char*) catqua);
  	if (catqui.not_empty())
  		_categorie.add((const char*) catqui);
  	if (catses.not_empty())
  		_categorie.add((const char*) catses);
		// filtro per tipizzazione
		TString filtro;
		filtro << "TCS->S6 == \"I\"";
  	const TString16 gruppoab0 	= _msk->get(F_GRUPPOAB0);
  	const TString16 rhantid 		= _msk->get(F_RHANTID);
  	const TString16 kell 				= _msk->get(F_KELL);
  	const TString16 fenotiporh	=	_msk->get(F_FENOTIPORH);
  	const TString16 du					=	_msk->get(F_DU);
		if (gruppoab0.not_empty())
    	filtro << " && (" << SOG_GRUPPOAB0 << " == \"" << gruppoab0 << "\")";
		if (rhantid.not_empty())
    	filtro << " && (" << SOG_RHANTID << " == \"" << rhantid << "\")";
		if (kell.not_empty())
    	filtro << " && (" << SOG_KELL << " == \"" << kell << "\")";
		if (fenotiporh.not_empty())
    	filtro << " && (" << SOG_FENOTIPORH << " == \"" << fenotiporh << "\")";
		if (du.not_empty())
    	filtro << " && (" << SOG_DU << " == \"" << du << "\")";
		_giorni  = _msk->get_long(F_GIORNI);
		_tipodon = _msk->get(F_TIPODON);
		_unica   = _msk->get_bool(F_UNICA);
		_datalimite   = _msk->get_date(F_DATALIMITE);
		// filtro per sezioni		  
		filtra_sezioni();
		if (_unica)
			((TSorted_cursor*)current_cursor())->change_order("90->COGNOME|90->NOME|90->DATANASC");
		else			
			((TSorted_cursor*)current_cursor())->change_order("90->CODSEZ|90->CODSOT|90->COGNOME|90->NOME|90->DATANASC");
		// filtra solo idonei
    current_cursor()->setfilter(filtro, TRUE);
    // filtra per categorie, tipizzazione e data donazione
    current_cursor()->set_filterfunction (filter_func_urgenze, TRUE);
		reset_print();
		printer().footerlen(0);
    crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TStampaUrgenze::crea_intestazione()
{
  reset_header();
  if (_tipostampa == elenco)
  {
  	TString sep(132);
  	sep = "CONVOCAZIONI URGENTI per ";
  	sep << _tipodon;
		sep.center_just();
  	set_header(2, "@0g%s", (const char*) sep);
  	TString16 data_stampa = _data_stampa.string();
  	set_header(2,"@0g%10s", (const char*) data_stampa);
  	sep = "";
  	sep << "Pag. @#";
  	set_header(2, "@120g%s", (const char*) sep);  
  	sep = "";
  	sep.fill('-');
  	set_header(3, (const char *) sep);
  	set_header(4,"@0gCodice@9gC.@12gCognome e nome@38gNato il@49gGr.@53gRh@57gFen.Rh@64gKellDu@72gUlt.donaz.@86gTelefono abit.@101gTelefono altro@116gUlt.idon.");
  	set_header(5,"@0gTessera@38gSe/So@75gTipo@86gTelefono lav.@116gInt.SI  AF");
  	set_header(6,"@0g--------@9g--@12g-------------------------@38g----------@49g--- --- ------ --- ---@72g----------@86g--------------@101g--------------@116g----------");
  	printer().footerlen(3);
	}  	
}

bool TStampaUrgenze::user_create()
{
  _rel = new TRelation(LF_SOGGETTI);
  _rel->add("TCS", "CODTAB==STATO",1,0,ALIAS_TCS);
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,0,ALIAS_LCP);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM");
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT");
  add_cursor(new TSorted_cursor(_rel,"","",3));
  _msk = new TMask("at4200a");
  TConfig config(CONFIG_STUDIO);
  TString16 etformato = config.get("EtFormato");
  _etlarghezza = config.get_int("EtLarghezza");
  _etcolonne = config.get_int("EtColonne");
	_form_eti = new TUrgenze_form(etformato);
  return TRUE;
}

bool TStampaUrgenze::user_destroy()
{
  delete _form_eti;
  delete _msk;
  delete _rel;
  return TRUE;
}

int at4200(int argc, char* argv[])
{
  TStampaUrgenze a;
  a.run(argc, argv, "Urgenze");
  return 0;
}
