#include <form.h>
#include <mask.h>
#include <printapp.h>
#include <tabutil.h>
#include <utility.h>

#include "soggetti.h"
#include "sezioni.h"
#include "convoc.h"
#include "rconvoc.h"
#include <comuni.h>

#include "at4.h"
#include "at4400a.h"
#include "at4100b.h"
#include "at4100c.h"
#include "atlib.h"

#define ALIAS_LDN	400
#define ALIAS_LCPDOM	100		// localita' postale di domicilio
#define ALIAS_COMDOM 	501		// comune di domicilio
#define ALIAS_COMLDN  503   // comune per punti di prelievo
#define ALIAS_LCPLDN  102   // localita' postale per punti di prelievo

enum ts { undefined=0, sintetico=1, completo=2, etichette=3, cartoline=4 };

// definizione form per etichette
class TStConvoc_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TStConvoc_form(): TForm() {};
  TStConvoc_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TStConvoc_form() {};
};

class TConvocazioniPerPunto : public TPrintapp
{
  TRelation*				_rel;
  TMask*      			_msk;
  TStConvoc_form*		_form_eti;
  TStConvoc_form*   _form_car;
  TAssoc_array			_categorie;
  int 							_cur;
  TParagraph_string _cognome_nome, _dencom;
  TDate 						_data_stampa;
  TDate							_dataini, _datafin, _dataatt;
  bool 							_aggiorna, _persezione, _tutti, _perdata;
  char 							_gruppiaz;
	TString16 				_punto,_tipo, _ordina;
	TString16					_codsez, _codsot, _sezini,_sotini, _sezfin,_sotfin;
	TString16					_orario, _datacart;
	TString80					_invitoper, _data, _presso1, _presso2, _presso3, _presso4;
	TString80					_note, _intest1, _intest2, _intest3, _intest4;
	bool 							_usasez;
	ts								_tipostampa;
	int 							_contxsez, _contxdata, _contatore;
	int								_etlarghezza, _etcolonne;

  static bool filter_func_perpunto(const TRelation* rel);
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page(int file, int counter);
  virtual print_action postprocess_print(int file, int counter);
  ts dati_cartoline();

public:
  void crea_intestazione();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  void header_data(const TDate data);
  void footer_sezione();
  void footer_data();
  void fine_stampa();
  void dati_sezione(const TString16 codsez, const TString16 codsot);
  TMask& app_mask() { return *_msk; }
  TConvocazioniPerPunto() : _data_stampa(TODAY), _cognome_nome("",35), _dencom("",30) {}
};

HIDDEN inline TConvocazioniPerPunto& app() { return (TConvocazioniPerPunto&) main_app(); }

TCursor* TStConvoc_form::cursor() const { return app().current_cursor(); }

TRelation* TStConvoc_form::relation() const { return cursor()->relation(); }

bool TConvocazioniPerPunto::filter_func_perpunto(const TRelation* rel)
{   
	bool filtrato = TRUE;
	TLocalisamfile& sog = rel->lfile(LF_SOGGETTI); 
	// filtro per categorie
	TAssoc_array& categorie = app()._categorie;
	if (categorie.items() != 0)
	{
		const TString16 cat = sog.get(SOG_CATDON);
		filtrato = categorie.is_key((const char*) cat);
	}	
	return filtrato;
}	

ts TConvocazioniPerPunto::dati_cartoline()
{
	TMask msk("at4100b");
 	if (msk.run() == K_ENTER) 
 	{ 
 		_invitoper = msk.get(F_INVITOPER);
 		_data			 = msk.get(F_DATA);
 		_presso1   = msk.get(F_PRESSO1);
 		_presso2   = msk.get(F_PRESSO2);
 		_presso3   = msk.get(F_PRESSO3);
 		_presso4   = msk.get(F_PRESSO4);
 		_datacart  = msk.get(F_DATACART);
 		_orario  	 = msk.get(F_ORARIO);
 		_note   	 = msk.get(F_NOTE);
 		_usasez		 = msk.get_bool(F_USASEZ);
 		return cartoline;
 	}
 	else
 		return undefined;
}

void TConvocazioniPerPunto::fine_stampa()
{
	// stampa totale soggetti a fine stampa
	
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		set_footer(3,"TOTALE SOGGETTI CONVOCATI %d", _contatore);
		printer().formfeed(); 
		reset_footer();				
}

void TConvocazioniPerPunto::set_page(int file, int cnt)
{
	// costruzione etichette
	switch (_tipostampa)
	{
		case etichette:
		{
			TPrint_section& corpo = _form_eti->get_body();
			for (int r=1;r<=_etcolonne;r++)
			{                     
				if (current_cursor()->pos()<current_cursor()->items())
				{ 
			  	preprocess_page(file, cnt);
					force_setpage(TRUE);
					corpo.update();
  				for (word i = 0; i < corpo.height(); i++)
  				{
  					TPrintrow& riga = corpo.row(i);
  					TString256 riga1 = riga.row();
						riga1.cut(_etlarghezza);
  					int colonna = ((r-1)*_etlarghezza);
  					TString16 formato;
  					formato << '@' << colonna << "g";
  					riga1.insert(formato,0);
  					set_row(i+1,riga1);
					}
					if (r < _etcolonne)  		
						++(*current_cursor());
				}					
			}				
		}
		break;
		case cartoline:
		{
			TPrint_section& corpo = _form_car->get_body();
			corpo.update();
  		for (word i = 0; i < corpo.height(); i++)
  		{
  			TPrintrow& riga = corpo.row(i);
  			set_row(i+1,riga);
			}  			
			force_setpage(TRUE);	// serve perch alla prossima etichetta rifaccia la setpage
														// altrimenti stampa sempre la stessa etichetta
		}
		break;
		case sintetico:
		{
			set_row(1,"@0g#a", &_cognome_nome);
			set_row(1,"@36g@S", FLD(LF_SOGGETTI,SOG_CATDON));
			set_row(1,"@40g@S", FLD(LF_SOGGETTI,SOG_TESSAVIS));
			set_row(1,"@48g@S", FLD(LF_SOGGETTI,SOG_CODSEZ));
			set_row(1,"@51g@S", FLD(LF_SOGGETTI,SOG_CODSOT));
			set_row(1,"@54g@ld", FLD(LF_SOGGETTI,SOG_DATAULTDON));
			set_row(1,"@65g@S", FLD(LF_SOGGETTI,SOG_TIPOULTDON));
			set_row(1,"@68g@ld", FLD(LF_RCONVOC,RCV_DATACONV));
			set_row(1,"@79g@S", FLD(LF_RCONVOC,RCV_CHIAMATA));
			set_row(1,"@84g@S", FLD(LF_CONVOC,COV_DATA));
			set_row(1,"@95g@S", FLD(LF_SOGGETTI,SOG_DATAPROSSI));
		}
		break;
		case completo:
		{
			set_row(1,"@0g#a", &_cognome_nome);
			set_row(3,"@0g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
			set_row(1,"@36g@S", FLD(LF_SOGGETTI,SOG_CATDON));
			set_row(1,"@40g@S", FLD(LF_SOGGETTI,SOG_TESSAVIS));
			set_row(2,"@40g@S", FLD(LF_SOGGETTI,SOG_CODSEZ));
			set_row(2,"@43g@S", FLD(LF_SOGGETTI,SOG_CODSOT));
			set_row(1,"@48g@S", FLD(LF_SOGGETTI,SOG_DOM_INDIR));
			set_row(2,"@48g#a", &_dencom);
			set_row(1,"@79g@ld", FLD(LF_SOGGETTI,SOG_DATAULTDON));
			set_row(2,"@83g@S", FLD(LF_SOGGETTI,SOG_TIPOULTDON));
			set_row(3,"@79g@S", FLD(LF_SOGGETTI,SOG_DATAPROSSI));
			set_row(1,"@90g@ld", FLD(LF_RCONVOC,RCV_DATACONV));
			set_row(2,"@94g@S", FLD(LF_RCONVOC,RCV_CHIAMATA));
			set_row(1,"@101g@S", FLD(LF_SOGGETTI,SOG_TELABI));
			set_row(2,"@101g@S", FLD(LF_SOGGETTI,SOG_TELLAV));
			set_row(3,"@101g@S", FLD(LF_SOGGETTI,SOG_TELALT));
			set_row(1,"@117g@S", FLD(LF_SOGGETTI,SOG_GRUPPOAB0));
			set_row(2,"@117g@S", FLD(LF_SOGGETTI,SOG_RHANTID));
			set_row(4,"");
		}
		break;
	}					
}

print_action TConvocazioniPerPunto::postprocess_print(int file, int counter)
{ 
	if (_contxsez > 0)
		footer_sezione();
	if (_contxdata > 0)
		footer_data();
	if ((_contatore > 0) && ((_contatore != _contxsez) || (!_persezione)) && ((_contatore != _contxdata) || (!_perdata)))
		fine_stampa();		
	return NEXT_PAGE;
}

bool TConvocazioniPerPunto::preprocess_page(int file, int counter)
{
	if (_aggiorna)
	{
		current_cursor()->curr().put(RCV_STAMPATO,"X");
		current_cursor()->file().rewrite();
	}
	if ((_tipostampa==sintetico) || (_tipostampa==completo))
	{
		TString80 nome = current_cursor()->curr(LF_SOGGETTI).get(SOG_COGNOME);
		nome << " ";
		nome << current_cursor()->curr(LF_SOGGETTI).get(SOG_NOME);
		_cognome_nome = nome;
		if (_tipostampa==completo)
		{
			TString256 localita = "";
			localita = current_cursor()->curr(LF_SOGGETTI).get(SOG_DOM_CAP);			
			localita << ' ';
			if (current_cursor()->curr(LF_SOGGETTI).get(SOG_DOM_CODLOC).not_empty())
			{
				localita << current_cursor()->curr(-ALIAS_LCPDOM).get("S0");
				localita << " - ";		
			}
			localita << current_cursor()->curr(LF_COMUNI).get(COM_DENCOM);
			localita.trim();
			_dencom = localita;
		}		
		// salto pagina se cambio sezione
		const TString16 codsez = current_cursor()->curr(LF_SOGGETTI).get(SOG_CODSEZ);		
		const TString16 codsot = current_cursor()->curr(LF_SOGGETTI).get(SOG_CODSOT);		
		if ((_codsez!=codsez)||(_codsot!=codsot))
		{
			if (_codsez != "**")
				footer_sezione();
			_contxsez = 0;				
			_codsez = codsez;
			_codsot = codsot;
			header_sezione(codsez, codsot);
		}
		const TDate data = current_cursor()->curr().get_date(RCV_DATACONV);
		if (_dataatt!=data)
		{
			if (_dataatt.ok())
				footer_data();
			_contxdata = 0;				
			_dataatt = data;
			header_data(data);
		}
		_contxdata++;
		_contxsez++;
		_contatore++;
	}
	if (_tipostampa==etichette)		
		if (printer().rows_left() < _form_eti->get_body().height())
			printer().formfeed();
  return TRUE;
}

void TConvocazioniPerPunto::footer_sezione()
{ 
	if (_persezione)
	{
		// stampa totale sezione
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		set_footer(3,"TOTALE CONVOCATI SEZIONE    %d", _contxsez);
		printer().formfeed(); 
		reset_footer();				
	}		
}

void TConvocazioniPerPunto::footer_data()
{ 
	if (_perdata)
	{
		// stampa totale sezione
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		set_footer(3,"TOTALE CONVOCATI DEL GIORNO %d", _contxdata);
		printer().formfeed(); 
		reset_footer();				
	}		
}

void TConvocazioniPerPunto::header_sezione(const TString16 codsez, const TString16 codsot)
{
	if (_persezione)
	{
		const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
		const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
		TString intestazione(80);
		intestazione = "Sezione: ";
		intestazione << codsez;
		if (codsot.not_empty())
		{
			intestazione << "/";
			intestazione << codsot;
		}		
		intestazione << " ";
		intestazione << densez;
		if (densot.not_empty())
		{
			intestazione << "/";
			intestazione << densot;
		}		
		intestazione.center_just();
		set_header(1,"@0g%s", (const char*) intestazione);
	}	 	
	return;
}

void TConvocazioniPerPunto::header_data(const TDate data)
{
	if (_perdata)
	{
		TString intestazione(80);
		intestazione = "Data cartolina: ";
		intestazione << data.string();
		intestazione.center_just();
		set_header(1,"@0g%s", (const char*) intestazione);
	}	 	
	return;
}

void TConvocazioniPerPunto::dati_sezione(const TString16 codsez, const TString16 codsot)
{
	_intest1 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ1);
	_intest2 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ2);
	_intest3 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ3);
	_intest4 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ4);
	if (_usasez)
	{
		if (_presso1.blank())
		{
			_presso1 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE1);
			_presso2 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE2);
			_presso3 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE3);
			_presso4 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE4);
		}			
		if (_orario.blank())
			_orario = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVORA);
		if (_note.blank())
			_note = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVNOTE);
	}
	return;
}

bool TConvocazioniPerPunto::set_print(int m)
{ 
  TPrinter& p = printer();
	p.read_configuration();
	_tipostampa = undefined;
  KEY tasto = F_AZZERA;
  while (tasto == F_AZZERA)
  {                        
  	tasto = _msk->run();
  	if (tasto == F_AZZERA)
  		_msk->reset();
	}  	
  switch (tasto)
  {
	  case F_SINTETICO:
	  	_tipostampa = sintetico;
	  break;
	  case F_COMPLETO:
	  	_tipostampa = completo;
	  break;
	  case F_ETICHETTE:
  		_tipostampa = (configura_stampante(p, "AT_ETICHETTE", "etichette")) ? etichette : undefined;
		break;  	
	  case F_CARTOLINE:
  		_tipostampa = (configura_stampante(p, "AT_CARTOLINE", "cartoline")) ? cartoline : undefined;
	  break;
  }	
	if (_tipostampa != undefined)
  {
    _sezini     = _msk->get(F_SEZINI);
    _sotini			= _msk->get(F_SOTINI);
    _sezfin     = _msk->get(F_SEZFIN);
    _sotfin			= _msk->get(F_SOTFIN);
    _dataini 		= _msk->get(F_DATAINI);
    _datafin 		= _msk->get(F_DATAFIN);
    _punto      = _msk->get(F_PUNTO);
    _tipo				= _msk->get(F_TIPO);
    _gruppiaz   = _msk->get(F_SCELTAGRUPPI)[0];
    _aggiorna   = _msk->get_bool(F_AGGIORNA);
    _persezione = _msk->get_bool(F_PERSEZIONE);
    _tutti      = (_msk->get(F_TUTTI)[0] == 'T');
    _ordina     = _msk->get(F_ORDINA);
    _perdata = FALSE;
		TString256 chiave = "";
		if (_tipostampa==cartoline)
			chiave = "90->GRUPPOAZIE|";			
		switch (_ordina[0])
		{
			case 'A':
					chiave << "90->CODSEZ|90->CODSOT|UPPER(90->COGNOME)|UPPER(90->NOME)|116->DATACONV";
			break;
			case 'S':
				chiave <<"90->CODSEZ|90->CODSOT|90->DOM_CAP|UPPER(90->COGNOME)|UPPER(90->NOME)|116->DATACONV";
			break;
			case 'C':
				chiave << "90->DOM_CAP|UPPER(90->COGNOME)|UPPER(90->NOME)|116->DATACONV";
			break;
			case 'D':
			{
				chiave << "116->DATACONV|UPPER(90->COGNOME)|UPPER(90->NOME)";
				_perdata = TRUE;
			}	
			break;
		}
		TString256 filtro = "";			
		filtro << "(116->ANNULLATO != \"X\")";
		if (!_tutti)
			filtro << " && (116->STAMPATO != \"X\")";
  	// filtro per punto di raccolta
  	if (_punto.not_empty())
  	{
			filtro << " && ";
			filtro << format("(115->PUNTO == \"%s\")",(const char*)_punto);
		}	
		// filtro per tipo
		if (_tipo.not_empty())
		{
			filtro << " && ";
			filtro << format("(115->TIPO == \"%s\")",(const char*)_tipo);
		}			
	 	if (_sezini.not_empty())
	 	{
			filtro << " && ";
	 		if (_sotini.not_empty())
	 		{                    
	 			filtro << "(";
	 			filtro << format("(90->CODSEZ > \"%s\")",(const char*)_sezini);
	 			filtro << " || ";
	 			filtro << "(" << format("(90->CODSEZ == \"%s\")",(const char*)_sezini);
	 			filtro << " && ";
	 			filtro << format("(90->CODSOT >= \"%s\")",(const char*)_sotini);
	 			filtro << ")";
	 			filtro << ")";
	 		}
	 		else
	 			filtro << format("(90->CODSEZ >= \"%s\")",(const char*)_sezini);
	 	}
	 	if (_sezfin.not_empty())
	 	{
			filtro << " && "; 
	 		if (_sotfin.not_empty())
	 		{          
	 			filtro << "(";
	 			filtro << format("(90->CODSEZ < \"%s\")",(const char*)_sezfin);
	 			filtro << " || ";
	 			filtro << "(" << format("(90->CODSEZ == \"%s\")",(const char*)_sezfin);
	 			filtro << " && ";
	 			filtro << format("(90->CODSOT <= \"%s\")",(const char*)_sotfin);
	 			filtro << ")";
	 			filtro << ")";
	 		}
	 		else
	 			filtro << format("(90->CODSEZ <= \"%s\")",(const char*)_sezfin);
	 	}
	 	if (_gruppiaz == 'S')  // solo appartenenti a gruppi aziendali
    {
 			filtro << " && ";
  		filtro << format("(90->GRUPPOAZIE != \"\")");
    }
    else
		 	if (_gruppiaz == 'E')  // escludi gli appartenenti a gruppi aziendali
	    {
	 			filtro << " && ";
	  		filtro << format("(90->GRUPPOAZIE == \"\")");
	    }
		// filtro per data
	  TRectype da(LF_RCONVOC);
  	TRectype a (LF_RCONVOC);
  	if (_dataini.ok())
			da.put(RCV_DATACONV, _dataini);
  	if (_datafin.ok())
			a.put(RCV_DATACONV, _datafin);
	  _cur = add_cursor(new TSorted_cursor(_rel, (const char*) chiave,"", 2, &da, &a));
		current_cursor()->setfilter((const char*) filtro, TRUE);
		// filtro per categorie
		// filtro per categorie	
		_categorie.destroy();	
  	const TString16 catpri = _msk->get(F_CAT1);
  	const TString16 catsec = _msk->get(F_CAT2);
  	const TString16 catter = _msk->get(F_CAT3);
  	const TString16 catqua = _msk->get(F_CAT4);
  	const TString16 catqui = _msk->get(F_CAT5);
  	const TString16 catses = _msk->get(F_CAT6);
  	if (catpri.not_empty())
  		_categorie.add((const char*) catpri);
  	if (catsec.not_empty())
  		_categorie.add((const char*) catsec);
  	if (catter.not_empty())
  		_categorie.add((const char*) catter);
  	if (catqua.not_empty())
  		_categorie.add((const char*) catqua);
  	if (catqui.not_empty())
  		_categorie.add((const char*) catqui);
  	if (catses.not_empty())
  		_categorie.add((const char*) catses);
		current_cursor()->set_filterfunction(filter_func_perpunto);

  	_codsez = "**";
  	_codsot = "**";
  	_dataatt = NULLDATE;
    _contxsez = 0; 
    _contxdata = 0;
    _contatore = 0;
    reset_files(); 
    add_file(LF_RCONVOC);
		reset_print();
		printer().footerlen(0);		
		crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TConvocazioniPerPunto::crea_intestazione()
{
  reset_header();
  if ((_tipostampa==sintetico) || (_tipostampa==completo))
  {
  	TString sep(80);
  	sep = "ELENCO CONVOCATI";
  	if (_tipo.not_empty())
  	{
  		sep << " per ";
  		sep << _tipo;
		}  		
		if (_dataini.ok())
		{
  		sep << " dal ";
 			sep << _dataini.string();
		}
		if (_datafin.ok())
		{ 			
 			sep << " al ";
 			sep << _datafin.string();
		}
		sep.center_just();
  	set_header(2, "@0g%s", (const char*) sep);
	 	TString16 data_stampa = _data_stampa.string();
	 	set_header(2,"@0g%10s", (const char*) data_stampa);
		sep = "";
	 	sep << "Pag. @#";
	 	set_header(2, "@80g%s", (const char*) sep);  	
		sep = "";
		if (_punto.not_empty())
		{         
 			sep << "presso ";
 			sep << _punto;
 			sep << " ";
			sep << cache().get("LDN", _punto).get("S0");
		} 			
		sep.center_just();
  	set_header(3, "@0g%s", (const char*) sep);
  	sep = "";
	 	char ora[16];
 		_strtime(ora);
 		set_header(3,"@0g%8s", ora);  	
 		if (_gruppiaz == 'S')
 			sep << "SOLO APPARTENENTI A GRUPPI AZIENDALI";
 		else
	 		if (_gruppiaz == 'E')
 				sep << "ESCLUSI GLI APPARTENENTI A GRUPPI AZIENDALI";
		sep.center_just();
  	set_header(4, "@0g%s", (const char*) sep);
		sep = "";
 		if (_tipostampa==sintetico)
 		{
		  set_header(5,"@0gCognome e nome@36gCat.@40gTessera@48gSe/So@54gUltima donaz.@68gData cart.@79gTipo@84gData conv.@95gData pr.");
	  	set_header(6,"@0g-----------------------------------@36g---@40g-------@48g-- --@54g---------- --@68g----------@79g----@84g----------@95g----------");
		}
		else	  	
 		{
		  set_header(5,"@0gCognome e nome@36gCat.@40gTessera@48gIndirizzo@79gUltima donaz.@90gData cart.@101gTel.abitazione@117gGr.AB0");
		  set_header(6,"@40gSe/So@79gTipo u.d.@90gTipo cart.@101gTel.lavoro@117gRh/Antid");
		  set_header(7,"@0gData nascita@101gTel.altro");
	  	set_header(8,"@0g-----------------------------------@36g---@40g-------@48g------------------------------@79g----------@90g----------@101g--------------@117g--------");
		}
  	printer().footerlen(3);
	}  	
}

bool TConvocazioniPerPunto::user_create()
{
  _rel = new TRelation(LF_RCONVOC);
  _rel->add(LF_SOGGETTI, "CODICE==CODICE");
  _rel->add(LF_CONVOC, "NUMERO==NUMERO");
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,LF_SOGGETTI,ALIAS_LCPDOM);
  _rel->add("LDN", "CODTAB==PUNTO",1,LF_CONVOC,ALIAS_LDN);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM",1,LF_SOGGETTI, ALIAS_COMDOM);
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT",1,LF_SOGGETTI);
  _rel->add("LCP", "CODTAB==S7",1,-ALIAS_LDN,ALIAS_LCPLDN);
  _rel->add(LF_COMUNI, "COM==S6",1,-ALIAS_LDN, ALIAS_COMLDN);
  _msk = new TMask("at4400a");
  TConfig config(CONFIG_STUDIO);
  TString16 etformato = config.get("EtFormato");
  TString16 caformato = config.get("CaFormato");
  _etlarghezza = config.get_int("EtLarghezza");
  _etcolonne = config.get_int("EtColonne");
  _form_eti = new TStConvoc_form(etformato);
  //_form_car = new TStConvoc_form("ATCARTO1");
  _form_car = new TStConvoc_form(caformato);
  return TRUE;
}

bool TConvocazioniPerPunto::user_destroy()
{
  delete _form_car;
  delete _form_eti;
  delete _msk;
  delete _rel;
  return TRUE;
}

int at4400(int argc, char* argv[])
{
  TConvocazioniPerPunto a;
  a.run(argc, argv, "Stampa convocati su punto di prelievo");
  return 0;
}
