#include <form.h>
#include <mask.h>
#include <printapp.h>
#include <recarray.h>
#include <tabutil.h>
#include <utility.h>

#include "at5.h"

#define ALIAS_TCS	200
#define ALIAS_LCPRES	101		// localita' postale di residenza
#define ALIAS_LCPDOM	100		// localita' postale di domicilio
#define ALIAS_COMRES 	502		// comune di residenza
#define ALIAS_COMDOM 	501		// comune di domicilio
#define ALIAS_COMNAS 	500		// comune di nascita

// nomi campi maschera
#include "at5200a.h"
          
// nomi dei campi
#include "soggetti.h"
#include "contsan.h"
#include "sezioni.h"
#include <comuni.h>

enum ts { undefined = 0, elenco = 1, etichette = 2 };

// definizione form per etichette
class TEti_sitsanitaria_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TEti_sitsanitaria_form(): TForm() {};
  TEti_sitsanitaria_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TEti_sitsanitaria_form() {};
};

class TSitSanitaria : public TPrintapp
{
	TMask*						_msk;
	TRelation*   			_rel;
	TLocalisamfile*		_contsan;
	TRecord_array* 		_scontrolli;
  TEti_sitsanitaria_form*	_form_eti;
  TAssoc_array		  _categorie;
	int 							_cur;	
	int								_contatore, _totale;
	TDate							_data_stampa;
	TString16					_codsez, _codsot;
	TString16					_tipo1, _tipo2, _tipo3, _tipo4, _tipo5;
  TString16					_lettini, _lettfin;	
  TParagraph_string _cognome_nome, _rigastampa, _dencom;
	ts 	      				_tipostampa;
	int								_etlarghezza, _etcolonne;

	static bool filter_func_sitsanitaria(const TRelation* rel);

protected:
	virtual bool user_create();
	virtual bool user_destroy();
	virtual bool set_print(int m);
	virtual void set_page(int file, int cnt);
	virtual bool preprocess_page(int file, int counter);
  virtual print_action postprocess_print(int file, int counter);
	
public:
  void filtra_sezioni();
  void crea_intestazione();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  void footer_sezione();
  void fine_stampa();
	TSitSanitaria() : _data_stampa(TODAY), _cognome_nome("",51), _rigastampa("", 71), _dencom("",51)  {}
};

HIDDEN inline TSitSanitaria& app() { return (TSitSanitaria&) main_app(); }

TCursor* TEti_sitsanitaria_form::cursor() const { return app().current_cursor(); }

TRelation* TEti_sitsanitaria_form::relation() const { return cursor()->relation(); }

print_action TSitSanitaria::postprocess_print(int file, int counter)
{ 
	if (_contatore > 0)
		footer_sezione();
	fine_stampa();				
	return NEXT_PAGE;
}

void TSitSanitaria::fine_stampa()
{
	// stampa totale soggetti a fine stampa
	reset_footer();
	TString sep(80);
	sep.fill('-');
	set_footer(2, (const char *) sep);
	if (_totale > 0)
	{
		if (_totale != _contatore)
		{
			set_footer(3,"TOTALE SOGGETTI STAMPATI %d", _totale);
			printer().formfeed();		
		}			
	}			
	else		
	{
		set_footer(3,"NON CI SONO SOGGETTI DA STAMPARE");		
		printer().formfeed();		
	}		
	reset_footer(); 
}

void TSitSanitaria::footer_sezione()
{
	if (_contatore > 0)
	{
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		set_footer(3,"SOGGETTI STAMPATI PER SEZIONE %d", _contatore);
		printer().formfeed(); 
		reset_footer();
	}				
}

void TSitSanitaria::filtra_sezioni()
{
 	const TString16 sezini = _msk->get(F_SEZINI);
 	const TString16 sotini = _msk->get(F_SOTINI);
 	const TString16 sezfin = _msk->get(F_SEZFIN);
 	const TString16 sotfin = _msk->get(F_SOTFIN);
  TRectype da(LF_SOGGETTI);
  TRectype a(LF_SOGGETTI);   
  if (sezini.not_empty())
		da.put(SOG_CODSEZ, sezini);
  if (sotini.not_empty())
		da.put(SOG_CODSOT, sotini);
  if (sezfin.not_empty())
		a.put(SOG_CODSEZ, sezfin);
  if (sotfin.not_empty())
		a.put(SOG_CODSOT, sotfin);
	current_cursor()->setregion(da, a);
}	

bool TSitSanitaria::preprocess_page(int file, int counter)
{ 
	if (_tipostampa == elenco)
	{
		TRectype& recsog = current_cursor()->curr();
		TString80 nome = recsog.get(SOG_COGNOME);
		nome << " ";
		nome << recsog.get(SOG_NOME);
		_cognome_nome = nome;
		TString256 localita = "";
		localita = recsog.get(SOG_DOM_CAP);			
		localita << ' ';
		if (recsog.get(SOG_DOM_CODLOC).not_empty())
		{
			localita << current_cursor()->curr(-ALIAS_LCPDOM).get("S0");
			localita << " - ";		
		}
		localita << current_cursor()->curr(LF_COMUNI).get(COM_DENCOM);
		localita.trim();
		_dencom = localita;
		TDate data1(NULLDATE), data2(NULLDATE), data3(NULLDATE), data4(NULLDATE), data5(NULLDATE);
		TRectype* key = new TRectype(LF_CONTSAN);
		long codice = recsog.get_long(SOG_CODICE);
		key->put(CON_CODICE, codice);
		int err = _scontrolli->read(key);
		for (int r=1; r<=_scontrolli->rows(); r++)
		{
			const TRectype& riga = _scontrolli->row(r);
			TString16 tipocon = riga.get(CON_TIPOCON);
			TDate datacon = riga.get(CON_DATACON);
			if (tipocon == _tipo1 && _tipo1.not_empty())
				data1 = datacon;
			if (tipocon == _tipo2 && _tipo2.not_empty())
				data2 = datacon;
			if (tipocon == _tipo3 && _tipo3.not_empty())
				data3 = datacon;
			if (tipocon == _tipo4 && _tipo4.not_empty())
				data4 = datacon;
			if (tipocon == _tipo5 && _tipo5.not_empty())
				data5 = datacon;
		}					   
		TString80 rigastampa;
		rigastampa = "";
		rigastampa << "@72g"<<data1;
		rigastampa << "@83g"<<data2;
		rigastampa << "@94g"<<data3;
		rigastampa << "@105g"<<data4;
		rigastampa << "@116g"<<data5;
		_rigastampa = rigastampa;
		// salto pagina se cambio sezione
		const TString16 codsez = recsog.get(SOG_CODSEZ);		
		const TString16 codsot = recsog.get(SOG_CODSOT);		
		if ((_codsez!=codsez)||(_codsot!=codsot))
		{
			if ((_codsez != "**") && (_contatore > 0))
				footer_sezione();
			_contatore = 0;			
			_codsez = codsez;
			_codsot = codsot;
			header_sezione(codsez, codsot);
		}
		if (printer().rows_left() < 4)
			printer().formfeed();
	}		
	if (_tipostampa==etichette)		
		if (printer().rows_left() < _form_eti->get_body().height())
			printer().formfeed();
	_contatore++;			
	_totale++;
  return TRUE;
}

void TSitSanitaria::set_page(int file, int cnt)
{ 
	switch (_tipostampa)
	{
		case etichette:
		{
			TPrint_section& corpo = _form_eti->get_body();
			corpo.update();
  		for (word i = 0; i < corpo.height(); i++)
  		{
  			TPrintrow& riga = corpo.row(i);
  			set_row(i+1,riga);
			}  			
			force_setpage(TRUE);	// serve perch alla prossima etichetta rifaccia la setpage
														// altrimenti stampa sempre la stessa etichetta
		}
		break;
		case elenco:
		{
			set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
			set_row(2,"@0g@S", FLD(LF_SOGGETTI,SOG_TESSAVIS));
			set_row(3,"@0g@S", FLD(LF_SOGGETTI,SOG_CATDON));
			
			set_row(1,"@9g#a", &_cognome_nome);
			set_row(2,"@9g@S", FLD(LF_SOGGETTI,SOG_DOM_INDIR));
			set_row(3,"@9g#a", &_dencom);

			set_row(1,"@61g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
			set_row(2,"@61g@S", FLD(LF_SOGGETTI,SOG_GRUPPOAB0));
			set_row(2,"@66g@S", FLD(LF_SOGGETTI,SOG_RHANTID));

			set_row(1,"@72g@S", FLD(LF_SOGGETTI,SOG_TELABI));
			set_row(1,"@87g@S", FLD(LF_SOGGETTI,SOG_TELLAV));
			set_row(1,"@102g@S", FLD(LF_SOGGETTI,SOG_TELALT));
			
			set_row(3,"@72g#a", &_rigastampa);
			
			set_row(4,"");
		}			
		break;
	}		
}

bool TSitSanitaria::filter_func_sitsanitaria(const TRelation* rel)
{   
	bool filtrato = TRUE;
	TRectype& recsog = rel->lfile().curr();
	// filtro per categorie
	TAssoc_array& categorie = app()._categorie;
	if (categorie.items() != 0)
	{
		const TString16 cat = recsog.get(SOG_CATDON);
		filtrato = categorie.is_key((const char*) cat);
	}	
	// filtro per iniziale cognome		
	if (filtrato && (!app()._lettini.blank()) && (!app()._lettfin.blank()))
	{
		TString80 cognome = recsog.get(SOG_COGNOME);
		int leni = app()._lettini.len();
		TString16 cognomei = cognome.left(leni);
		int lenf = app()._lettfin.len();
		TString16 cognomef = cognome.left(lenf);
		if (!(cognomei >= app()._lettini && cognomef <= app()._lettfin))
			filtrato = FALSE;
	}	
	return filtrato;
}

void TSitSanitaria::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(132);
	intestazione = "Sezione: ";
	intestazione << codsez;
	intestazione << "/";
	intestazione << codsot;
	intestazione << " ";
	intestazione << densez;
	if (densot.not_empty())
	{
		intestazione << "/";
		intestazione << densot;
	}		
	intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

bool TSitSanitaria::set_print(int)
{
	_tipostampa = undefined;
  KEY tasto;
  tasto = _msk->run();
  switch (tasto)
  {
  case F_ELENCO:
  	_tipostampa = elenco;
  	_codsez = "**";
  	_codsot = "**";
  	break;
  case F_ETICHETTE:
  	_tipostampa = etichette;	
  	break;
  }	
	if (_tipostampa != undefined)
  {
		_tipo1 = _msk->get(F_TIPO1);
		_tipo2 = _msk->get(F_TIPO2);
		_tipo3 = _msk->get(F_TIPO3);
		_tipo4 = _msk->get(F_TIPO4);
		_tipo5 = _msk->get(F_TIPO5);
		_lettini = _msk->get(F_LETTINI);
		_lettini.left(15);
		_lettfin = _msk->get(F_LETTFIN);
		_lettfin.left(15);		
		_contatore = 0;
		_totale = 0;
    reset_files(); 
    add_file(LF_SOGGETTI);
		filtra_sezioni();
	  // filtro per categorie
	  _categorie.destroy();
  	const TString16 catpri = _msk->get(F_CAT1);
  	const TString16 catsec = _msk->get(F_CAT2);
  	const TString16 catter = _msk->get(F_CAT3);
  	const TString16 catqua = _msk->get(F_CAT4);
  	const TString16 catqui = _msk->get(F_CAT5);
  	const TString16 catses = _msk->get(F_CAT6);
	  if (catpri.not_empty())
  		_categorie.add((const char*) catpri);
  	if (catsec.not_empty())
  		_categorie.add((const char*) catsec);
  	if (catter.not_empty())
  		_categorie.add((const char*) catter);
  	if (catqua.not_empty())
  		_categorie.add((const char*) catqua);
  	if (catqui.not_empty())
  		_categorie.add((const char*) catqui);
  	if (catses.not_empty())
  		_categorie.add((const char*) catses);
		current_cursor()->set_filterfunction(filter_func_sitsanitaria,TRUE);
		reset_print();		
		printer().footerlen(0);		
    crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TSitSanitaria::crea_intestazione()
{
	reset_header();
	TString sep(132);
 	sep = "STAMPA SITUAZIONE SANITARIA";	
	sep.center_just();
 	set_header(2, "@0g%s", (const char*) sep);
 	TDate data_stampa = _data_stampa.string();
  set_header(2,"@0g%10s", (const char*) data_stampa);
  sep = "";
  sep << "Pag. @#";
  set_header(2, "@125g%s", (const char*) sep);  
  	
  sep = "";
  sep.fill('-');
  set_header(3, (const char *) sep);
 	set_header(4,"@0gCodice@9gCognome e nome@61gNato il@72gTelefono abit.@87gTelefono lav.@102gTelefono altro");
 	set_header(5,"@0gTessera@9gIndirizzo");
	TString256 riga = "@0gCat.@61gGruppo Rh@72gUltimo ";
	riga << _tipo1 << "@83gUltimo " << _tipo2 << "@94gUltimo " << _tipo3 << "@105gUltimo " << _tipo4 << "@116gUltimo " << _tipo5;
 	set_header(6,riga);
 	set_header(7,"@0g--------@9g---------------------------------------------------@61g----------@72g------------------------------------------------------");
 	printer().footerlen(3);
}

bool TSitSanitaria::user_create()
{
	_msk = new TMask("at5200a");
	_rel = new TRelation(LF_SOGGETTI);
	_rel->add(LF_SEZIONI,"CODSEZ==CODSEZ|CODSOT==CODSOT");
  _rel->add("TCS", "CODTAB==STATO",1,0,ALIAS_TCS);
  _rel->add("LCP", "CODTAB==RES_CODLOC",1,LF_SOGGETTI,ALIAS_LCPRES);
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,LF_SOGGETTI,ALIAS_LCPDOM);
  _rel->add(LF_COMUNI, "COM==RES_CODCOM",1,LF_SOGGETTI,ALIAS_COMRES);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM",1,LF_SOGGETTI,ALIAS_COMDOM);
  _rel->add(LF_COMUNI, "COM==COMNASC",1,LF_SOGGETTI,ALIAS_COMNAS);
  
	//cursore ordinamento per sezione+sottogruppo+cognome+nome  
  _cur = add_cursor(new TCursor(_rel, "", 3));
	_contsan = new TLocalisamfile(LF_CONTSAN);
	_scontrolli = new TRecord_array(LF_CONTSAN, CON_PROGCON);
  TConfig config(CONFIG_STUDIO);
  TString16 etformato = config.get("EtFormato");
  _etlarghezza = config.get_int("EtLarghezza");
  _etcolonne = config.get_int("EtColonne");
	_form_eti = new TEti_sitsanitaria_form(etformato);
	
	return TRUE;
}	

bool TSitSanitaria::user_destroy()	
{
	delete _form_eti;
	delete _scontrolli;
	delete _contsan;
	delete _rel;
	delete _msk;
	return TRUE;
}

int at5200(int argc, char* argv[])
{
	TSitSanitaria a;
	a.run(argc, argv, "Stampa situazione sanitaria");
	return 0;
}