#include <form.h>
#include <mask.h>
#include <printapp.h>
#include <recarray.h>
#include <tabutil.h>
#include <utility.h>

#include "at5.h"

#define ALIAS_TCS	200
#define ALIAS_LCPRES	101		// localita' postale di residenza
#define ALIAS_LCPDOM	100		// localita' postale di domicilio
#define ALIAS_COMRES 	502		// comune di residenza
#define ALIAS_COMDOM 	501		// comune di domicilio
#define ALIAS_COMNAS 	500		// comune di nascita

// nomi campi maschera
#include "at5300a.h"
#include "at4100b.h"
#include "at4100c.h"

#include "atlib.h"
          
// nomi dei campi
#include "soggetti.h"
#include "contsan.h"
#include "sezioni.h"
#include <comuni.h>

enum ts { undefined = 0, elenco = 1, etichette = 2, cartoline = 3 };

// definizione form per etichette
class TEti_turnicon_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TEti_turnicon_form(): TForm() {};
  TEti_turnicon_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TEti_turnicon_form() {};
};

class TTurniControlli : public TPrintapp
{
	TMask*						_msk;
	TRelation*   			_rel;
	TLocalisamfile*		_contsan;
	TRecord_array* 		_scontrolli;
  TEti_turnicon_form*	_form_eti;
  TEti_turnicon_form*	_form_car;
  TAssoc_array		  _categorie, _tipicon;
	int 							_cur;	
	int								_contatore, _totale;
	TDate							_data_stampa;
	bool							_maifatto;
	bool 							_usasez;
	TDate							_primadel, _dataini, _datafin;
  TParagraph_string _cognome_nome, _rigastampa, _dencom;
	ts 	      				_tipostampa;
	TString16					_codsez, _codsot, _orario, _datacart;
	TString80					_invitoper, _data, _presso1, _presso2, _presso3, _presso4;
	TString80					_note, _intest1, _intest2, _intest3, _intest4;
	int								_etlarghezza, _etcolonne;

	static bool filter_func_turnicon(const TRelation* rel);

protected:
	virtual bool user_create();
	virtual bool user_destroy();
	virtual bool set_print(int m);
	virtual void set_page(int file, int cnt);
	virtual bool preprocess_page(int file, int counter);
  virtual print_action postprocess_print(int file, int counter);
  ts dati_cartoline();
	
public:
  void filtra_sezioni();
  void crea_intestazione();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  void dati_sezione(const TString16 codsez, const TString16 codsot);
  void footer_sezione();
  void fine_stampa();
	TTurniControlli() : _data_stampa(TODAY), _cognome_nome("",51), _rigastampa("", 71), _dencom("",51)  {}
};

HIDDEN inline TTurniControlli& app() { return (TTurniControlli&) main_app(); }

TCursor* TEti_turnicon_form::cursor() const { return app().current_cursor(); }

TRelation* TEti_turnicon_form::relation() const { return cursor()->relation(); }

ts TTurniControlli::dati_cartoline()
{
	TMask msk("at4100b");
 	if (msk.run() == K_ENTER) 
 	{ 
 		_invitoper = msk.get(F_INVITOPER);
 		_data			 = msk.get(F_DATA);
 		_presso1   = msk.get(F_PRESSO1);
 		_presso2   = msk.get(F_PRESSO2);
 		_presso3   = msk.get(F_PRESSO3);
 		_presso4   = msk.get(F_PRESSO4);
 		_datacart  = msk.get(F_DATACART);
 		_orario  	 = msk.get(F_ORARIO);
 		_note   	 = msk.get(F_NOTE);
 		_usasez		 = msk.get_bool(F_USASEZ);
 		return cartoline;
 	}
 	else
 		return undefined;
}

print_action TTurniControlli::postprocess_print(int file, int counter)
{ 
	if (_contatore > 0)
		footer_sezione();
	fine_stampa();				
	return NEXT_PAGE;
}

void TTurniControlli::dati_sezione(const TString16 codsez, const TString16 codsot)
{
	_intest1 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ1);
	_intest2 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ2);
	_intest3 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ3);
	_intest4 = current_cursor()->curr(LF_SEZIONI).get(SEZ_INTESTAZ4);
	if (_usasez)
	{
		if (_presso1.blank())
		{
			_presso1 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE1);
			_presso2 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE2);
			_presso3 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE3);
			_presso4 = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVSEDE4);
		}			
		if (_orario.blank())
			_orario = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVORA);
		if (_note.blank())
			_note = current_cursor()->curr(LF_SEZIONI).get(SEZ_CONVNOTE);
	}
	return;
}

void TTurniControlli::fine_stampa()
{
	// stampa totale soggetti a fine stampa
	reset_footer();
	TString sep(80);
	sep.fill('-');
	set_footer(2, (const char *) sep);
	if (_totale > 0)
	{
		if (_totale != _contatore)
		{
			set_footer(3,"TOTALE SOGGETTI STAMPATI %d", _totale);
			printer().formfeed();		
		}			
	}			
	else		
	{
		set_footer(3,"NON CI SONO SOGGETTI DA STAMPARE");		
		printer().formfeed();		
	}		
	reset_footer(); 
}

void TTurniControlli::footer_sezione()
{
	if (_contatore > 0)
	{
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		set_footer(3,"TOTALE SOGGETTI PER SEZIONE %d", _contatore);
		printer().formfeed(); 
		reset_footer();
	}				
}

void TTurniControlli::filtra_sezioni()
{
 	const TString16 sezini = _msk->get(F_SEZINI);
 	const TString16 sotini = _msk->get(F_SOTINI);
 	const TString16 sezfin = _msk->get(F_SEZFIN);
 	const TString16 sotfin = _msk->get(F_SOTFIN);
  TRectype da(LF_SOGGETTI);
  TRectype a(LF_SOGGETTI);   
  if (sezini.not_empty())
		da.put(SOG_CODSEZ, sezini);
  if (sotini.not_empty())
		da.put(SOG_CODSOT, sotini);
  if (sezfin.not_empty())
		a.put(SOG_CODSEZ, sezfin);
  if (sotfin.not_empty())
		a.put(SOG_CODSOT, sotfin);
	current_cursor()->setregion(da, a);
}	

bool TTurniControlli::preprocess_page(int file, int counter)
{ 
	if (_tipostampa == elenco)
	{
		TRectype& recsog = current_cursor()->curr();
		TString80 nome = recsog.get(SOG_COGNOME);
		nome << " ";
		nome << recsog.get(SOG_NOME);
		_cognome_nome = nome;
		TString256 localita = "";
		localita = recsog.get(SOG_DOM_CAP);			
		localita << ' ';
		if (recsog.get(SOG_DOM_CODLOC).not_empty())
		{
			localita << current_cursor()->curr(-ALIAS_LCPDOM).get("S0");
			localita << " - ";		
		}
		localita << current_cursor()->curr(LF_COMUNI).get(COM_DENCOM);
		localita.trim();
		_dencom = localita;
		TDate data(NULLDATE);
		TRectype* key = new TRectype(LF_CONTSAN);
		long codice = recsog.get_long(SOG_CODICE);
		key->put(CON_CODICE, codice);
		int err = _scontrolli->read(key);
		for (int r=1; r<=_scontrolli->rows(); r++)
		{
			const TRectype& riga = _scontrolli->row(r);
			TString16 tipocon = riga.get(CON_TIPOCON);
			TDate datacon = riga.get(CON_DATACON);
			//if ((tipocon == _tipo1 && _tipo1.not_empty()) || (tipocon == _tipo2 && _tipo2.not_empty()))
			if (_tipicon.is_key((const char*)tipocon))
				data = datacon;
		}					   
		TString80 rigastampa;
		rigastampa = "";
		if (data.ok())
			rigastampa << "@72g"<<data;
		else
			rigastampa << "@72g"<<"mai effettuato";
		_rigastampa = rigastampa;
		// salto pagina se cambio sezione
		const TString16 codsez = recsog.get(SOG_CODSEZ);		
		const TString16 codsot = recsog.get(SOG_CODSOT);		
		if ((_codsez!=codsez)||(_codsot!=codsot))
		{
			if ((_codsez != "**") && (_contatore > 0))
				footer_sezione();
			_contatore = 0;			
			_codsez = codsez;
			_codsot = codsot;
			header_sezione(codsez, codsot);
		}
		if (printer().rows_left() < 4)
			printer().formfeed();
	}		
	if (_tipostampa==etichette)		
		if (printer().rows_left() < _form_eti->get_body().height())
			printer().formfeed();
	_contatore++;			
	_totale++;
  return TRUE;
}

void TTurniControlli::set_page(int file, int cnt)
{ 
	switch (_tipostampa)
	{
		case cartoline:
		{
			TPrint_section& corpo = _form_car->get_body();
			const TString16 codsez = current_cursor()->curr().get(SOG_CODSEZ);		
			const TString16 codsot = current_cursor()->curr().get(SOG_CODSOT);		
			if ((_codsez!=codsez)||(_codsot!=codsot))
			{
				_codsez = codsez;
				_codsot = codsot;
				dati_sezione(codsez, codsot);
				corpo.reset();
			  TForm_item& intest1 = corpo.find_field(CAR_INTEST1);
			  intest1.set(_intest1);
			  TForm_item& intest2 = corpo.find_field(CAR_INTEST2);
			  intest2.set(_intest2);
			  TForm_item& intest3 = corpo.find_field(CAR_INTEST3);
			  intest3.set(_intest3);
			  TForm_item& intest4 = corpo.find_field(CAR_INTEST4);
			  intest4.set(_intest4);
			  TForm_item& invito = corpo.find_field(CAR_INVITOPER);
	  		invito.set(_invitoper);
			  TForm_item& data = corpo.find_field(CAR_DATA);
	  		data.set(_data);
			  TForm_item& presso1 = corpo.find_field(CAR_PRESSO1);
	  		presso1.set(_presso1);
			  TForm_item& presso2 = corpo.find_field(CAR_PRESSO2);
	  		presso2.set(_presso2);
			  TForm_item& presso3 = corpo.find_field(CAR_PRESSO3);
	  		presso3.set(_presso3);
			  TForm_item& presso4 = corpo.find_field(CAR_PRESSO4);
	  		presso4.set(_presso4);
			  TForm_item& orario = corpo.find_field(CAR_ORARIO);
	  		orario.set(_orario);
			  TForm_item& datacart = corpo.find_field(CAR_DATACART);
	  		datacart.set(_datacart);
			  TForm_item& note = corpo.find_field(CAR_NOTE);
	  		note.set(_note);
			}
			corpo.update();
  		for (word i = 0; i < corpo.height(); i++)
  		{
  			TPrintrow& riga = corpo.row(i);
  			set_row(i+1,riga);
			}  			
			force_setpage(TRUE);	// serve perch alla prossima etichetta rifaccia la setpage
														// altrimenti stampa sempre la stessa etichetta
		}
		break;
		case etichette:
		{
			TPrint_section& corpo = _form_eti->get_body();
			corpo.update();
  		for (word i = 0; i < corpo.height(); i++)
  		{
  			TPrintrow& riga = corpo.row(i);
  			set_row(i+1,riga);
			}  			
			force_setpage(TRUE);	// serve perch alla prossima etichetta rifaccia la setpage
														// altrimenti stampa sempre la stessa etichetta
		}
		break;
		case elenco:
		{
			set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
			set_row(2,"@0g@S", FLD(LF_SOGGETTI,SOG_TESSAVIS));
			set_row(3,"@0g@S", FLD(LF_SOGGETTI,SOG_CATDON));
			
			set_row(1,"@9g#a", &_cognome_nome);
			set_row(2,"@9g@S", FLD(LF_SOGGETTI,SOG_DOM_INDIR));
			set_row(3,"@9g#a", &_dencom);

			set_row(1,"@61g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
			set_row(2,"@61g@S", FLD(LF_SOGGETTI,SOG_GRUPPOAB0));
			set_row(2,"@66g@S", FLD(LF_SOGGETTI,SOG_RHANTID));

			set_row(1,"@72g@S", FLD(LF_SOGGETTI,SOG_TELABI));
			set_row(1,"@87g@S", FLD(LF_SOGGETTI,SOG_TELLAV));
			set_row(1,"@102g@S", FLD(LF_SOGGETTI,SOG_TELALT));
			
			set_row(3,"@72g#a", &_rigastampa);
			
			set_row(4,"");
		}			
		break;
	}		
}

bool TTurniControlli::filter_func_turnicon(const TRelation* rel)
{   
	bool filtrato = TRUE;
	TRectype& recsog = rel->lfile().curr();
	// filtro per categorie
	TAssoc_array& categorie = app()._categorie;
	if (categorie.items() != 0)
	{
		const TString16 cat = recsog.get(SOG_CATDON);
		filtrato = categorie.is_key((const char*) cat);
	}	
	// filtro per data di nascita
	if (filtrato)
	{ 
		const TDate datanasc = recsog.get_date(SOG_DATANASC);
		if (app()._dataini.ok())
			filtrato = datanasc >= app()._dataini;
		if (filtrato)
			if (app()._datafin.ok())
				filtrato = datanasc <= app()._datafin;
	}
	if (filtrato)
	{                                        
		bool fatto = FALSE;
		TRectype* key = new TRectype(LF_CONTSAN);
		long codice = recsog.get_long(SOG_CODICE);
		key->put(CON_CODICE, codice);
		int err = app()._scontrolli->read(key);
		for (int r=1; r<=app()._scontrolli->rows(); r++)
		{
			const TRectype& riga = app()._scontrolli->row(r);
			TString16 tipocon = riga.get(CON_TIPOCON);
			//if ((tipocon == app()._tipo1) || (tipocon == app()._tipo2))
			if (app()._tipicon.is_key((const char*) tipocon))
			{                  
				fatto = TRUE;              
				if (app()._primadel.ok())
				{ 
					TDate datacon = riga.get(CON_DATACON);
					if (datacon >= app()._primadel)
						filtrato = FALSE;
				}
				else
					filtrato = FALSE;
			}				
	  }
	  if (!app()._maifatto && !fatto)
	  	filtrato = FALSE;
	}	
	return filtrato;
}

void TTurniControlli::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(100);
	intestazione = "Sezione: ";
	intestazione << codsez;
	intestazione << "/";
	intestazione << codsot;
	intestazione << " ";
	intestazione << densez;
	if (densot.not_empty())
	{
		intestazione << "/";
		intestazione << densot;
	}		
	intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

bool TTurniControlli::set_print(int)
{
  TPrinter& p = printer();
	p.read_configuration();
	_tipostampa = undefined;
  KEY tasto;
  tasto = _msk->run();
  switch (tasto)
  {
  case F_ELENCO:
  	_tipostampa = elenco;
  	_codsez = "**";
  	_codsot = "**";
  break;
  case F_ETICHETTE:
  	_tipostampa = (configura_stampante(p, "AT_ETICHETTE", "etichette")) ? etichette : undefined;
  break;
  case F_CARTOLINE:
  	_tipostampa = (configura_stampante(p, "AT_CARTOLINE", "cartoline")) ? dati_cartoline() : undefined;
  break;
  }	
	if (_tipostampa != undefined)
  {
  	_tipicon.destroy();
  	TString16 tipo;
		tipo = _msk->get(F_TIPO1);
		if (tipo.not_empty())
  		_tipicon.add((const char*) tipo);
		tipo = _msk->get(F_TIPO2);
		if (tipo.not_empty())
  		_tipicon.add((const char*) tipo);
		tipo = _msk->get(F_TIPO3);
		if (tipo.not_empty())
  		_tipicon.add((const char*) tipo);
		tipo = _msk->get(F_TIPO4);
		if (tipo.not_empty())
  		_tipicon.add((const char*) tipo);
		tipo = _msk->get(F_TIPO5);
		if (tipo.not_empty())
  		_tipicon.add((const char*) tipo);
		tipo = _msk->get(F_TIPO6);
		if (tipo.not_empty())
  		_tipicon.add((const char*) tipo);
		tipo = _msk->get(F_TIPO7);
		if (tipo.not_empty())
  		_tipicon.add((const char*) tipo);
		tipo = _msk->get(F_TIPO8);
		if (tipo.not_empty())
  		_tipicon.add((const char*) tipo);
		tipo = _msk->get(F_TIPO9);
		if (tipo.not_empty())
  		_tipicon.add((const char*) tipo);
		tipo = _msk->get(F_TIPO10);
		if (tipo.not_empty())
  		_tipicon.add((const char*) tipo);
		_maifatto = _msk->get_bool(F_MAIFATTO);
		_primadel = _msk->get_date(F_PRIMADEL);
		_dataini = _msk->get_date(F_DATAINI);
		_datafin = _msk->get_date(F_DATAFIN);
		_contatore = 0;
		_totale = 0;
    reset_files(); 
    add_file(LF_SOGGETTI);
		filtra_sezioni();
	  // filtro per categorie
	  _categorie.destroy();
  	const TString16 catpri = _msk->get(F_CAT1);
  	const TString16 catsec = _msk->get(F_CAT2);
  	const TString16 catter = _msk->get(F_CAT3);
  	const TString16 catqua = _msk->get(F_CAT4);
  	const TString16 catqui = _msk->get(F_CAT5);
  	const TString16 catses = _msk->get(F_CAT6);
	  if (catpri.not_empty())
  		_categorie.add((const char*) catpri);
  	if (catsec.not_empty())
  		_categorie.add((const char*) catsec);
  	if (catter.not_empty())
  		_categorie.add((const char*) catter);
  	if (catqua.not_empty())
  		_categorie.add((const char*) catqua);
  	if (catqui.not_empty())
  		_categorie.add((const char*) catqui);
  	if (catses.not_empty())
  		_categorie.add((const char*) catses);
		current_cursor()->set_filterfunction(filter_func_turnicon,TRUE);
		reset_print();		
		printer().footerlen(0);		
    crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TTurniControlli::crea_intestazione()
{     
	reset_header();
  if (_tipostampa == elenco)
  {
		TString sep(132);
	 	sep = "STAMPA TURNI PER CONTROLLI SANITARI "; 	
		char* tc;
	  for (tc = (char*)_tipicon.first_item(); tc != NULL; tc = (char*) _tipicon.succ_item())
	  {
	    sep << _tipicon.get_hashobj()->key();
	    sep << " ";
	 	}
	 	if (_maifatto)
	 		sep << " - mai effettuato";
		if (_primadel.ok())
			sep << " - effettuato prima del " << _primadel.string(); 		
		sep.center_just();
	 	set_header(2, "@0g%s", (const char*) sep);
	 	TDate data_stampa = _data_stampa.string();
	  set_header(2,"@0g%10s", (const char*) data_stampa);
	  sep = "";
	  sep << "Pag. @#";
	  set_header(2, "@125g%s", (const char*) sep);  
	  	
	  sep = "";
	  sep.fill('-');
	  set_header(3, (const char *) sep);
	 	set_header(4,"@0gCodice@9gCognome e nome@61gNato il@72gTelefono abit.@87gTelefono lav.@102gTelefono altro");
	 	set_header(5,"@0gTessera@9gIndirizzo");
	 	set_header(6,"@0gCat.@61gGruppo Rh@72gData ult. controllo scelto");
	 	set_header(7,"@0g--------@9g---------------------------------------------------@61g----------@72g------------------------------------------------------");
	 	printer().footerlen(3);
	}	 	
}

bool TTurniControlli::user_create()
{
	_msk = new TMask("at5300a");
	_rel = new TRelation(LF_SOGGETTI);
	_rel->add(LF_SEZIONI,"CODSEZ==CODSEZ|CODSOT==CODSOT");
  _rel->add("TCS", "CODTAB==STATO",1,0,ALIAS_TCS);
//  _rel->add("LCP", "CODTAB==RES_CODLOC",1,LF_SOGGETTI,ALIAS_LCPRES);
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,LF_SOGGETTI,ALIAS_LCPDOM);
//  _rel->add(LF_COMUNI, "COM==RES_CODCOM",1,LF_SOGGETTI,ALIAS_COMRES);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM",1,LF_SOGGETTI,ALIAS_COMDOM);
  _rel->add(LF_COMUNI, "COM==COMNASC",1,LF_SOGGETTI,ALIAS_COMNAS);
  
	//cursore ordinamento per sezione+sottogruppo+cognome+nome  
  _cur = add_cursor(new TCursor(_rel, "", 3));
	_contsan = new TLocalisamfile(LF_CONTSAN);
	_scontrolli = new TRecord_array(LF_CONTSAN, CON_PROGCON);
  TConfig config(CONFIG_STUDIO);
  TString16 etformato = config.get("EtFormato");
  _etlarghezza = config.get_int("EtLarghezza");
  _etcolonne = config.get_int("EtColonne");
	_form_eti = new TEti_turnicon_form(etformato);
	_form_car = new TEti_turnicon_form("ATCARTO1");
	return TRUE;
}	

bool TTurniControlli::user_destroy()	
{
	delete _form_car;
	delete _form_eti;
	delete _scontrolli;
	delete _contsan;
	delete _rel;
	delete _msk;
	return TRUE;
}

int at5300(int argc, char* argv[])
{
	TTurniControlli a;
	a.run(argc, argv, "Stampa turni per controlli");
	return 0;
}