#include <form.h>
#include <mask.h>
#include <printapp.h>
#include <utility.h>

#include "soggetti.h"
#include "sezioni.h"
#include <comuni.h>

#include "at7.h"
#include "at7600a.h"
#include "atlib.h"

#define ALIAS_LCPDOM	100
#define ALIAS_TCS			200

enum ts { undefined=0, sintetico=1, completo=2, etichette=3, lettere=4 };

// definizione form per etichette
class TEti_consenso_form : public TForm
{
public:
	
	virtual TCursor* cursor() const;
	virtual TRelation* relation() const;
	TPrint_section& get_body() { return section('B'); } ;
  TEti_consenso_form(): TForm() {};
  TEti_consenso_form(const char* form, const char * code = "", int editlevel = 0, const char* desc = "")
    						: TForm(form,code,editlevel,desc) {};
  virtual ~TEti_consenso_form() {};
};

class TStampaConsenso : public TPrintapp
{
  TRelation*					_rel;
  TMask*      				_msk;
  TEti_consenso_form*	_form_eti;
  TEti_consenso_form*	_form_let;
  TAssoc_array		    _categorie;
  TParagraph_string   _cognome_nome, _dencom;
  TDate 							_data_stampa;
	ts 	      					_tipostampa;
	TString16						_codsez, _codsot;
	int									_etlarghezza, _etcolonne, _totale, _contatore;
	char								_consenso;

  static bool filter_func_consenso(const TRelation* rel);
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page(int file, int counter);
  virtual print_action postprocess_print(int file, int counter);

public:
  void crea_intestazione();
  void filtra_sezioni();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  void footer_sezione();
  void fine_stampa();
  
  TMask& app_mask() { return *_msk; }
  TStampaConsenso() : _data_stampa(TODAY), _cognome_nome("",25), _dencom("",30) {}
};

HIDDEN inline TStampaConsenso& app() { return (TStampaConsenso&) main_app(); }

TCursor* TEti_consenso_form::cursor() const { return app().current_cursor(); }

TRelation* TEti_consenso_form::relation() const { return cursor()->relation(); }

void TStampaConsenso::filtra_sezioni()
{
 	const TString16 sezini = _msk->get(F_SEZINI);
 	const TString16 sotini = _msk->get(F_SOTINI);
 	const TString16 sezfin = _msk->get(F_SEZFIN);
 	const TString16 sotfin = _msk->get(F_SOTFIN);
  TRectype da(LF_SOGGETTI);
  TRectype a (LF_SOGGETTI);
  if (sezini.not_empty())
		da.put(SOG_CODSEZ, sezini);
  if (sotini.not_empty())
		da.put(SOG_CODSOT, sotini);
  if (sezfin.not_empty())
		a.put(SOG_CODSEZ, sezfin);
  if (sotfin.not_empty())
		a.put(SOG_CODSOT, sotfin);
	current_cursor()->setregion(da, a);
}	

void TStampaConsenso::set_page(int file, int cnt)
{
	// costruzione etichette
	switch (_tipostampa)
	{
		case etichette:
		{
			TPrint_section& corpo = _form_eti->get_body();
			corpo.update();
  		for (word i = 0; i < corpo.height(); i++)
  		{
  			TPrintrow& riga = corpo.row(i);
  			set_row(i+1,riga);
			}  			
			force_setpage(TRUE);	// serve perch alla prossima etichetta rifaccia la setpage
														// altrimenti stampa sempre la stessa etichetta
		}
		break;
		case lettere:
		{
			TPrint_section& corpo = _form_let->get_body();
			corpo.update();
  		for (word i = 0; i < corpo.height(); i++)
  		{
  			TPrintrow& riga = corpo.row(i);
  			set_row(i+1,riga);
			}  			
			force_setpage(TRUE);	// serve perch alla prossima etichetta rifaccia la setpage
														// altrimenti stampa sempre la stessa etichetta
		}
		break;
		case sintetico:
		{
			set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
			set_row(1,"@9g@S", FLD(LF_SOGGETTI,SOG_CATDON));
			set_row(1,"@12g#a", &_cognome_nome);
			set_row(1,"@38g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
			set_row(1,"@49g@6,rs", FLD(LF_SOGGETTI,SOG_TESSAVIS));
			set_row(1,"@56g@ld", FLD(LF_SOGGETTI,SOG_DATAULTDON));
			set_row(1,"@67g@S", FLD(LF_SOGGETTI,SOG_TIPOULTDON));
			set_row(1,"@70g@S", FLD(LF_SOGGETTI,SOG_CONSENSO));
			set_row(2,"");
		}
		break;
		case completo:
		{
			set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
			set_row(2,"@0g@8,rs", FLD(LF_SOGGETTI,SOG_TESSAVIS));
			set_row(3,"@0g@8,rs", FLD(LF_SOGGETTI,SOG_CATDON));

			set_row(1,"@9g@S",FLD(LF_SOGGETTI,SOG_COGNOME));
			set_row(2,"@9g@S",FLD(LF_SOGGETTI,SOG_NOME));
			set_row(3,"@9g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));

			set_row(1,"@35g@S", FLD(LF_SOGGETTI,SOG_DOM_INDIR));
			set_row(2,"@35g#a", &_dencom);
	
			set_row(1,"@66g@S", FLD(LF_SOGGETTI,SOG_TELABI));
			set_row(2,"@66g@S", FLD(LF_SOGGETTI,SOG_TELLAV));
			set_row(3,"@66g@S", FLD(LF_SOGGETTI,SOG_TELALT));

			set_row(1,"@82gGr @S", FLD(LF_SOGGETTI,SOG_GRUPPOAB0));
			set_row(2,"@82gRh @S", FLD(LF_SOGGETTI,SOG_RHANTID));
			set_row(3,"@82gFe @S", FLD(LF_SOGGETTI,SOG_FENOTIPORH));
			set_row(4,"@82gKe @S", FLD(LF_SOGGETTI,SOG_KELL));
			set_row(5,"@82gDu @S", FLD(LF_SOGGETTI,SOG_DU));

			set_row(1,"@94g@ld", FLD(LF_SOGGETTI,SOG_DATAULTDON));
			set_row(2,"@94g@S", FLD(LF_SOGGETTI,SOG_TIPOULTDON));
			set_row(4,"@94g@pn", FLD(LF_SOGGETTI,SOG_TOTDON));

			set_row(1,"@105g@S", FLD(LF_SOGGETTI,SOG_CONSENSO));

			set_row(6,"");
		}
		break;
	}					
}

bool TStampaConsenso::filter_func_consenso(const TRelation* rel)
{   
	bool filtrato = TRUE;
	TRectype recsog = rel->lfile().curr();
	// filtro per categorie
	TAssoc_array& categorie = app()._categorie;
	if (categorie.items() != 0)
	{
		const TString16 cat = recsog.get(SOG_CATDON);
		filtrato = categorie.is_key((const char*) cat);
	}	
  return filtrato;
}

bool TStampaConsenso::preprocess_page(int file, int counter)
{
	if ((_tipostampa==sintetico) || (_tipostampa==completo))
	{
		TRectype& recsog = current_cursor()->curr();
		TString80 nome = recsog.get(SOG_COGNOME);
		nome << ' ';
		nome << recsog.get(SOG_NOME);
		_cognome_nome = nome;
		TString256 localita = "";
		localita << recsog.get(SOG_DOM_CAP);
		localita << ' ';
		if (recsog.get(SOG_DOM_CODLOC).not_empty())
			localita << current_cursor()->curr(-ALIAS_LCPDOM).get("S0");
		else
			localita << current_cursor()->curr(LF_COMUNI).get(COM_DENCOM);
		localita << ' ' << current_cursor()->curr(LF_COMUNI).get(COM_PROVCOM);
		localita.trim();
		_dencom = localita;                
		// salto pagina se cambio sezione
		const TString16 codsez = recsog.get(SOG_CODSEZ);		
		const TString16 codsot = recsog.get(SOG_CODSOT);		
		if ((_codsez!=codsez)||(_codsot!=codsot))
		{
			if (_codsez != "**")
				footer_sezione();
			_contatore = 0;	
			_codsez = codsez;
			_codsot = codsot;
			header_sezione(codsez, codsot);
		}
	}		
	if (_tipostampa==etichette)		
		if (printer().rows_left() < _form_eti->get_body().height())
			printer().formfeed();
	_contatore++;			
	_totale++;
  return TRUE;
}

print_action TStampaConsenso::postprocess_print(int file, int counter)
{ 
	if (_contatore > 0)
		footer_sezione();
	fine_stampa();	
	return NEXT_PAGE;
}

void TStampaConsenso::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(132);
	intestazione = "Sezione: ";
	intestazione << codsez;
	if (codsot.not_empty())
	{
		intestazione << '/';
		intestazione << codsot;
	}		
	intestazione << ' ';
	intestazione << densez;
	if (densot.not_empty())
	{
		intestazione << '/';
		intestazione << densot;
	}		
	intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}
                          
void TStampaConsenso::footer_sezione()
{
	// stampa totale soggetti appartenenti alla sezione
  if (_tipostampa==sintetico || _tipostampa==completo)
	{
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		set_footer(3,"TOTALE SOGGETTI PER SEZIONE %d", _contatore);
		printer().formfeed(); 
		reset_footer();
	}	
}
                          
void TStampaConsenso::fine_stampa()
{
	// stampa totale soggetti a fine stampa
  if (_tipostampa==sintetico || _tipostampa==completo)
	{
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(2, (const char *) sep);
		if (_totale > 0 && _totale != _contatore)
		{
			set_footer(3,"TOTALE SOGGETTI STAMPATI %d", _totale);
			printer().formfeed();		
		}			
		reset_footer(); 
	}	
}

bool TStampaConsenso::set_print(int m)
{       
  TPrinter& p = printer();
	p.read_configuration();
	_tipostampa = undefined;
  KEY tasto;
  tasto = _msk->run();
  switch (tasto)
  {
  case F_SINTETICO:
  {
  	_tipostampa = sintetico;
  	_codsez = _codsot = "**";
  	_totale = _contatore = 0;
	}
  break;
  case F_COMPLETO:
  {
  	_tipostampa = completo;
  	_codsez = _codsot = "**";
  	_totale = _contatore = 0;
	}  	
  break;
  case F_ETICHETTE:
  	_tipostampa = (configura_stampante(p, "AT_ETICHETTE", "etichette")) ? etichette : undefined;
  break;
  case F_LETTERE:
  	_tipostampa = lettere;	
  break;
  }	
	if (_tipostampa != undefined)
  {
    reset_files(); 
    add_file(LF_SOGGETTI);
    // filtro per sezioni selezionate
	  filtra_sezioni();
    _consenso = _msk->get(F_CONSENSO)[0];
	  TString filtro;
		// filtro per consenso
		if (_consenso=='S')
	    current_cursor()->setfilter("(90->CONSENSO==\"X\")", TRUE);
	  else
	  	if (_consenso=='N')
		  	current_cursor()->setfilter("(90->CONSENSO!=\"X\")", TRUE);
	  // filtro per categorie
	  _categorie.destroy();
  	const TString16 catpri = _msk->get(F_CAT1);
  	const TString16 catsec = _msk->get(F_CAT2);
  	const TString16 catter = _msk->get(F_CAT3);
  	const TString16 catqua = _msk->get(F_CAT4);
  	const TString16 catqui = _msk->get(F_CAT5);
  	const TString16 catses = _msk->get(F_CAT6);
	  if (catpri.not_empty())
  		_categorie.add((const char*) catpri);
  	if (catsec.not_empty())
  		_categorie.add((const char*) catsec);
  	if (catter.not_empty())
  		_categorie.add((const char*) catter);
  	if (catqua.not_empty())
  		_categorie.add((const char*) catqua);
  	if (catqui.not_empty())
  		_categorie.add((const char*) catqui);
  	if (catses.not_empty())
  		_categorie.add((const char*) catses);
		current_cursor()->set_filterfunction(filter_func_consenso);
		reset_print();
		printer().footerlen(0);
		crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TStampaConsenso::crea_intestazione()
{
  reset_header();
  if (_tipostampa==sintetico || _tipostampa==completo)
  {
  	TString sep(132);
  	sep = "SOGGETTI PER CONSENSO ";           
  	if (_consenso=='S')
  		sep << "Si";
  	else	
  		if (_consenso=='N')
	  		sep << "No";
	  	else
	  		sep << "Indifferente";	
		sep.center_just();
  	set_header(2, "@0g%s", (const char*) sep);
  	const TString16 data_stampa = _data_stampa.string();
  	set_header(2,"@0g%10s", (const char*) data_stampa);
  	sep = "";
  	sep << "Pag. @#";
  	set_header(2, "@120g%s", (const char*) sep);  
  	sep = "";
  	sep.fill('-');
  	set_header(3, (const char *) sep);
  	if (_tipostampa==sintetico)
  	{
	  	sep.fill('-',100);
  		set_header(4, (const char *) sep);
  		set_header(5,"@0gCodice@9gC.@12gCognome e nome@38gNato il@49gTess.@56gData/Tipo u.d.@70Consenso");
  		set_header(6,"@0g--------@9g--@12g-------------------------@38g----------@49g------@56g---------- --@70g--------");
		}
		else  			
		{
	  	sep.fill('-');
  		set_header(4, (const char *) sep);
  		set_header(5,"@0gCodice@9gCognome@35gIndirizzo@66gTel.abitazione@82gTipiz@94gUlt.donaz.@105gConsenso");
	  	set_header(6,"@0gTessera@9gNome@66gTel.lavoro@82gzazione@94gTipo");
	  	set_header(7,"@0gCateg.@9gData nascita@66gTel.altro");
	  	set_header(8,"@94gTot.don.");
  		set_header(10,"@0g--------@9g-------------------------@35g------------------------------@66g--------------@82g------------@94g----------@105g--------");
		}
		printer().footerlen(3);
	}  	
}

bool TStampaConsenso::user_create()
{
  _rel = new TRelation(LF_SOGGETTI);
  _rel->add("TCS", "CODTAB==STATO",1,0,ALIAS_TCS);
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,0,ALIAS_LCPDOM);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM");
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT");
  add_cursor(new TCursor(_rel,"", 3));
  _msk = new TMask("at7600a");
  TConfig config(CONFIG_STUDIO);
  TString16 etformato = config.get("EtFormato");
  _etlarghezza = config.get_int("EtLarghezza");
  _etcolonne = config.get_int("EtColonne");
	_form_eti = new TEti_consenso_form(etformato);
	_form_let = new TEti_consenso_form("ATLETTER");
  return TRUE;
}

bool TStampaConsenso::user_destroy()
{
	delete _form_let;
  delete _form_eti;
  delete _msk;
  delete _rel;
  return TRUE;
}

int at7600(int argc, char* argv[])
{
  TStampaConsenso a;
  a.run(argc, argv, "Soggetti per consenso");
  return 0;
}
