#include <form.h>
#include <mask.h>
#include <printapp.h>

#include "soggetti.h"
#include "sezioni.h"
#include <comuni.h>

#include "at7.h"
#include "at7700a.h"

#define ALIAS_LCP	100
#define ALIAS_TCS	200

class TStampaNonDom : public TPrintapp
{
  static bool filter_func_nondom(const TRelation* rel);

  TRelation*        _rel;
  TMask*            _msk;
  TAssoc_array			_catdon;
  TAssoc_array			_catnondon;
  TParagraph_string _cognome_nome, _dencom;
  TDate 						_data_stampa;
  TString16					_codsez, _codsot, _codcom;
  TString16					_tipostampa;
  int								_contatore;
  
protected:
  virtual bool user_create();
  virtual bool user_destroy();
  virtual bool set_print(int m);
  virtual void set_page(int file, int cnt);
  virtual bool preprocess_page (int file, int counter);

public:
  void crea_intestazione();
  void filtra_sezioni();
  void header_sezione(const TString16 codsez, const TString16 codsot);
  void header_comune(const TString16 codcom);
  void footer_stampa();
  TMask& app_mask() { return *_msk; }

  TStampaNonDom() : _data_stampa(TODAY), _cognome_nome("",25), _dencom("",50) {}
};

HIDDEN inline TStampaNonDom& app() { return (TStampaNonDom&) main_app(); }

void TStampaNonDom::filtra_sezioni()
{
 	const TString16 sezini = _msk->get(F_SEZINI);
 	const TString16 sotini = _msk->get(F_SOTINI);
 	const TString16 sezfin = _msk->get(F_SEZFIN);
 	const TString16 sotfin = _msk->get(F_SOTFIN);
  TRectype da(LF_SOGGETTI);
  TRectype a(LF_SOGGETTI);   
	if (sezini.not_empty())  
		da.put(SOG_CODSEZ, sezini);
  if (sotini.not_empty())
		da.put(SOG_CODSOT, sotini);
  if (sezfin.not_empty())
		a.put(SOG_CODSEZ, sezfin);
  if (sotfin.not_empty())
		a.put(SOG_CODSOT, sotfin);
	current_cursor()->setregion(da, a);
}	

void TStampaNonDom::set_page(int file, int cnt)
{
	set_row(1,"@0g@pn", FLD(LF_SOGGETTI,SOG_CODICE,"########"));
	set_row(2,"@0g@8,rs", FLD(LF_SOGGETTI,SOG_TESSAVIS));
	set_row(1,"@9g@S", FLD(LF_SOGGETTI,SOG_CATDON));
	set_row(1,"@12g#a", &_cognome_nome);
	set_row(1,"@38g@ld", FLD(LF_SOGGETTI,SOG_DATANASC));
	set_row(2,"@38g@S", FLD(LF_SOGGETTI,SOG_CODSEZ));
	set_row(2,"@41g@S", FLD(LF_SOGGETTI,SOG_CODSOT));
	set_row(1,"@49g@S", FLD(LF_SOGGETTI,SOG_DOM_INDIR));
	set_row(1,"@100g@S", FLD(LF_SOGGETTI,SOG_TELABI));
	set_row(1,"@116g@S", FLD(LF_SOGGETTI,SOG_TELLAV));
	set_row(2,"@49g#a", &_dencom);
	set_row(2,"@116g@S", FLD(LF_SOGGETTI,SOG_TELALT));
	set_row(3,"");
}

bool TStampaNonDom::filter_func_nondom(const TRelation* rel)
{   
	bool filtrato = FALSE;
	TRectype& recsog = rel->curr();
	TRectype& recsez = rel->curr(LF_SEZIONI);
	// filtro per categorie
	TAssoc_array& categorie = app()._catdon;
	if (categorie.items() != 0)
	{
		const TString16 cat = recsog.get(SOG_CATDON);
		filtrato = categorie.is_key((const char*) cat);
	}	
	if (!filtrato)
	{	
		// filtro per categorie non donatori
		TAssoc_array& categorie = app()._catnondon;
		if (categorie.items() != 0)
		{
			const TString16 cat1 = recsog.get(SOG_CATNOND1);
			const TString16 cat2 = recsog.get(SOG_CATNOND2);			
			filtrato = (categorie.is_key((const char*) cat1) || categorie.is_key((const char*) cat2)) ;
		}	
	}		
	if (app()._catdon.items() == 0 && app()._catnondon.items() == 0)
		filtrato = TRUE;
	if (filtrato)
	{
		TString16 comsez = recsez.get(SEZ_COM);
		TString16 comsog = recsog.get(SOG_DOM_CODCOM);
		filtrato = ((comsez != comsog) && comsez.not_empty() && comsog.not_empty());
	}		
  return filtrato;
}

void TStampaNonDom::header_sezione(const TString16 codsez, const TString16 codsot)
{
	const TString80 densez = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSEZ);
	const TString80 densot = current_cursor()->curr(LF_SEZIONI).get(SEZ_DENSOT);
	TString intestazione(132);
	intestazione = "Sezione: ";
	intestazione << codsez;
	intestazione << '/';
	intestazione << codsot;
	intestazione << ' ';
	intestazione << densez;
	if (densot.not_empty())
	{
		intestazione << '/';
		intestazione << densot;
	}		
	intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

void TStampaNonDom::header_comune(const TString16 codcom)
{
	TString80 dencom = current_cursor()->curr(LF_COMUNI).get(COM_DENCOM);
	dencom << ' ';
	dencom << current_cursor()->curr(LF_COMUNI).get(COM_PROVCOM);
	TString intestazione(132);
	intestazione = "Comune: ";
	intestazione << codcom;
	intestazione << ' ';
	intestazione << dencom;
	intestazione.center_just();
	set_header(1,"@0g%s", (const char*) intestazione);
	return;
}

void TStampaNonDom::footer_stampa()
{
	// stampa totale soggetti a fine pagina
	if (_contatore > 0)
	{
		reset_footer();
		TString sep(80);
		sep.fill('-');
		set_footer(1, (const char *) sep);
		set_footer(2,"TOTALE %d", _contatore);
		printer().formfeed(); 
		reset_footer();
	}		
}

bool TStampaNonDom::preprocess_page(int file, int counter)
{
	TRectype& recsog = current_cursor()->curr();
	TString80 nome = recsog.get(SOG_COGNOME);
	nome << ' ';
	nome << recsog.get(SOG_NOME);
	_cognome_nome = nome;
	TString256 localita = "";
	localita << current_cursor()->curr(-ALIAS_LCP).get("S6");
	if (localita.not_empty())
	{
		localita << ' ';
		localita << current_cursor()->curr(-ALIAS_LCP).get("S0");
		localita << " - ";		
	}
	else
	{
		localita = current_cursor()->curr(LF_COMUNI).get(COM_CAPCOM);
		localita << ' ';
	}			
	localita << current_cursor()->curr(LF_COMUNI).get(COM_DENCOM);
	localita << ' ';
	localita << current_cursor()->curr(LF_COMUNI).get(COM_PROVCOM);
	localita.trim();
	_dencom = localita;
	// salto pagina se cambio sezione o comune
	if (_tipostampa == "01")
	{
		const TString16 codsez = recsog.get(SOG_CODSEZ);		
		const TString16 codsot = recsog.get(SOG_CODSOT);		
		if ((_codsez!=codsez)||(_codsot!=codsot))
		{
			if (_codsez != "**")
				//printer().formfeed();
				footer_stampa();
			_contatore = 0;				
			_codsez = codsez;
			_codsot = codsot;
			header_sezione(codsez, codsot);
		}
	}
	else				
	{
		const TString16 codcom = recsog.get(SOG_DOM_CODCOM);		
		if (_codcom!=codcom)
		{
			if (_codcom != "**")
				//printer().formfeed();
				footer_stampa();
			_contatore = 0;				
			_codcom = codcom;
			header_comune(codcom);
		}
	}
	_contatore++;
	if (printer().rows_left()<3)
		printer().formfeed();
  return TRUE;
}

bool TStampaNonDom::set_print(int m)
{
  KEY tasto;
  tasto = _msk->run();
  if (tasto == K_ENTER)
  {
  	_contatore = 0;
  	_codsez = "**";
  	_codsot = "**";
  	_codcom = "****";
  	_tipostampa = _msk->get(F_TIPOSTAMPA);
  	TString80 chiave = "";
  	if (_tipostampa == "01")
  		chiave = "90->CODSEZ|90->CODSOT|90->COGNOME|90->NOME";
		else  		
  		chiave = "90->DOM_CODCOM|90->COGNOME|90->NOME";
    reset_files(); 
    add_file(LF_SOGGETTI);
		filtra_sezioni();
		_catdon.destroy();
	  const TString16 catpri = _msk->get(F_CAT1);
	  const TString16 catsec = _msk->get(F_CAT2);
	  const TString16 catter = _msk->get(F_CAT3);
	  const TString16 catqua = _msk->get(F_CAT4);
	  const TString16 catqui = _msk->get(F_CAT5);
	  const TString16 catses = _msk->get(F_CAT6);
	  if (catpri.not_empty())
	  	_catdon.add((const char*) catpri);
	  if (catsec.not_empty())
	  	_catdon.add((const char*) catsec);
	  if (catter.not_empty())
	  	_catdon.add((const char*) catter);
	  if (catqua.not_empty())
	  	_catdon.add((const char*) catqua);
	  if (catqui.not_empty())
	  	_catdon.add((const char*) catqui);
	  if (catses.not_empty())
	  	_catdon.add((const char*) catses);
		_catnondon.destroy();
	  const TString16 catnpri = _msk->get(F_CATN1);
	  const TString16 catnsec = _msk->get(F_CATN2);
	  const TString16 catnter = _msk->get(F_CATN3);
	  const TString16 catnqua = _msk->get(F_CATN4);
	  const TString16 catnqui = _msk->get(F_CATN5);
	  const TString16 catnses = _msk->get(F_CATN6);
	  if (catnpri.not_empty())
	  	_catnondon.add((const char*) catnpri);
	  if (catnsec.not_empty())
	  	_catnondon.add((const char*) catnsec);
	  if (catnter.not_empty())
	  	_catnondon.add((const char*) catnter);
	  if (catnqua.not_empty())
	  	_catnondon.add((const char*) catnqua);
	  if (catnqui.not_empty())
	  	_catnondon.add((const char*) catnqui);
	  if (catnses.not_empty())
	  	_catnondon.add((const char*) catnses);
		current_cursor()->set_filterfunction (filter_func_nondom,TRUE);
		((TSorted_cursor*)current_cursor())->change_order(chiave);
		reset_print();
		printer().footerlen(0);		
    crea_intestazione();
    return TRUE;
  }
  else
    return FALSE;
}

void TStampaNonDom::crea_intestazione()
{
  reset_header();
	TString sep(132);
	sep = "STAMPA SOGGETTI NON DOMICILIATI NEL COMUNE DELLA SEZIONE";
	sep.center_just();
	set_header(2, "@0g%s", (const char*) sep);
	TString16 data_stampa = _data_stampa.string();
	set_header(2,"@0g%10s", (const char*) data_stampa);
	sep = "";
	sep << "Pag. @#";
	set_header(2, "@120g%s", (const char*) sep);  
	sep = "";
	sep.fill('-');
	set_header(4, (const char *) sep);
	set_header(5,"@0gCodice@9gC.@12gCognome e nome@38gNato il@49gIndirizzo@100gTelefono abit.@116gTelefono lavoro");
	set_header(6,"@0gTessera@38gSe/So@49gCAP/Localit/Comune/Prov.@116gTelefono altro");
	set_header(7,"@0g--------@9g--@12g-------------------------@38g----------@49g--------------------------------------------------@100g---------------@116g---------------");
 	printer().footerlen(3);
}

bool TStampaNonDom::user_create()
{
  _rel = new TRelation(LF_SOGGETTI);
  _rel->add("LCP", "CODTAB==DOM_CODLOC",1,0,ALIAS_LCP);
  _rel->add(LF_COMUNI, "COM==DOM_CODCOM");
  _rel->add(LF_SEZIONI, "CODSEZ==CODSEZ|CODSOT==CODSOT");
  add_cursor(new TSorted_cursor(_rel,"","",3));
  _msk = new TMask("at7700a");
  return TRUE;
}

bool TStampaNonDom::user_destroy()
{
  delete _msk;
  delete _rel;
  return TRUE;
}

int at7700(int argc, char* argv[])
{
  TStampaNonDom a;
  a.run(argc, argv, "Soggetti non domiciliati nel comune di sez.");
  return 0;
}
