// cglib03.h

#ifndef __CGLIB03_H
#define __CGLIB03_H

#ifndef __TABUTIL_H
#include <tabutil.h>
#endif

// Defines per TInteressi_IVA_table. Servono per comporre i nomi dei campi

#define P_1_TRIMESTRE 0
#define P_2_TRIMESTRE 1
#define P_3_TRIMESTRE 2
#define P_4_TRIMESTRE 3
#define P_ANNUALE     4
#define I_PERIODICO   5
#define I_ANNUALE     6
#define I_ACCONTOIVA  7
#define I_INTEGRATIVO 8
#define I_CESSAZIONE  9
#define I_ART74       10
#define B_LIQ_DIFF    11
#define B_LIQ_NORM    12

class TInteressi_IVA_table : public TTable
{
public:
  int  read(int anno, int mese);
  //int  status() { return _ver->status();}
  real get(int what);
  TInteressi_IVA_table();
  virtual ~TInteressi_IVA_table();
};

struct TDociva : public TObject
{
  TString _descrdoc;
  real _totdociva;
  int _ndoc;

  TDociva(const char* _descrdociva, const real& _importo,const int _natdoc) 
    :_descrdoc(_descrdociva),_totdociva(_importo),_ndoc(_natdoc) { }
  virtual ~TDociva() { }
};

class TDociva_array : public TArray
{
public:
  bool add_riga(const char* _descrdociva,const real& _importo,const int _natdoc);
  TDociva& riga(int i) { return (TDociva&)(*this)[i]; }
};

struct TRigaiva : public TObject
{
  real     _imponibile, _imposta;    
  real     _imponibilep, _impostap;
  TString  _codiva;
  int      _tipodet, _tipocr, _tipoatt;
  bool     _intra;
  
  TRigaiva (const real& imponibile, const real& imposta, const real& imponibilep, const real& impostap, const char* codiva, const int tipodet, const int tipocr, bool intra, const int tipoatt):
  _imponibile(imponibile),_imposta(imposta),_imponibilep(imponibilep),_impostap(impostap),_codiva(codiva),_tipodet(tipodet),_tipocr(tipocr),_intra(intra),_tipoatt(tipoatt)
  {} 
};

// TRigaiva_array 
// Tabella per il calcolo degli imponibili Iva
class TRigaiva_array : public TArray
{
public:
  bool add_riga(const real& imponibile, const real& imposta, const real& imponibilep, const real& impostap, const char* codiva, const int tipodet, const int tipocr, bool intra, const int tipoatt = 0); 
  TRigaiva& riga(int i) { return (TRigaiva&)(*this)[i]; }
};

struct TTipodoc : public TObject
{
  TString _tipodoc, _descrdoc;
  real    _totdoc;
  
  TTipodoc (const char* tipodoc, const char* descrdoc, const real& totdoc):
  _tipodoc(tipodoc),_descrdoc(descrdoc),_totdoc(totdoc) {}
};

class TTipodoc_array : public TArray
{
public:
  bool add_riga(const char* tipodoc, const char* descrdoc, const real& totdoc);  
  TTipodoc& riga(int i) { return (TTipodoc&)(*this)[i]; }
};

struct TRiga : public TObject
{
  real     _imponibile, _imposta, _implordo;    //totali del periodo
  real     _imponibilep, _impostap, _implordop; //totali progressivi
  TString  _codiva;
  
  TRiga (const real& imponibile, const real& imposta, const real& implordo, const real& imponibilep, const real& impostap, const real& implordop, const char* codiva):
  _imponibile(imponibile),_imposta(imposta),_implordo(implordo),_codiva(codiva),_imponibilep(imponibilep),_impostap(impostap),_implordop(implordop) {} 
};

class TRiga_array : public TArray
{
public:
  bool add_riga(const real& imponibile, const real& imposta, const real& implordo, const real& imponibilep, const real& impostap, const real& implordop, const char* codiva); 
  TRiga& riga(int i) { return (TRiga&)(*this)[i]; }
};  

///////////////////////////////////////////////////////////
// Arrotondamenti IVA
///////////////////////////////////////////////////////////

enum TRound_mode { rm_unknown, rm_round, rm_ceil, rm_millelire };

class TIva_round : public TObject
{                
  static TRound_mode _def_mode;
  static int _def_decimals;
  TRound_mode _mode;
  int _decimals;

public: // Low level functions (rare usage)
  void set_default_mode(TRound_mode m, int d);   
  void set_mode(TRound_mode m, int d);

public: // High level functions (common usage) 
  virtual bool ok() const { return _mode != rm_unknown; }

  // Call this before the first call to round!
  void set_default_iva_mode(int year, bool declaration = FALSE, long firm = 0L); 
  
  void round(real& n) const;
  TIva_round();
};

#endif
