/* DEC	*StatMean(pDst,pSrc,n)
 *
 * ARGUMENT
 *	DEC	*pDst;
 *	DEC	**pSrc;
 *	int n;
 *
 * DESCRIPTION
 *	Sets pDst to the mean of the n DEC elements of the array pSrc.
 *   n <= 0 is not allowed.  pDst is unchanged if an error occurs.
 *
 * SIDE	EFFECTS
 *	None.
 *
 * RETURNS
 *	Returns	pointer	to the average if successful, otherwise
 *   returns GM_NULL.
 *
 * POSSIBLE ERROR CODES
 *
 *	GM_NULLPOINTER
 *	GM_ARGVAL
 *	GM_UNDERFLOW
 *
 * AUTHOR
 *	Jared Levy
 *   Copyright (C) 1987-1990 Greenleaf Software	Inc.  All rights reserved.
 *
 * MODIFICATIONS
 *
 */

#include <stdio.h>
#include "gm.h"
#include "gmsystem.h"

DEC	*StatMean(pDst,pSrc,n)
DEC	*pDst;
DEC	**pSrc;
int n;
{
	int i;
	DEC	*p, dsum, *sum=&dsum, ddn, *dn=&ddn;

/*  source must	be supplied !! */
	_MacStart(GM_DMEAN);

	if (!pSrc) {
		_MacErr(GM_NULLPOINTER);
		_MacRet(GM_NULL);
	}

	if (n<=0)  {
		_MacErr(GM_ARGVAL);
		_MacRet(GM_NULL);
	}

	for (i=0; i<n; i++)
		_MacInVarD(pSrc[i]);

	_MacOutVarD(pDst);

	p = pSrc[0];
	_MacDCopy(sum, p);


	for (i=1;i<n;i++)  {
		p = pSrc[i];
		(void) _AddDec80Bit(sum, sum, p);
		/* overflow impossible here since sum is 80 bits */
	}

	(void) ConvLongToDecimal(dn, (long) n);
	pDst = DivideDecimal(pDst, sum,	dn);

	_MacRet(pDst);
}
