#include <xvt.h>

#include <checks.h>
#include <keys.h>

#define buildmsg() char msg[256];va_list argptr;va_start(argptr,fmt);vsprintf(msg,fmt,argptr);va_end(argptr)

// @doc EXTERNAL

// @msg fatal_box | Crea una finestra di ERRORE FATALE con il relativo messaggio
bool fatal_box(
  const char* fmt, ...)  // @parm Messaggio da stampare nella finestra
  //       (il formato e' come nella printf del C)
  // @comm Il programma viene interrotto al momento in cui si e' verificato l'errore.
{
  buildmsg();
  xvt_dm_post_fatal_exit(msg);
  return false;
}

// @doc EXTERNAL

// @msg error_box | Crea una finestra di ERRORE con il relativo messaggio
bool error_box(
  const char* fmt,   // @parm Messaggio da stampare nella finestra
  ...)               // @parmvar Uno o piu' parametri corrispondenti ai codici in <p fmt>
  // @comm Se si opera in ambiente Windows crea la finestra con il bottone OK
  //       e l'icona punto esclamativo.
{
  buildmsg();                                                            
  xvt_dm_post_error(msg);
  return false;
}

// @doc EXTERNAL
// @msg warning_box | Crea una finestra di ATTENZIONE con il relativo messaggio
bool warning_box(
  const char* fmt,   // @parm Messaggio da stampare nella finestra
  ...)               // @parmvar Uno o piu' parametri corrispondenti ai codici in <p fmt>
  // @comm Se si opera in ambiente Windows crea la finestra con il bottone OK
  //       e l'icona punto di domanda.
{
  buildmsg();
  xvt_dm_post_warning(msg);
  return 0;
}

// @doc EXTERNAL
// @msg message_box | Crea una finestra di INFORMAZIONE con relativo il messaggio
bool message_box(
  const char* fmt,  // @parm Messaggio da stampare nella finestra
  ...)              // @parmvar Uno o piu' parametri corrispondenti ai codici in <p fmt>
  // @comm Se si opera in ambiente Windows crea la finestra con il bottone OK
  //       e l'icona informazioni.
{
  buildmsg();
  xvt_dm_post_message(msg);
  return false;
}

// @doc EXTERNAL
// @msg sorry_box | Crea una finestra di SCUSE con il relativo messaggio
bool sorry_box(
  const char* fmt, // @parm Messaggio da stampare nella finestra
  ...)             // @parmvar Uno o piu' parametri corrispondenti ai codici in <p fmt>
  // @comm Se si opera in ambiente Windows crea la finestra con il bottone OK
  //       e l'icona informazioni.
{
  buildmsg();
  xvt_dm_post_note(msg);
  return false;
}

// @msg yesno_box | Crea una finestra di RICHIESTA con il relativo messaggio
bool noyes_box(
  const char* fmt,   // @parm Messaggio da stampare nella finestra
  ...)               // @parmvar Uno o piu' parametri corrispondenti ai codici in <p fmt>
{
  buildmsg();
  ASK_RESPONSE r = xvt_dm_post_ask("No", "Si", NULL, msg);
  return r == RESP_DEFAULT;
}

// @doc EXTERNAL
// @msg yesno_box | Crea una finestra di RICHIESTA con il relativo messaggio
bool yesno_box(
  const char* fmt,   // @parm Messaggio da stampare nella finestra
  ...)                           // @parmvar Uno o piu' parametri corrispondenti ai codici in <p fmt>
  // @comm Se si opera in ambiente Windows crea la finestra con i bottni SI e NO
  //       e l'icona col punto di domanda.
  //
  // @rdesc Ritorna il risultato della richiesta:
  // 
  // @flag 1 | Se viene premuto il taso SI
  // @flag 0 | Se viene premuto il taso NO
{
  buildmsg();
  ASK_RESPONSE r = xvt_dm_post_ask("Si", "No", NULL, msg);
  return r == RESP_DEFAULT;
}

// @doc EXTERNAL
// @msg yesnofatal_box | Crea una finestra di RICHIESTA con il relativo
//                        messaggio, ma permette di terminare il programma
bool yesnofatal_box(
  const char* fmt,   // @parm Messaggio da stampare nella finestra
  ...)                           // @parmvar Uno o piu' parametri corrispondenti ai codici in <p fmt>

  // @comm Se si opera in ambiente Windows crea la finestra con i bottoni SI e NO
  //       e l'icona col punto di domanda.
  //       <nl>A differenza della <m yesno_box> permette di terminare il programma
  //             se viene premuto il tasto SI
  //
  // @rdesc Ritorna il risultato della richiesta:
  // 
  // @flag 1 | Se viene premuto il tasto SI
  // @flag 0 | Se viene premuto il tasto NO
  //
  // @xref <m yesno_box> <m yesnofatal_box>
{
  buildmsg();

#ifdef DBG  
  char s[256]; sprintf(s, "%s\nContinuare ugualmente?", msg);
  const bool ret = yesno_box("%s", s);
  if (!ret) 
#endif
  fatal_box(msg);

  return FALSE;
}

// @doc EXTERNAL
// @msg yesnocancel_box | Crea una finestra di RICHIESTA con il relativo messaggio
int yesnocancel_box(
  const char* fmt,   // @parm Messaggio da stampare nella finestra
  ...)               // @parmvar Uno o piu' parametri corrispondenti ai codici in <p fmt>
  // @comm Se si opera in ambiente Windows crea la finestra con i bottoni SI e NO
  //       e l'icona col punto di domanda.
  //
  // @rdesc Ritorna il risultato della richiesta:
  // 
  // @flag K_YES | Se viene premuto il tasto SI
  // @flag K_NO | Se viene premuto il tasto NO
  // @flag K_ESC | Se viene premuto il tosto ESC
  //
  // @xref <m yesno_box> <m yesnofatal_box>
{
  buildmsg();
  ASK_RESPONSE r = xvt_dm_post_ask("Si", "No", "Annulla", msg);
  return r == RESP_DEFAULT ? K_YES : (r == RESP_2 ? K_NO : K_ESC);
}

// @doc EXTERNAL
// @msg __trace | Permette di mandare dei messaggi nel file trace.log
bool __trace(
  const char* fmt,   // @parm Messaggio da stampare
  ...)               // @parmvar Uno o piu' parametri corrispondenti ai codici in <p fmt>
  // @comm Usato in fase di debug, permette di vedere se il programma ha
  //             fatto certe operazioni (deve essere definito il simbolo TRC in
  //             fase di compilazione).
  //
  // @rdesc Ritorna non 0 se riesca ad aprire il file trace.log
{
  static FILE* f = NULL;
  if (f == NULL)
    f = fopen("trace.log", "w");
  if (f != NULL)
  {
    buildmsg();
    fprintf(f, "%s\n", msg);
    fflush(f);
  }
  return f != NULL;  
}

bool __tracemem(const char* msg)
{
  unsigned long mem = xvt_sys_get_free_memory_kb();
  return __trace("%5lu Kb - %s", mem, msg);
}

