#include <applicat.h>
#include <utility.h>

#include "../ve/veconf.h"
#include "mglib.h"
#include "mg3100.h"
#include "mg3frm.h"  


class TForm_giacprezzi : public TForm_stampemg
{
  TConfig *conf_ditta;
  TCondizione_vendita * listino;

public:
  void set_codlistino(const char * codcm,const char * cod);
  bool validate(TForm_item &cf, TToken_string &s);
  TForm_giacprezzi(const char *name,const char *code);
  virtual ~TForm_giacprezzi();
};

bool TForm_giacprezzi::validate(TForm_item &cf, TToken_string &s)
{ 
  const TFixed_string code = s.get(0); // prende il primo parametro, il codice del messaggio
  if (code == "_USER") 
  {
    const TFixed_string subcode = s.get();
    if (subcode=="PREZZO") 
    {
      const TString80 codart = relation()->lfile(LF_ANAMAG).get("CODART");
      const bool found = listino->ricerca(codart, ZERO);
      cf.set(found ? listino->get_prezzo().string() : "");
      return TRUE;
    }
  }
  return TForm_stampemg::validate(cf, s);
}

void TForm_giacprezzi::set_codlistino(const char * codcatv,const char * codlist)
{
  listino->put_listino(codlist,codcatv);
}

TForm_giacprezzi::TForm_giacprezzi(const char *name,const char *code) :
  TForm_stampemg(name,code)
{
  listino=new TCondizione_vendita(new TConfig(CONFIG_DITTA,""));
  listino->set_anamag(relation()->lfile(LF_ANAMAG));
  listino->set_umart(relation()->lfile(LF_UMART));
}

TForm_giacprezzi::~TForm_giacprezzi()
{
  delete &(listino->config_ditta());
  delete listino;
}



class TStampadatist_mask: public TStampemg_mask
{
  static bool handle_ragg(TMask_field &fld, KEY k); // handler 
  static bool handle_liv(TMask_field &fld, KEY k); // handler 
  static bool handle_subord(TMask_field &fld, KEY k); // handler 
  static bool handle_filt(TMask_field &fld, KEY k); // handler 
public:
  TStampadatist_mask();
  virtual ~TStampadatist_mask(){};
};

TStampadatist_mask::TStampadatist_mask():
  TStampemg_mask("mg3100")
{
  TConfig prassid(CONFIG_DITTA, "ve"); // apre il file di configurazione della ditta corrente
  if (prassid.get_bool("GES", NULL, A_LISTINI))
    field(F_CATVENLISTINO).enable(prassid.get_bool("GESLISCV"));
  else {
    field(F_CATVENLISTINO).disable(); field(F_CODLISTINO).disable();
  }
  
  set_handler(F_TOLIVELLOART, handle_liv);
  set_handler(F_TOLIVELLOGIAC, handle_liv);
  set_handler(F_ORDINE, handle_ragg);
  set_handler(F_ORDINEART,handle_subord);
  set_handler(F_FILTRO,handle_filt);
  if (!livelli_articolo().enabled()) {
    TMask_field &f1=field(F_RAGGCODICE);
    f1.reset(); f1.check();f1.hide();
    field(F_FROMLIVELLOART).hide();
    field(F_TOLIVELLOART).hide();
  }
  if (!livelli_giacenza().enabled()) {
    TMask_field &f2=field(F_RAGGLIVGIAC);
    f2.reset(); f2.check();f2.hide();
    field(F_FROMLIVELLOGIAC).hide();
    field(F_TOLIVELLOGIAC).hide();
  }
}

bool TStampadatist_mask::handle_filt(TMask_field &fld, KEY k)
{
  if (k == K_SPACE) 
  {
    TMask& m = fld.mask();
    const bool articles = m.get(F_ORDINE)=="A";
    if (articles)
    {
      switch (fld.get()[0]) 
      {
        case '0':
        case 'P':
        case 'N':
        case 'S':
          m.show(F_ORDINEMAG);
          m.hide(F_ORDINEART);
          m.send_key(K_SPACE, F_ORDINEMAG);
        break;
        case 'T':
          m.show(F_ORDINEART);
          m.hide(F_ORDINEMAG);
          m.send_key(K_SPACE, F_ORDINEART);
        default:
        break;
      }
    }
  }
 return TRUE;
}

bool TStampadatist_mask::handle_subord(TMask_field &fld, KEY k)
{
  if (k == K_SPACE) 
  {
    TStampadatist_mask &m=(TStampadatist_mask &)fld.mask();
    
    TOperable_field &fraggcod=(TOperable_field &)m.field(F_RAGGCODICE);
    TOperable_field &fragggiac=(TOperable_field &)m.field(F_RAGGLIVGIAC);
    switch (fld.get()[0]) {
      case 'C':
        if (m.get(F_ORDINE)=="A") 
        {
          fraggcod.enable();m.send_key(K_SPACE,F_RAGGCODICE);
        } else {
          if (!m.get_bool(F_DETTAGLIODEP))
          {
            fraggcod.reset(); fraggcod.disable();
            fragggiac.reset(); fragggiac.disable();
            fraggcod.do_message(0);fragggiac.do_message(0);
          }
        }
      break;
      case 'M':
      case 'D':
        fraggcod.reset(); fraggcod.disable();
        fraggcod.do_message(0);
        fragggiac.reset(); fragggiac.disable();
        fragggiac.do_message(0);
        //m.field(F_FROMLIVELLOART).set(m.artlev().last_level());
        //m.disable(F_FROMLIVELLOART);
        //m.field(F_TOLIVELLOART).set("0");
        //m.disable(F_TOLIVELLOART);
        //m.field(F_FROMLIVELLOGIAC).set("0");
        //m.disable(F_FROMLIVELLOGIAC);
        //m.field(F_TOLIVELLOGIAC).set("0");
        //m.disable(F_TOLIVELLOGIAC);
      break;
    }
  }
  return TRUE;
}

bool TStampadatist_mask::handle_liv(TMask_field &fld, KEY k)
{
  if (k == K_TAB) 
  {
    TOperable_field &ftotmag=(TOperable_field &)fld.mask().field(F_TOTALIMAGAZZINI);
    TOperable_field &fdettmag=(TOperable_field &)fld.mask().field(F_DETTAGLIOMAG);
    TOperable_field &fdettgiac=(TOperable_field &)fld.mask().field(F_RAGGLIVGIAC);
    if (fld.mask().get_int(F_TOLIVELLOART))
    {
      fdettgiac.reset();
      fdettgiac.disable();
    } else
      fdettgiac.enable();
    if (fld.mask().get(F_ORDINE)=="A") 
    {
      if (fld.mask().get_int(F_TOLIVELLOART) || fld.mask().get_int(F_TOLIVELLOGIAC))
      {
        // raggruppamenti prima del dettaglio magazzini
        ftotmag.reset(); ftotmag.disable();ftotmag.do_message(0);
      } else {
        ftotmag.enable();
      }
    }
  }
  return TRUE;
}

bool TStampadatist_mask::handle_ragg(TMask_field &fld, KEY k)
{
  if (k == K_SPACE) 
  {
    TStampadatist_mask &m=(TStampadatist_mask &)fld.mask();

    TOperable_field &ftotmag=(TOperable_field &)fld.mask().field(F_TOTALIMAGAZZINI);
    TOperable_field &fraggcod=(TOperable_field &)fld.mask().field(F_RAGGCODICE);
    TOperable_field &fragggiac=(TOperable_field &)fld.mask().field(F_RAGGLIVGIAC);

    fraggcod.enable();
    fragggiac.enable();
    if (fld.mask().get(F_ORDINE)=="M")
    {
      ftotmag.enable();
      if (!fld.mask().get_bool(F_DETTAGLIODEP))
      {
        fraggcod.reset(); fraggcod.disable();
        fragggiac.reset(); fragggiac.disable();
        fraggcod.do_message(0);fragggiac.do_message(0);
        return TRUE;
      }
    } else {
      bool enabmag=m.magazz_ini().gestmultimag() || m.magazz_ini().gestdep();
      m.enable(F_TOTALIMAGAZZINI,enabmag && m.field(F_TOTALIMAGAZZINI).enabled());
      m.enable(F_DETTAGLIOMAG,enabmag && m.field(F_DETTAGLIOMAG).enabled());
    }
  }
  return TRUE;
}


// STAMPA anagrafiche
class TStampa_datistorici : public TSkeleton_application
{
  TStampadatist_mask * _mask;
  TCursor * _cur;
  TForm_giacprezzi * _form; // to be moved into TPrint_application

protected:
  virtual bool create();
  virtual bool destroy();
  virtual void main_loop();
  virtual void on_firm_change();

  void setprint_permagazzini();
  void setprint_perarticoli();
  void setprint_perarticoli_all();

public:
  TStampa_datistorici() {}
};

void TStampa_datistorici ::on_firm_change()
{
  _mask->enable_livellicodice();
}

bool TStampa_datistorici::create()
{
  open_files(LF_ANAMAG,LF_MAG,LF_TAB,LF_TABCOM,LF_UMART, 0);
  _mask = new TStampadatist_mask();
  return TSkeleton_application::create();
}

bool TStampa_datistorici::destroy()
{
  delete _mask;
  return TSkeleton_application::destroy();
}

void TStampa_datistorici::setprint_permagazzini()
{
  TRectype darec(LF_MAG),arec(LF_MAG);
  TString cfilter,sortexp,joinexp;
  _form = new TForm_giacprezzi("mg3100b", "");
  _cur = _form->cursor();
  // ********************
  // setta il filtro sul cursore
  TForm_item & it_filtro=_form->find_field('H',odd_page,FF_FILTRO);
  TForm_item & rangea=_form->find_field('H',odd_page,FF_RANGEARTIC);
  TString srangea;
  if (*_mask->get(F_DAART))
  {
    cfilter << "(CODART[1," << _mask->get(F_DAART).len() << "]>=" <<'"' << _mask->get(F_DAART)<< "\")&&" ;
    srangea << "da \"" << _mask->get(F_DAART) << "\" ";
  }
  if (*_mask->get(F_AART))
  {
    cfilter << "(CODART[1," << _mask->get(F_AART).len() << "]<=" <<'"' << _mask->get(F_AART)<< "\")&&" ;
    srangea << "fino a " << _mask->get(F_AART)  << "\" ";
  }
  rangea.set(srangea.empty() ? "Tutti gli articoli" : format("Articoli  %s", (const char *)srangea));
  switch (*_mask->get(F_FILTRO)) {
    case '0': // non nulli
      cfilter << "(STR(" << LF_MAG << "->GIAC!=\"0\"))&&";
      it_filtro.set("non nulle");
    break;
    case 'P': // positivi
      cfilter << "(STR(" << LF_MAG << "->GIAC>\"0\"))&&";
      it_filtro.set("positive");
    break;
    case 'N': // negativi
      cfilter << "(STR(" << LF_MAG << "->GIAC<\"0\"))&&";
      it_filtro.set("negative");
    break;
    case 'S': // sottoscorta
      cfilter << "(STR(GIAC<SCORTAMIN))&&";
      it_filtro.set("sottoscorta");
    break;
  }
  if (cfilter.not_empty())
    cfilter.cut(cfilter.len()-2); 

  // abilita/disabilita le sezioni dei totali
  _form->setdett_permag(
    _mask->get_bool(F_RAGGCODICE),
    _mask->get_int(F_FROMLIVELLOART),
    _mask->get_int(F_TOLIVELLOART),
    _mask->get_bool(F_RAGGLIVGIAC),
    _mask->get_int(F_FROMLIVELLOGIAC),
    _mask->get_int(F_TOLIVELLOGIAC),
    TRUE,
    _mask->get_bool(F_TOTALIDEPOSITI),
    _mask->get_bool(F_DETTAGLIODEP));
  
  if (!_mask->get_bool(F_TOTALIDEPOSITI))
  {
    ((TSorted_cursor *)_cur)->change_order("ANNOES|CODMAG[1,3]|CODART|LIVELLO");
    _form->find_field('B',odd_page,"GRUPPI_DEPOSITO").setcondition("CODMAG[1,3]",_strexpr);
  }
  TString16 ena=((_mask->get_bool(F_STAMPAGIAC) && *_mask->get(F_FILTRO)!='S')? "1": "0");
  _form->find_field('H',odd_page,FF_FLAGGIAC).set(ena );
  _form->find_field('B',odd_page,FF_FLAGGIAC).set(ena);
  ena=((_mask->get_bool(F_STAMPAGIAC) && *_mask->get(F_FILTRO)=='S')? "1": "0") ;
  _form->find_field('H',odd_page,FF_FLAGSOTTOS).set(ena);
  _form->find_field('B',odd_page,FF_FLAGSOTTOS).set(ena);
  ena=_mask->get(F_PREZZIVALORI);
  _form->find_field('H',odd_page,FF_PREZZIVALORI).set(ena);
  _form->find_field('B',odd_page,FF_PREZZIVALORI).set(ena);
  ena=_mask->get_bool(F_VALMEDI) ? "1": "0";
  _form->find_field('H',odd_page,FF_FLAGVALMEDI).set(ena);
  _form->find_field('B',odd_page,FF_FLAGVALMEDI).set(ena);
  _form->set_codlistino(_mask->get(F_CATVENLISTINO),_mask->get(F_CODLISTINO));

  // ********************
  // range di stampa magazzini
  TForm_item & rangem=_form->find_field('H',odd_page,FF_RANGEMAGAZ);
  TString srangem;
  if (!_mask->get(F_DAMAG).blank())
  {
    darec.put("CODMAG",_mask->get(F_DAMAG));
    srangem << "da \"" << _mask->get(F_DAMAG) << "\" ";
  }
  if (!_mask->get(F_AMAG).blank())
  {
    arec.put("CODMAG",_mask->get(F_AMAG));
    srangem << "fino a \"" <<  _mask->get(F_AMAG) << "\" ";
  }
  rangem.set(srangem.empty() ? "Tutti i magazzini" : format("Magazzini %s", (const char *)srangem));
  darec.put("ANNOES",_mask->get(F_ANNOES));
  arec.put("ANNOES",_mask->get(F_ANNOES));
  _cur->setfilter(cfilter);
  _cur->setregion(darec,arec);
}

void TStampa_datistorici::setprint_perarticoli()
{
  char subordine=*_mask->get(F_ORDINEMAG);
  TRectype darec(LF_MAG),arec(LF_MAG);
  TString filter;
  char filtro_articoli=*_mask->get(F_FILTRO);
  _form = new TForm_giacprezzi("mg3100a", "");

  // ********************
  // range di stampa magazzini
  TString srangem;
  TForm_item & rangem=_form->find_field('H',odd_page,FF_RANGEMAGAZ);
  _cur = _form->cursor();
  if (*_mask->get(F_DAMAG))
    srangem << "da \"" << _mask->get(F_DAMAG) << "\" ";
  if (*_mask->get(F_AMAG))
    srangem << "fino a \"" << _mask->get(F_AMAG) << "\"";

  rangem.set(srangem.blank()? "Tutti i magazzini" : format("Magazzini %s",(const char *)srangem));

  // ********************
  // filtro giacenze
  TForm_item & it_filtro=_form->find_field('H',odd_page,FF_FILTRO);
  switch (filtro_articoli) {
    case '0': // non nulli
      filter << "(STR(" << LF_MAG << "->GIAC!=\"0\"))&&";
      it_filtro.set("non nulle");
    break;
    case 'P': // positivi
      filter << "(STR(" << LF_MAG << "->GIAC>\"0\"))&&";
      it_filtro.set("positive");
    break;
    case 'N': // negativi
      filter << "(STR(" << LF_MAG << "->GIAC<\"0\"))&&";
      it_filtro.set("negative");
    break;
    case 'S': // sottoscorta
      filter << "(STR("<< LF_MAG << "->GIAC < " <<LF_MAG << "->SCORTAMIN))&&";
      it_filtro.set("sottoscorta");
    break;
  }
  if (_mask->get(F_DAART).not_empty())
    filter << "(CODART>=\"" << _mask->get(F_DAART) << "\")&&";
  if (_mask->get(F_AART).not_empty())
    filter << "(CODART<=\"" << _mask->get(F_AART) << "\")&&";
  
  if (filter.not_empty())
    filter.rtrim(2); 
  
  // ******************
  // sotto-ordinamenti e range di stampa articoli 
  TString srangea,srangecm,espr;
  TForm_item & rangecm=_form->find_field('H',odd_page,FF_RANGECATMER);
  TForm_item & rangea=_form->find_field('H',odd_page,FF_RANGEARTIC);
  if (!_mask->get(F_DAART).blank())
    srangea << "da \"" << _mask->get(F_DAART) << "\" ";
  if (!_mask->get(F_AART).blank())
    srangea << "fino a \"" << _mask->get(F_AART) << "\"";
  rangea.set(srangea.empty() ? "Tutti gli articoli" : format("Articoli  %s", (const char *)srangea));
  switch (subordine)
  {
    case 'C':
    case 'D':
      _form->find_field('B',odd_page,"H_CATMER").hide();
      _form->find_field('B',odd_page,"TOT_CATMER").hide();
      _form->set_ordering(mg_normale);
    break;
    case 'M':
    case 'S':
      _form->find_field('B',odd_page,"H_CATMER").show();
      _form->find_field('B',odd_page,"TOT_CATMER").show();
      _form->set_ordering(subordine == 'M' ? mg_cat_merc : mg_scat_merc);
      
      espr.format("%d->%s", LF_ANAMAG, subordine == 'M' ? "GRMERC[1,3]" : "GRMERC");
      _form->find_field('B',odd_page,"GRUPPI_CATMER").setcondition(espr,_strexpr); 
      
      if (!_mask->get(F_DASCATMER).blank())
      {
        srangecm << "da \"" << _mask->get(F_DASCATMER) << "\" ";
        filter << "&&(" << espr << ">=\"" << _mask->get(F_DASCATMER) << "\")";
      }
      if (!_mask->get(F_ASCATMER).blank())
      {
        srangecm << "fino a \"" << _mask->get(F_ASCATMER) << "\"";
        filter << "&&(" << espr << "<=\"" << _mask->get(F_ASCATMER) << "\")";
      }
      rangecm.set(srangecm.empty() ? "Tutte le categorie merceologiche" : format("Categorie merc. %s", (const char *)srangecm));
    break;
  }
  
  _form->setdett_perart(
    _mask->get_bool(F_RAGGCODICE),
    _mask->get_int(F_FROMLIVELLOART),
    _mask->get_int(F_TOLIVELLOART),
    _mask->get_bool(F_RAGGLIVGIAC),
    _mask->get_int(F_FROMLIVELLOGIAC),
    _mask->get_int(F_TOLIVELLOGIAC),
    _mask->get_bool(F_TOTALIMAGAZZINI),
    _mask->get_bool(F_DETTAGLIOMAG));

  TString16 ena=((_mask->get_bool(F_STAMPAGIAC) && *_mask->get(F_FILTRO)!='S')? "1": "0");
  _form->find_field('H',odd_page,FF_FLAGGIAC).set(ena );
  _form->find_field('B',odd_page,FF_FLAGGIAC).set(ena);
  ena=((_mask->get_bool(F_STAMPAGIAC) && *_mask->get(F_FILTRO)=='S')? "1": "0") ;
  _form->find_field('H',odd_page,FF_FLAGSOTTOS).set(ena);
  _form->find_field('B',odd_page,FF_FLAGSOTTOS).set(ena);
  _form->find_field('H',odd_page,FF_PREZZIVALORI).set(_mask->get(F_PREZZIVALORI));
  _form->find_field('B',odd_page,FF_PREZZIVALORI).set(_mask->get(F_PREZZIVALORI));

  ena=_mask->get_bool(F_VALMEDI) ? "1": "0";
  _form->find_field('H',odd_page,FF_FLAGVALMEDI).set(ena);
  _form->find_field('B',odd_page,FF_FLAGVALMEDI).set(ena);

  _form->set_codlistino(_mask->get(F_CATVENLISTINO),_mask->get(F_CODLISTINO));
  
  darec.put("ANNOES",_mask->get(F_ANNOES));
  darec.put("CODMAG",_mask->get(F_DAMAG));
  arec.put("ANNOES",_mask->get(F_ANNOES));
  arec.put("CODMAG",_mask->get(F_AMAG));

  const bool order_by_group = subordine == 'M' || subordine == 'S';

  _cur->setregion(darec,arec);
  _cur->setfilter(filter, order_by_group);

  if (order_by_group)
  {
    TString so;
    so.format("ANNOES|%d->GRMERC|CODART|LIVELLO|CODMAG", LF_ANAMAG);
    ((TSorted_cursor*) _cur)->change_order(so);
  }
}  

// ORDINAMENTO 
void TStampa_datistorici::setprint_perarticoli_all()
{
  TRectype darec(LF_ANAMAG),arec(LF_ANAMAG);
  TString cfilter,filter,joinexp;
  TString sortexp("ANNOES|CODART");
  if (_mask->get_bool(F_RAGGCODICE))
  {
    if (_mask->get_int(F_TOLIVELLOART))
      sortexp << "[" << livelli_articolo().packed_length(_mask->get_int(F_TOLIVELLOART)) << "]";
  }
  if (_mask->get_bool(F_RAGGLIVGIAC))
  {
    sortexp << "|LIVELLO" ;
    if (_mask->get_int(F_TOLIVELLOGIAC))
     sortexp << "[1," << livelli_articolo().packed_length(_mask->get_int(F_TOLIVELLOGIAC)) << "]";
  }
  sortexp <<   "|CODMAG";
  char filtro_articoli=*_mask->get(F_FILTRO);
  char subordine=*_mask->get(F_ORDINEART);

  _form = new TForm_giacprezzi("mg3100aa", "");
  // ********************
  // range di stampa magazzini
  TForm_item & rangem=_form->find_field('H',odd_page,FF_RANGEMAGAZ);
  TString srangem;
  _cur = _form->cursor();
  if (*_mask->get(F_DAMAG))
  {
    filter << "(" << LF_MAG <<"->CODMAG[1,3]>=" <<'"' << _mask->get(F_DAMAG)<< "\")&&" ;
    srangem << "da \"" << _mask->get(F_DAMAG) << "\" ";
  }
  if (*_mask->get(F_AMAG))
  {
    filter << "(" << LF_MAG <<"->CODMAG[1,3]<=" <<'"' << _mask->get(F_AMAG)<< "\")&&" ;
    srangem << "fino a \"" << _mask->get(F_AMAG) << "\"";
  }
  rangem.set(srangem.blank()? "Tutti i magazzini" : format("Magazzini %s",(const char *)srangem));
  // ********************
  // filtro giacenze
  TForm_item & it_filtro=_form->find_field('H',odd_page,FF_FILTRO);
  switch (filtro_articoli) {
    case '0': // non nulli
      filter << "(STR(" << LF_MAG << "->GIAC!=\"0\"))&&";
      it_filtro.set("non nulle");
    break;
    case 'P': // positivi
      filter << "(STR(" << LF_MAG << "->GIAC>\"0\"))&&";
      it_filtro.set("positive");
    break;
    case 'N': // negativi
      filter << "(STR(" << LF_MAG << "->GIAC<\"0\"))&&";
      it_filtro.set("negative");
    break;
    case 'S': // sottoscorta
      filter << "(STR("<< LF_MAG << "->GIAC < " <<LF_MAG << "->SCORTAMIN))&&";
      it_filtro.set("sottoscorta");
    break;
  }
  if (filter.not_empty())
    filter.cut(filter.len()-2); 

  // ******************
  // sotto-ordinamenti e range di stampa articoli 
  TForm_item & rangecm=_form->find_field('H',odd_page,FF_RANGECATMER);
  TForm_item & rangea=_form->find_field('H',odd_page,FF_RANGEARTIC);
  TString srangea,srangecm;
  switch (subordine)
  {
    case 'C':
      _cur->setkey(1);
      darec.put("CODART",_mask->get(F_DAART));
      arec.put("CODART",_mask->get(F_AART));
      _form->find_field('B',odd_page,"H_CATMER").hide();
      _form->find_field('B',odd_page,"TOT_CATMER").hide();
      _form->set_ordering(mg_normale);
      if (!_mask->get(F_DAART).blank())
        srangea << "da \"" << _mask->get(F_DAART) << "\" ";
      if (!_mask->get(F_AART).blank())
        srangea << "fino a \"" << _mask->get(F_AART) << "\"";
    break;
    case 'D':
      _cur->setkey(2);
      darec.put("DESCR",_mask->get(F_DADES));
      arec.put("DESCR",_mask->get(F_ADES));
      _form->find_field('B',odd_page,"H_CATMER").disable();
      _form->find_field('B',odd_page,"TOT_CATMER").hide();
      _form->set_ordering(mg_normale);
      if (!_mask->get(F_DADES).blank())
        srangea << "da \"" << _mask->get(F_DADES) << "\" ";
      if (!_mask->get(F_ADES).blank())
        srangea << "fino a \"" << _mask->get(F_ADES) << "\"";
    break;
    case 'M':
    case 'S':
      _cur->setkey(3);
      darec.put("GRMERC",_mask->get(F_DACATMER));
      arec.put("GRMERC",_mask->get(F_ACATMER));
      _form->find_field('B',odd_page,"H_CATMER").show();
      _form->find_field('B',odd_page,"TOT_CATMER").show();
      _form->set_ordering(subordine == 'M' ? mg_cat_merc : mg_scat_merc);

      if (!_mask->get(F_DAART).blank())
      {
        srangea << "da \"" << _mask->get(F_DAART) << "\" ";
        cfilter << "(CODART[1," << _mask->get(F_DAART).len() << "]>=" <<'"' << _mask->get(F_AART)<< "\")&&" ;
      }
      if (!_mask->get(F_AART).blank())
      {
        srangea << "fino a \"" << _mask->get(F_AART) << "\"";
        cfilter << "(CODART[1," << _mask->get(F_AART).len() << "]<=" <<'"' << _mask->get(F_AART)<< "\")&&" ;
      }
      if (!_mask->get(F_DACATMER).blank())
        srangecm << "da \"" << _mask->get(F_DACATMER) << "\" ";
      if (!_mask->get(F_ACATMER).blank())
        srangecm << "fino a \"" << _mask->get(F_ACATMER) << "\"";
      rangecm.set(srangecm.empty() ? "Tutte le categorie merceologiche" : format("Categorie merc. %s", (const char *)srangecm));
    break;
  }
  rangea.set(srangea.empty() ? "Tutti gli articoli" : format("Articoli  %s", (const char *)srangea));

  _form->setdett_perart(
    _mask->get_bool(F_RAGGCODICE),
    _mask->get_int(F_FROMLIVELLOART),
    _mask->get_int(F_TOLIVELLOART),
    _mask->get_bool(F_RAGGLIVGIAC),
    _mask->get_int(F_FROMLIVELLOGIAC),
    _mask->get_int(F_TOLIVELLOGIAC),
    _mask->get_bool(F_TOTALIMAGAZZINI),
    _mask->get_bool(F_DETTAGLIOMAG));

  TString16 ena=((_mask->get_bool(F_STAMPAGIAC) && *_mask->get(F_FILTRO)!='S')? "1": "0");
  _form->find_field('H',odd_page,FF_FLAGGIAC).set(ena );
  _form->find_field('B',odd_page,FF_FLAGGIAC).set(ena);
  ena=((_mask->get_bool(F_STAMPAGIAC) && *_mask->get(F_FILTRO)=='S')? "1": "0") ;
  _form->find_field('H',odd_page,FF_FLAGSOTTOS).set(ena);
  _form->find_field('B',odd_page,FF_FLAGSOTTOS).set(ena);
  _form->find_field('H',odd_page,FF_PREZZIVALORI).set(_mask->get(F_PREZZIVALORI));
  _form->find_field('B',odd_page,FF_PREZZIVALORI).set(_mask->get(F_PREZZIVALORI));

  ena=_mask->get_bool(F_VALMEDI) ? "1": "0";
  _form->find_field('H',odd_page,FF_FLAGVALMEDI).set(ena);
  _form->find_field('B',odd_page,FF_FLAGVALMEDI).set(ena);

  _form->set_codlistino(_mask->get(F_CATVENLISTINO),_mask->get(F_CODLISTINO));

  joinexp << "ANNOES==\"" << _mask->get(F_ANNOES) << "\"|CODART==CODART";

  TSortedfile *mag;
  // !?!?!! ATTENZIONE : Modifica temporanea
  mag= new TSortedfile(LF_MAG,NULL,sortexp,"",1);
  // il filtro viene qui settato DOPO la creazione del Sortedfile a causa di una bug
  // sulla libreria nei TSorted_file / TCursor
  mag->cursor().setfilter(filter);
  // FINE MODIFICA ; rimettere il filtro nel costruttore quando il bug sar risolto 
  _cur->relation()->replace(mag,1,joinexp);
  if (cfilter.not_empty())
  {
    cfilter.cut(cfilter.len()-2); 
    _cur->setfilter(cfilter);
  }
  _cur->setregion(darec,arec);
}


void TStampa_datistorici::main_loop()
{
  while (_mask->run() == K_ENTER) 
  {
    if (_mask->magazz_ini().gestmag(TRUE))
    {
      if (_mask->get(F_ORDINE)[0]=='A')  
      {
        if (_mask->get(F_FILTRO)[0] == 'T')
          setprint_perarticoli_all();
        else
          setprint_perarticoli();
        
        const bool sp = _mask->field(F_SALTOPAG).active()&&_mask->get_bool(F_SALTOPAG);  
        TForm_subsection& ss = _form->find_field('B', odd_page, "H_CATMER");
        ss.set_newpage(sp);
      } 
      else 
      {                      
        setprint_permagazzini();
      }   

      _form->print();
      delete _form;
    }
  } // while true  
}


int mg3100(int argc, char* argv[])
{
  TStampa_datistorici a;// derivata da Application e con uso di form
  a.run(argc,argv,"Stampa dati storici di giacenza");
  return 0;
}
