#include <applicat.h>
#include <utility.h>

#include "mglib.h"
#include "mg3200.h"
#include "mg3frm.h"

class TStampadisp_mask: public TStampemg_mask
{
  static bool handle_ragg(TMask_field &fld, KEY k); // handler 
  static bool handle_liv(TMask_field &fld, KEY k); // handler 
  static bool handle_subord(TMask_field &fld, KEY k); // handler 

public:
  TStampadisp_mask();
  virtual ~TStampadisp_mask(){};
};
TStampadisp_mask::TStampadisp_mask():
  TStampemg_mask("mg3200")
{
  TCodgiac_livelli giaclev;
  TCodart_livelli  artlev;
  
  set_handler(F_TOLIVELLOART, handle_liv);
  set_handler(F_TOLIVELLOGIAC, handle_liv);
  set_handler(F_ORDINE, handle_ragg);
  set_handler(F_DETTAGLIODEP, handle_ragg);
  set_handler(F_DETTAGLIOMAG, handle_ragg);
  set_handler(F_ORDINEART,handle_subord);
  if (!artlev.enabled()) {
    TMask_field &f1=field(F_RAGGCODICE);
    f1.reset(); f1.check();f1.hide();
    field(F_FROMLIVELLOART).hide();
    field(F_TOLIVELLOART).hide();
  }
  if (!giaclev.enabled()) {
    TMask_field &f2=field(F_RAGGLIVGIAC);
    f2.reset(); f2.check();f2.hide();
    field(F_FROMLIVELLOGIAC).hide();
    field(F_TOLIVELLOGIAC).hide();
  }
}

bool TStampadisp_mask::handle_subord(TMask_field &fld, KEY k)
{
  if (k == K_SPACE) 
  {
  TOperable_field &fraggcod=(TOperable_field &)fld.mask().field(F_RAGGCODICE);
  TOperable_field &fragggiac=(TOperable_field &)fld.mask().field(F_RAGGLIVGIAC);
  switch (fld.get()[0]) {
    case 'C':
      if (fld.mask().get(F_ORDINE)=="A") 
        fraggcod.enable();
      else {
        if (!fld.mask().get_bool(F_DETTAGLIODEP))
        {
          fraggcod.reset(); fraggcod.disable();
          fragggiac.reset(); fragggiac.disable();
          fraggcod.do_message(0);fragggiac.do_message(0);
        }
      }
    break;
    case 'M':
    case 'D':
      fraggcod.reset(); fraggcod.check();fraggcod.disable();
    break;
    }
  }
  return TRUE;
}

bool TStampadisp_mask::handle_liv(TMask_field &fld, KEY k)
{
  if (k == K_TAB) 
  {
    TOperable_field &ftotmag=(TOperable_field &)fld.mask().field(F_TOTALIMAGAZZINI);
    TOperable_field &fdettmag=(TOperable_field &)fld.mask().field(F_DETTAGLIOMAG);
    TOperable_field &fdettgiac=(TOperable_field &)fld.mask().field(F_RAGGLIVGIAC);
    if (fld.mask().get(F_ORDINE)=="A") 
    {
      if (fld.mask().get_int(F_TOLIVELLOART))
      {
        fdettgiac.reset();
        fdettgiac.disable();
      } else
        fdettgiac.enable();
      if (fld.mask().get_int(F_TOLIVELLOART) || fld.mask().get_int(F_TOLIVELLOGIAC))
      {
        // raggruppamenti prima del dettaglio magazzini
        ftotmag.reset(); ftotmag.disable();ftotmag.do_message(0);
      } else {
        ftotmag.enable();
      }
    }
  }
  return TRUE;
}

bool TStampadisp_mask::handle_ragg(TMask_field &fld, KEY k)
{
  if (k == K_SPACE) 
  {
    TStampadisp_mask & m=(TStampadisp_mask &)fld.mask();

    TOperable_field &ftotmag=(TOperable_field &)m.field(F_TOTALIMAGAZZINI);
    TOperable_field &fraggcod=(TOperable_field &)m.field(F_RAGGCODICE);
    TOperable_field &fragggiac=(TOperable_field &)m.field(F_RAGGLIVGIAC);
    if (m.get(F_ORDINE)=="M")
    {
      ftotmag.enable();
      if (!m.get_bool(F_DETTAGLIODEP))
      {
        fraggcod.reset(); fraggcod.disable();
        fragggiac.reset(); fragggiac.disable();
        fraggcod.do_message(0);fragggiac.do_message(0);
        return TRUE;
      }
    } else {
      bool enabmag=m.magazz_ini().gestmultimag() || m.magazz_ini().gestdep();
      m.enable(F_TOTALIMAGAZZINI,enabmag);
      m.enable(F_DETTAGLIOMAG,enabmag);
    } 
    fraggcod.enable();
    fragggiac.enable();
  }
  return TRUE;
}


// mg3200       Stampa 
class TStampa_disp : public TSkeleton_application
{
  TArray * _files;
  TStampadisp_mask * _mask;
  TCursor * _cur;
  TForm_stampemg * _form; // to be moved into TPrint_application

protected:
  virtual bool print_one(int file)  {return TRUE;}
  virtual bool create();
  virtual bool destroy();
  virtual void main_loop();
  virtual void on_firm_change();

  void setprint_permagazzini();
  void setprint_perarticoli();

public:
  TStampa_disp() {}
};


bool TStampa_disp::create()
{
  _mask = new TStampadisp_mask;
  _files = new TArray();
  _files->add(new TLocalisamfile(LF_MAG));
  return TSkeleton_application::create();
}

bool TStampa_disp::destroy()
{
  delete _mask;
  delete _files;
  return TSkeleton_application::destroy();
}

void TStampa_disp::on_firm_change()
{
  _mask->enable_livellicodice();
}


void TStampa_disp::setprint_permagazzini()
{
  char subordine=*_mask->get(F_ORDINEART);
  TRectype darec(LF_MAG),arec(LF_MAG);
  TString cfilter,sortexp,joinexp;
  _form = new TForm_stampemg("mg3200b", "");
  _cur  = _form->cursor();
  TSorted_cursor & cur= (TSorted_cursor & )*_cur;
  // ********************
  // range di stampa articoli
  // setta il filtro sul cursore
  TForm_item & rangea=_form->find_field('H',odd_page,FF_RANGEARTIC);
  TString srangea;
  
  if (*_mask->get(F_DAART))
  {
    cfilter << "(CODART[1," << _mask->get(F_DAART).len() << "]>=" <<'"' << _mask->get(F_DAART)<< "\")&&" ;
    srangea << "da \"" << _mask->get(F_DAART) << "\" ";
  }
  if (*_mask->get(F_AART))
  {
    cfilter << "(CODART[1," << _mask->get(F_AART).len() << "]<=" <<'"' << _mask->get(F_AART)<< "\")&&" ;
    srangea << "fino a \"" << _mask->get(F_AART) << "\"";
  }
  if (cfilter.not_empty())
    cfilter.cut(cfilter.len()-2); 

  rangea.set(srangea.empty() ? "Tutti gli articoli" : format("Articoli  %s",(const char *)srangea));
  // ***************
  // gestione dei livelli di raggruppamento e di detaglio
  // (abilita/disabilita le sezioni dei totali/header)
  
  _form->setdett_permag(
    _mask->get_bool(F_RAGGCODICE),
    _mask->get_int(F_FROMLIVELLOART),
    _mask->get_int(F_TOLIVELLOART),
    _mask->get_bool(F_RAGGLIVGIAC),
    _mask->get_int(F_FROMLIVELLOGIAC),
    _mask->get_int(F_TOLIVELLOGIAC),
    TRUE,
    _mask->get_bool(F_TOTALIDEPOSITI),
    _mask->get_bool(F_DETTAGLIODEP));

  if (!_mask->get_bool(F_TOTALIDEPOSITI))
  {
    cur.change_order("ANNOES|CODMAG[1,3]|CODART|LIVELLO");
    _form->find_field('B',odd_page,"GRUPPI_DEPOSITO").setcondition("CODMAG[1,3]",_strexpr);
  }
  // ********************
  // range di stampa magazzini
  TForm_item & rangem=_form->find_field('H',odd_page,FF_RANGEMAGAZ);
  TString srangem;

  darec.put("ANNOES",_mask->get(F_ANNOES));
  arec.put("ANNOES",_mask->get(F_ANNOES));
  darec.put("CODMAG",_mask->get(F_DAMAG));
  arec.put("CODMAG",_mask->get(F_AMAG));
  if (*_mask->get(F_DAMAG))
    srangem << "da \"" <<  _mask->get(F_DAMAG) << "\" ";
  if (*_mask->get(F_AMAG))
    srangem << "fino a \"" <<  _mask->get(F_AMAG) << "\" ";
  rangem.set(srangem.empty() ? "Tutti i magazzini" : format("Magazzini %s",(const char *)srangem));

  cur.setfilter(cfilter);
  cur.setregion(darec,arec);
}

// ORDINAMENTO 
void TStampa_disp::setprint_perarticoli()
{
  TRectype darec(LF_ANAMAG),arec(LF_ANAMAG);
  TString cfilter,filter,joinexp;
  TString sortexp("ANNOES|CODART");
  if (_mask->get_bool(F_RAGGCODICE))
  {
    if (_mask->get_int(F_TOLIVELLOART))
      sortexp << "[" << livelli_articolo().packed_length(_mask->get_int(F_TOLIVELLOART)) << "]";
  }
  if (_mask->get_bool(F_RAGGLIVGIAC))
  {
    sortexp << "|LIVELLO" ;
    if (_mask->get_int(F_TOLIVELLOGIAC))
     sortexp << "[1," << livelli_articolo().packed_length(_mask->get_int(F_TOLIVELLOGIAC)) << "]";
  }
  sortexp <<   "|CODMAG";
  char subordine=*_mask->get(F_ORDINEART);
  _form = new TForm_stampemg("mg3200a", "");
  _cur = _form->cursor();
  // ********************
  // range di stampa magazzini
  TForm_item & rangem=_form->find_field('H',odd_page,FF_RANGEMAGAZ);
  TString srangem;

  if (*_mask->get(F_DAMAG))
  {
    filter << "(" << LF_MAG<< "->CODMAG[1,3]>=" <<'"' << _mask->get(F_DAMAG)<< "\")&&" ;
    srangem << "da \"" <<  _mask->get(F_DAMAG) << "\" ";
  }
  if (*_mask->get(F_AMAG))
  {
    filter << "(" << LF_MAG<< "->CODMAG[1,3]<=" <<'"' << _mask->get(F_AMAG)<< "\")&&" ;
    srangem << "fino a \"" <<  _mask->get(F_AMAG) << "\" ";
  }
  rangem.set(srangem.empty() ? "Tutti i magazzini" : format("Magazzini %s",(const char *)srangem));
  if (filter.not_empty())
    filter.cut(filter.len()-2); 
  // ********************
  // range di stampa articoli 
  TForm_item & rangecm=_form->find_field('H',odd_page,FF_RANGECATMER);
  TForm_item & rangea=_form->find_field('H',odd_page,FF_RANGEARTIC);
  TString srangea,srangecm;
  switch (subordine)
  {
    case 'C':
      _cur->setkey(1);
      darec.put("CODART",_mask->get(F_DAART));
      arec.put("CODART",_mask->get(F_AART));
      _form->find_field('B',odd_page,"H_CATMER").hide();
      _form->find_field('B',odd_page,"TOT_CATMER").hide();
      _form->set_ordering(mg_normale);
      if (!_mask->get(F_DAART).blank())
        srangea << "da \"" << _mask->get(F_DAART) << "\" ";
      if (!_mask->get(F_AART).blank())
        srangea << "fino a \"" << _mask->get(F_AART) << "\"";
    break;
    case 'D':
      _cur->setkey(2);
      darec.put("DESCR",_mask->get(F_DADES));
      arec.put("DESCR",_mask->get(F_ADES));
      _form->find_field('B',odd_page,"H_CATMER").disable();
      _form->find_field('B',odd_page,"TOT_CATMER").hide();
      _form->set_ordering(mg_normale);
      if (!_mask->get(F_DADES).blank())
        srangea << "da \"" << _mask->get(F_DADES) << "\" ";
      if (!_mask->get(F_ADES).blank())
        srangea << "fino a \"" << _mask->get(F_ADES) << "\"";
    break;
    case 'M':
      _cur->setkey(3);
      darec.put("GRMERC",_mask->get(F_DACATMER));
      arec.put("GRMERC",_mask->get(F_ACATMER));
      _form->find_field('B',odd_page,"H_CATMER").show();
      _form->find_field('B',odd_page,"TOT_CATMER").show();
      _form->set_ordering(mg_cat_merc);
      if (!_mask->get(F_DAART).blank())
      {
        srangea << "da \"" << _mask->get(F_DAART) << "\" ";
        cfilter << "(CODART[1," << _mask->get(F_DAART).len() << "]>=" <<'"' << _mask->get(F_AART)<< "\")&&" ;
      }
      if (!_mask->get(F_AART).blank())
      {
        srangea << "fino a \"" << _mask->get(F_AART) << "\"";
        cfilter << "(CODART[1," << _mask->get(F_AART).len() << "]<=" <<'"' << _mask->get(F_AART)<< "\")&&" ;
      }
      if (!_mask->get(F_DACATMER).blank())
        srangecm << "da \"" << _mask->get(F_DACATMER) << "\" ";
      if (!_mask->get(F_ACATMER).blank())
        srangecm << "fino a \"" << _mask->get(F_ACATMER) << "\"";
      rangecm.set(srangecm.empty() ? "Tutte le categorie merceologiche" : format("Categorie merc. %s", (const char *)srangecm));
    break;
  }
  rangea.set(srangea.empty() ? "Tutti gli articoli" : format("Articoli  %s",(const char *)srangea));
  
  _form->setdett_perart(
    _mask->get_bool(F_RAGGCODICE),
    _mask->get_int(F_FROMLIVELLOART),
    _mask->get_int(F_TOLIVELLOART),
    _mask->get_bool(F_RAGGLIVGIAC),
    _mask->get_int(F_FROMLIVELLOGIAC),
    _mask->get_int(F_TOLIVELLOGIAC),
    _mask->get_bool(F_TOTALIMAGAZZINI),
    _mask->get_bool(F_DETTAGLIOMAG));


  joinexp << "ANNOES==\"" << _mask->get(F_ANNOES) << "\"|CODART==CODART";

  TSortedfile *mag;
  // !?!?!! ATTENZIONE : Modifica temporanea
  mag= new TSortedfile(LF_MAG,NULL,sortexp,"",1);
  // il filtro viene qui settato DOPO la creazione del Sortedfile a causa di una bug
  // sulla libreria nei TSorted_file / TCursor
  mag->cursor().setfilter(filter);
  // FINE MODIFICA ; rimettere il filtro nel costruttore quando il bug sar risolto 
  _cur->relation()->replace(mag,1,joinexp);
  if (cfilter.not_empty())
  {
    cfilter.rtrim(2); 
    _cur->setfilter(cfilter);
  }
  _cur->setregion(darec,arec);
}


void TStampa_disp::main_loop()
{
  while (_mask->run() == K_ENTER) 
  {
    if (_mask->magazz_ini().gestmag(TRUE))
    {
      if (_mask->get(F_ORDINE)[0] == 'A')  
        setprint_perarticoli();
      else 
        setprint_permagazzini();
      _form->print();
      delete _form;
    }
  } // while true  
  return ;
}


int mg3200(int argc, char* argv[])
{
  TStampa_disp a;// derivata da Application e con uso di form
  a.run(argc,argv,"Stampa disponibilita'");
  return 0;
}
