#include <applicat.h>
#include <mask.h>
#include <progind.h>

#include "velib.h" 
#include "ve6200a.h"

#include <doc.h> 

class TLista_cf : public TLista_clifo 
{              
  char _tipo;
  
protected:
  virtual char tipo() const { return _tipo; }
  
public:
  TLista_cf(char tipo) { _tipo = tipo;}
};                                                  

class TFatturazione_bolle_app : public TSkeleton_application
{ 
  TString4 _default_selection;

protected:
  virtual bool create();
  virtual void main_loop();
  
  bool process(TIndwin& iw, TFatturazione_bolle& eld, TLista_documenti& din, const TDate& data_elab);
  void process_by_cli(const TMask& m);
  void process_by_doc(const TMask& m);
};

bool TFatturazione_bolle_app::create()
{
  open_files(LF_TABCOM, LF_TAB, LF_CLIFO, LF_OCCAS, LF_INDSP, LF_CFVEN, 
             LF_DOC, LF_RIGHEDOC, LF_ANAMAG, LF_MOVMAG, LF_RMOVMAG, LF_CONDV, LF_SVRIEP, 
             LF_AGENTI, LF_PERCPROV, LF_CAUSALI, 0);
                                            
  TConfig cfg(CONFIG_DITTA, "ve");
  _default_selection = cfg.get("FATBOLSEL");
             
  return TSkeleton_application::create();
}

bool TFatturazione_bolle_app::process(TIndwin& iw, TFatturazione_bolle& eld, 
                                      TLista_documenti& din, const TDate& data_elab)
{
  // Crea documenti di output
  TLista_documenti dout;
  bool ok = eld.elabora(din, dout, data_elab);
  if (ok) 
  {
    TString msg;
    msg << din.items() << " documenti raggruppati in " << dout.items(); 
    iw.set_text(msg);           // Messaggio sul cliente
    do_events();                // Attende visualizzazione

    int err = dout.write();     // Scrive documenti di output
    if (err == NOERR)
    {
      err = din.rewrite();      // Aggiorna stato dei documenti di input
      if (err != NOERR)
        ok = error_box("Errore %d durante la scrittura dei documenti raggruppati!", err);
    }
    else
      ok = error_box("Errore %d durante l'aggiornamento dei documenti da raggruppare!", err);
  }
  else
  {
    const long codcli = din[0].get_long(DOC_CODCF);
    ok = error_box("I documenti relativi al cliente %ld non sono stati elaborati.",
                   codcli);
  }  
  return ok;
}

void TFatturazione_bolle_app::process_by_cli(const TMask& m)
{
  TIndwin iw(0, "Inizializzazione ...\n ", TRUE, FALSE, 60);
  
  const TDate data_elab = m.get(F_DATA_ELAB); 
  const int anno = data_elab.year();
    
  const long dc = m.get_long(F_CODICE_CLIFO_DA);
  const long ac = m.get_long(F_CODICE_CLIFO_A);
  const int da = m.get_int(F_CODICE_AGENTE_DA);
  const int aa = m.get_int(F_CODICE_AGENTE_A);
  const TString16 dz(m.get(F_CODICE_ZONA_DA)); 
  const TString16 az(m.get(F_CODICE_ZONA_A));
    
  const TDate dd = m.get(F_DATA_DOCUMENTO_DA);
  TString st_da = m.get(F_DATA_DOCUMENTO_A);
  const TDate ad = st_da.not_empty() ? (const char*)st_da : data_elab;
  const TString& codnum = m.get(F_CODICE_NUMERAZIONE);
  const long dn = m.get_long(F_NUMERO_DOCUMENTO_DA);
  const long an = m.get_long(F_NUMERO_DOCUMENTO_A);
    
  TFatturazione_bolle eld(m.get(F_CODICE_ELAB));
  TToken_string tipidoc(24), statidoc(10);
  eld.tipi_validi(tipidoc); 
  eld.stati_validi(statidoc);
      
  TTipo_documento t(eld.tipo_iniziale(0)); 
  char tipocf(t.tipocf());
  TLista_cf clienti(tipocf);
  const int tot_cli = clienti.leggi(dc, ac, da, aa, dz, az);
    
  TString msg(80);
  for (int c = 0; c < tot_cli && !iw.iscancelled(); c++)
  { 
    const long codcli = clienti[c];     // Codice cliente in esame
    msg = "Elaborazione dei documenti del cliente "; 
    msg << codcli << " ...";
    iw.set_text(msg);                   // Messaggio sul cliente
    do_events();                        // Attende visualizzazione
      
    TLista_documenti din;               // Legge tutti i documenti di input
    din.read('D', tipocf, codcli, anno, tipidoc, statidoc, dd, ad, codnum, dn, an);
    if (din.items() > 0 && !iw.iscancelled())
    {
      bool ok = process(iw, eld, din, data_elab);
      if (!ok) // In caso di errore termina qui l'elaborazione.   
        break;   
    }
  }
}

void TFatturazione_bolle_app::process_by_doc(const TMask& m)
{
  TIndwin iw(0, "Inizializzazione ...\n ", TRUE, FALSE, 60);

  const TDate data_elab = m.get(F_DATA_ELAB); 
  const int anno = data_elab.year();
  long dc = m.get_long(F_CODICE_CLIFO_DA);
  long ac = m.get_long(F_CODICE_CLIFO_A);   
  
  const int da = m.get_int(F_CODICE_AGENTE_DA);
  const int aa = m.get_int(F_CODICE_AGENTE_A);
  const TString8 dz(m.get(F_CODICE_ZONA_DA)); 
  const TString8 az(m.get(F_CODICE_ZONA_A));
    
  const TDate dd = m.get(F_DATA_DOCUMENTO_DA);
  TDate ad = m.get(F_DATA_DOCUMENTO_A);
  if (!ad.ok()) ad = data_elab;
  
  const TString& codnum = m.get(F_CODICE_NUMERAZIONE);
  const long dn = m.get_long(F_NUMERO_DOCUMENTO_DA);
  const long an = m.get_long(F_NUMERO_DOCUMENTO_A);

  TFatturazione_bolle eld(m.get(F_CODICE_ELAB));
  TToken_string tipidoc(24), statidoc(10);
  eld.tipi_validi(tipidoc); 
  eld.stati_validi(statidoc);

  TTipo_documento t(eld.tipo_iniziale(0));
  char tipocf(t.tipocf());
  TLista_cf clienti(tipocf);
  const int tot_cli = clienti.leggi(dc, ac);

  TString msg(80);
  for (int c = 0; c < tot_cli && !iw.iscancelled(); c++)
  { 
    const long codcli = clienti[c];     // Codice cliente in esame
    msg = "Elaborazione dei documenti del cliente "; 
    msg << codcli << " ...";
    iw.set_text(msg);                   // Messaggio sul cliente
    do_events();                        // Attende visualizzazione
    
    TLista_documenti din, dout;         // Legge tutti i documenti di input
    din.read('D', tipocf, codcli, anno, tipidoc, statidoc, dd, ad, codnum, dn, an);
    
    for (int i = din.items()-1; i >= 0; i--)
    {              
      const TDocumento& doc = din[i];        
      
      bool to_del = FALSE;
      const int agente = doc.get_int(DOC_CODAG);
      if (agente > 0)
      {
        if (da > 0 && agente < da) to_del = TRUE;
        if (aa > 0 && agente > aa) to_del = TRUE;
      }
      const TString8 zona = doc.get(DOC_ZONA);
      if (zona.not_empty())
      {
        if (dz.not_empty() && zona < dz) to_del = TRUE;
        if (az.not_empty() && zona > az) to_del = TRUE;
      }
      if (to_del)
        din.destroy(i);
    }
    if (din.items() > 0 && !iw.iscancelled())
    {
      bool ok = process(iw, eld, din, data_elab);
      if (!ok) // In caso di errore termina qui l'elaborazione.   
        break;   
    }
  }
}
 
void TFatturazione_bolle_app::main_loop()
{
  TMask m("ve6200a");
  m.set(F_SELEZIONE, _default_selection);
  
  while (m.run() == K_ENTER)
  {                              
    const TString& select_from = m.get(F_SELEZIONE);  
    if (select_from != _default_selection)
    {
      TConfig cfg(CONFIG_DITTA, "ve"); 
      cfg.set("FATBOLSEL", _default_selection = select_from);
    }
    if (select_from == "D")
      process_by_doc(m);
    else  
      process_by_cli(m);
  }
}

int ve6200 (int argc, char **argv)
{
  TFatturazione_bolle_app a;
  a.run (argc, argv, "Fatturazione Bolle");
  return TRUE;
}
