#ifndef XVT_INCL_XVT
#define XVT_INCL_XVT

#ifdef XVT_INCL_NATIVE
#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#define STRICT
#include <windows.h>
#endif
#endif

#ifdef WIN32
#if XVAGADLL == 1
  #define XVTDLL __declspec(dllexport)
#else
  #define XVTDLL __declspec(dllimport)
#endif
#else
#define XVTDLL
#endif

#define XVAGA 1

#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "xvt_env.h"
#include "xvt_defs.h"
#include "xvt_help.h"
#include "xvt_menu.h"
#include "xvt_type.h"

#ifdef __cplusplus
extern "C" {
#endif

XVTDLL void     xvt_app_allow_quit(void);
XVTDLL void     xvt_app_create(int argc, char **argv, unsigned long flags, EVENT_HANDLER eh, XVT_CONFIG *config);
XVTDLL void     xvt_app_destroy(void);
XVTDLL BOOLEAN  xvt_app_escape(int esc_code, PRINT_RCD*, long* ph, long* pw, long* pvr, long* phr);
XVTDLL DRAW_CTOOLS* xvt_app_get_default_ctools(DRAW_CTOOLS* ct);
XVTDLL void    xvt_app_process_pending_events(void);

XVTDLL char*    xvt_cb_alloc_data(long size);
XVTDLL BOOLEAN  xvt_cb_close(void);
XVTDLL void     xvt_cb_free_data(void);
XVTDLL char*    xvt_cb_get_data(CB_FORMAT cbfmt, char *name, long *sizep);
XVTDLL BOOLEAN  xvt_cb_has_format(CB_FORMAT fmt, char *name);
XVTDLL BOOLEAN  xvt_cb_open(BOOLEAN writing);
XVTDLL BOOLEAN  xvt_cb_put_data(CB_FORMAT cbfmt, char *name, long size, PICTURE pic);

XVTDLL void    xvt_ctl_check_radio_button(WINDOW Win, WINDOW* Wins, int NbrWindows);
XVTDLL WINDOW  xvt_ctl_create_def(WIN_DEF *win_def_p, WINDOW parent_win, long app_data);
XVTDLL void    xvt_ctl_set_checked(WINDOW Win, BOOLEAN Check);

XVTDLL void  xvt_debug_printf(const char* fmt, ...);

XVTDLL ASK_RESPONSE  xvt_dm_post_ask(const char* Btn1, const char*Btn2, const char* Btn3, const char* xin_buffer);
XVTDLL COLOR         xvt_dm_post_choose_color(WINDOW win, COLOR c); // Added by guy
XVTDLL void          xvt_dm_post_error(const char *fmt);
XVTDLL void          xvt_dm_post_fatal_exit(const char *fmt);
XVTDLL FL_STATUS     xvt_dm_post_file_open(FILE_SPEC *fsp, const char *msg);
XVTDLL FL_STATUS     xvt_dm_post_file_save(FILE_SPEC *fsp, const char *msg);
XVTDLL BOOLEAN       xvt_dm_post_font_sel(WINDOW win, XVT_FNTID font_id, PRINT_RCD *precp, unsigned long reserved);
XVTDLL void          xvt_dm_post_message(const char *fmt);
XVTDLL void          xvt_dm_post_note(const char *fmt);
XVTDLL BOOLEAN       xvt_dm_post_page_setup(PRINT_RCD *precp);
XVTDLL char*         xvt_dm_post_string_prompt(const char* message, char* response, int response_len);
XVTDLL void          xvt_dm_post_warning(const char *fmt);

// Dongle support by AGA
XVTDLL BOOLEAN xvt_dongle_hl_crypt(unsigned short* data);
XVTDLL BOOLEAN xvt_dongle_hl_login(unsigned short address, const unsigned char* label, const unsigned char* password);
XVTDLL BOOLEAN xvt_dongle_hl_logout();
XVTDLL BOOLEAN xvt_dongle_hl_read(unsigned short reg, unsigned short* data);
XVTDLL BOOLEAN xvt_dongle_hl_read_block(unsigned char* data);
XVTDLL BOOLEAN xvt_dongle_hl_write(unsigned short reg, unsigned short data);
XVTDLL BOOLEAN xvt_dongle_sl_crypt(unsigned short* data);
XVTDLL BOOLEAN xvt_dongle_sl_login(const unsigned char* label, const unsigned char* password);
XVTDLL BOOLEAN xvt_dongle_sl_logout();
XVTDLL BOOLEAN xvt_dongle_sl_read_block(unsigned short reg, unsigned short size, unsigned short* data);
XVTDLL BOOLEAN xvt_dongle_sl_write_block(unsigned short reg, unsigned short size, const unsigned short* data);

XVTDLL void          xvt_dwin_clear(WINDOW win, COLOR col);
XVTDLL void          xvt_dwin_draw_arc(WINDOW Win, RCT* r, int sx, int sy, int ex, int ey);
XVTDLL void          xvt_dwin_draw_icon(WINDOW win, int x, int y, int rid);
XVTDLL void          xvt_dwin_draw_image(WINDOW win, XVT_IMAGE image, RCT* dest, RCT* source);
XVTDLL void          xvt_dwin_draw_oval(WINDOW Win, RCT* r);
XVTDLL void          xvt_dwin_draw_pie(WINDOW win, RCT *rctp, int start_x, int start_y, int stop_x, int stop_y);
XVTDLL void          xvt_dwin_draw_polygon(WINDOW win, PNT *lpnts, int npnts);
XVTDLL void          xvt_dwin_draw_rect(WINDOW win, RCT *rctp);
XVTDLL void          xvt_dwin_draw_dotted_rect(WINDOW win, RCT *rctp); // Added by Guy
XVTDLL void          xvt_dwin_draw_set_pos(WINDOW win, PNT pnt);
XVTDLL void          xvt_dwin_draw_text(WINDOW win, int x, int y, const char *s, int len);
XVTDLL RCT*          xvt_dwin_get_clip(WINDOW win, RCT* rct);
XVTDLL DRAW_CTOOLS*  xvt_dwin_get_draw_ctools(WINDOW win, DRAW_CTOOLS *ctoolsp);
XVTDLL XVT_FNTID     xvt_dwin_get_font(WINDOW win);
XVTDLL void          xvt_dwin_get_font_metrics(WINDOW win, int *leadingp, int *ascentp, int *descentp);
XVTDLL long          xvt_dwin_get_font_size_mapped(WINDOW win);
XVTDLL int           xvt_dwin_get_text_width(WINDOW win, const char *s, int len);
XVTDLL void          xvt_dwin_invalidate_rect(WINDOW win, RCT *rctp);
XVTDLL BOOLEAN       xvt_dwin_is_update_needed(WINDOW Win, RCT* rctp);
XVTDLL void          xvt_dwin_scroll_rect(WINDOW win, RCT *rctp, int dh, int dv);
XVTDLL void          xvt_dwin_set_back_color(WINDOW win, COLOR color);
XVTDLL void          xvt_dwin_set_cbrush(WINDOW win, CBRUSH* cbrush);
XVTDLL void          xvt_dwin_set_clip(WINDOW win, RCT* rct);
XVTDLL void          xvt_dwin_set_cpen(WINDOW win, CPEN* cpen);
XVTDLL void          xvt_dwin_set_draw_ctools(WINDOW win, DRAW_CTOOLS* xct);
XVTDLL void          xvt_dwin_set_draw_mode(WINDOW win, DRAW_MODE mode);
XVTDLL void          xvt_dwin_set_font(WINDOW win, XVT_FNTID font_id);
XVTDLL void          xvt_dwin_set_fore_color(WINDOW win, COLOR color);
XVTDLL void          xvt_dwin_set_std_cpen(WINDOW win, long flag);
XVTDLL void          xvt_dwin_draw_line(WINDOW win, PNT pnt);
XVTDLL void          xvt_dwin_update(WINDOW Win);

XVTDLL XVT_ERRSEV  xvt_errmsg_get_sev_id(XVT_ERRMSG err);

XVTDLL long  xvt_fmap_get_family_sizes(PRINT_RCD *precp, char *family, long *size_array, BOOLEAN *scalable, long max_sizes);
XVTDLL long  xvt_fmap_get_families(PRINT_RCD *precp, char **family_array, long max_families);

XVTDLL void                 xvt_font_copy(XVT_FNTID dest_font_id, XVT_FNTID src_font_id, XVT_FONT_ATTR_MASK mask);
XVTDLL XVT_FNTID            xvt_font_create(void);
XVTDLL void                 xvt_font_deserialize(XVT_FNTID font_id, char *buf);
XVTDLL void                 xvt_font_destroy(XVT_FNTID font_id);
XVTDLL BOOLEAN              xvt_font_get_family(XVT_FNTID font_id, char* buf, long max_buf);
XVTDLL BOOLEAN              xvt_font_get_family_mapped(XVT_FNTID font_id, char* buf, long max_buf);
XVTDLL void                 xvt_font_get_metrics(XVT_FNTID font_id, int *leadingp, int *ascentp, int *descentp);
XVTDLL BOOLEAN              xvt_font_get_native_desc(XVT_FNTID font_id, char *buf, long max_buf);
XVTDLL long                 xvt_font_get_size(XVT_FNTID font_id);
XVTDLL XVT_FONT_STYLE_MASK  xvt_font_get_style(XVT_FNTID font_id);
XVTDLL WINDOW               xvt_font_get_win(XVT_FNTID font_id);
XVTDLL BOOLEAN              xvt_font_is_mapped(XVT_FNTID font_id);
XVTDLL void                 xvt_font_map(XVT_FNTID font_id, WINDOW font_win );
XVTDLL void                 xvt_font_map_using_default(XVT_FNTID font_id);
XVTDLL void                 xvt_font_set_family(XVT_FNTID font_id, char* family);
XVTDLL void                 xvt_font_set_size(XVT_FNTID font_id, long size);
XVTDLL void                 xvt_font_set_style(XVT_FNTID font_id, XVT_FONT_STYLE_MASK mask);
XVTDLL long                 xvt_font_serialize(XVT_FNTID font_id, char *buf, long max_buf);
XVTDLL void                 xvt_font_unmap(XVT_FNTID font_id);

XVTDLL BOOLEAN  xvt_fsys_build_pathname(char *mbs, const char *volname, const char *dirname, const char *leafroot, const char *leafext, const char *leafvers);
XVTDLL BOOLEAN  xvt_fsys_convert_dir_to_str(DIRECTORY *dirp, char *path, int sz_path);
XVTDLL BOOLEAN  xvt_fsys_convert_str_to_dir(const char *path, DIRECTORY *dirp);
XVTDLL BOOLEAN  xvt_fsys_get_dir(DIRECTORY *dirp);
XVTDLL void     xvt_fsys_get_default_dir(DIRECTORY *dirp);
XVTDLL SLIST    xvt_fsys_list_files(char *type, char *pat, BOOLEAN dirs);
XVTDLL BOOLEAN  xvt_fsys_parse_pathname (const char *mbs, char *volname, char *dirname, char *leafroot, char *leafext, char *leafvers);
XVTDLL void     xvt_fsys_restore_dir();
XVTDLL void     xvt_fsys_save_dir();
XVTDLL BOOLEAN  xvt_fsys_set_dir(DIRECTORY *dirp);
// Added by Guy
XVTDLL unsigned long  xvt_fsys_get_disk_size(const char* path, char unit);
XVTDLL unsigned long  xvt_fsys_get_disk_free_space(const char* path, char unit);
XVTDLL BOOLEAN        xvt_fsys_is_removable_drive(const char* path);
XVTDLL BOOLEAN        xvt_fsys_is_network_drive(const char* path);
XVTDLL BOOLEAN        xvt_fsys_is_fixed_drive(const char* path);
XVTDLL BOOLEAN        xvt_fsys_test_disk_free_space(const char* path, unsigned long filesize);
XVTDLL BOOLEAN        xvt_fsys_mkdir(const char *pathname);
XVTDLL BOOLEAN        xvt_fsys_rmdir(const char *pathname);
XVTDLL BOOLEAN        xvt_fsys_removefile(const char *pathname);
XVTDLL BOOLEAN        xvt_fsys_access(const char *pathname, int mode);
XVTDLL BOOLEAN        xvt_fsys_file_exists(const char *pathname);

XVTDLL void           xvt_help_close_helpfile(XVT_HELP_INFO hi);
XVTDLL XVT_HELP_INFO  xvt_help_open_helpfile(FILE_SPEC *fs, unsigned long flags);
XVTDLL BOOLEAN        xvt_help_process_event(XVT_HELP_INFO hi, WINDOW win, EVENT *ev);

XVTDLL XVT_IMAGE         xvt_image_create(XVT_IMAGE_FORMAT format, short width, short height, long reserved);
XVTDLL void              xvt_image_destroy(XVT_IMAGE image);
XVTDLL COLOR             xvt_image_get_clut(XVT_IMAGE image, short index);
XVTDLL void              xvt_image_get_dimensions(XVT_IMAGE image, short *width, short *height);
XVTDLL XVT_IMAGE_FORMAT  xvt_image_get_format(XVT_IMAGE image);
XVTDLL short             xvt_image_get_ncolors(XVT_IMAGE image);
XVTDLL COLOR             xvt_image_get_pixel(XVT_IMAGE image, short x, short y);
XVTDLL XVT_IMAGE         xvt_image_read(const char *filenamep);
XVTDLL XVT_IMAGE         xvt_image_read_bmp(const char *filenamep);
XVTDLL void              xvt_image_set_clut(XVT_IMAGE image, short index, COLOR color);
XVTDLL void              xvt_image_set_ncolors(XVT_IMAGE image, short ncolors);
XVTDLL void              xvt_image_set_pixel(XVT_IMAGE image, short x, short y, COLOR color);
XVTDLL void              xvt_image_transfer(XVT_IMAGE dstimage, XVT_IMAGE srcimage, RCT *dstrctp, RCT *srcrctp);

XVTDLL DATA_PTR  xvt_mem_alloc(size_t size);
XVTDLL void      xvt_mem_free(DATA_PTR p);
XVTDLL DATA_PTR  xvt_mem_realloc(DATA_PTR p, size_t size);
XVTDLL DATA_PTR  xvt_mem_rep(DATA_PTR dst, DATA_PTR src, unsigned int srclen, long reps);
XVTDLL DATA_PTR  xvt_mem_zalloc(size_t size);

XVTDLL MENU_ITEM*  xvt_menu_get_tree(WINDOW win);
XVTDLL BOOLEAN     xvt_menu_popup(MENU_ITEM *menu_p, WINDOW win, PNT pos, XVT_POPUP_ALIGNMENT alignment, MENU_TAG item);
XVTDLL void        xvt_menu_set_font_sel(WINDOW win, XVT_FNTID font_id);
XVTDLL void        xvt_menu_set_item_checked(WINDOW win, MENU_TAG tag, BOOLEAN check);
XVTDLL void        xvt_menu_set_item_enabled(WINDOW win, MENU_TAG tag, BOOLEAN enable);
XVTDLL void        xvt_menu_set_item_title(WINDOW win, MENU_TAG tag, char* text);
XVTDLL void        xvt_menu_set_tree(WINDOW win, MENU_ITEM* tree);
XVTDLL void        xvt_menu_update(WINDOW win);

// Added by Guy
typedef const char* TRANSLATE_CALLBACK(const char* ita);
XVTDLL void  xvt_menu_translate_tree(WINDOW win, TRANSLATE_CALLBACK tc);

XVTDLL short        xvt_palet_add_colors(XVT_PALETTE palet, COLOR *colorsp, short numcolors);
XVTDLL short        xvt_palet_add_colors_from_image(XVT_PALETTE palet, XVT_IMAGE image);
XVTDLL XVT_PALETTE  xvt_palet_create(XVT_PALETTE_TYPE type, XVT_PALETTE_ATTR reserved);
XVTDLL void         xvt_palet_destroy(XVT_PALETTE palet);
XVTDLL short        xvt_palet_get_colors(XVT_PALETTE palet, COLOR *colorsp, short maxcolors);
XVTDLL short        xvt_palet_get_ncolors(XVT_PALETTE palet);
XVTDLL int          xvt_palet_get_tolerance(XVT_PALETTE p);
XVTDLL void         xvt_palet_set_tolerance(XVT_PALETTE p, int t);

XVTDLL void        xvt_print_close(void);
XVTDLL BOOLEAN     xvt_print_close_page(PRINT_RCD *precp);
XVTDLL PRINT_RCD*  xvt_print_create(int *sizep);
XVTDLL PRINT_RCD*  xvt_print_create_by_name(int* sizep, const char* name);
XVTDLL WINDOW      xvt_print_create_win(PRINT_RCD *precp, char *title);
XVTDLL void        xvt_print_destroy(PRINT_RCD *precp);
XVTDLL RCT*        xvt_print_get_next_band(void);
XVTDLL BOOLEAN     xvt_print_is_valid(PRINT_RCD *precp);
XVTDLL BOOLEAN     xvt_print_open(void);
XVTDLL BOOLEAN     xvt_print_start_thread (BOOLEAN (* print_fcn)(long), long data);
XVTDLL BOOLEAN     xvt_print_open_page(PRINT_RCD *precp);
// Added XVAGA
XVTDLL SLIST    xvt_print_list_devices();
XVTDLL BOOLEAN  xvt_print_set_default_device(const char* name);
XVTDLL BOOLEAN  xvt_print_get_default_device(char* name, int namesize);

XVTDLL int		 xvt_rect_get_height(RCT *rctp);
XVTDLL int		 xvt_rect_get_width(RCT *rctp);
XVTDLL BOOLEAN  xvt_rect_has_point(RCT *rctp, PNT pnt);
XVTDLL BOOLEAN  xvt_rect_intersect(RCT *drctp, RCT *rctp1, RCT *rctp2);
XVTDLL BOOLEAN  xvt_rect_is_empty(const RCT *rctp);
XVTDLL void     xvt_rect_offset(RCT *rctp, short dh, short dv);
XVTDLL void     xvt_rect_set(RCT *rctp, short left, short top, short right, short bottom);
XVTDLL void     xvt_rect_set_empty(RCT *rctp);
XVTDLL BOOLEAN  xvt_rect_set_pos(RCT *rctp, PNT pos);

XVTDLL void        xvt_res_free_menu_tree(MENU_ITEM* tree);
XVTDLL XVT_IMAGE   xvt_res_get_image(int rid);
XVTDLL MENU_ITEM*  xvt_res_get_menu(int rid);
XVTDLL char*       xvt_res_get_str(int rid, char *s, int sz_s);

XVTDLL int   xvt_sbar_get_pos(WINDOW win, SCROLL_TYPE t);
XVTDLL int   xvt_sbar_get_proportion(WINDOW win, SCROLL_TYPE t);
XVTDLL void  xvt_sbar_get_range(WINDOW win, SCROLL_TYPE t, int *minp, int *maxp);
XVTDLL void  xvt_sbar_set_pos(WINDOW win, SCROLL_TYPE t, int pos);
XVTDLL void  xvt_sbar_set_proportion(WINDOW win, SCROLL_TYPE t, int proportion);
XVTDLL void  xvt_sbar_set_range(WINDOW win, SCROLL_TYPE t, int min, int max);

XVTDLL void    xvt_scr_beep(void);
XVTDLL WINDOW  xvt_scr_get_focus_topwin(void);
XVTDLL WINDOW  xvt_scr_get_focus_vobj(void);
XVTDLL SLIST   xvt_scr_list_wins();
XVTDLL void    xvt_scr_set_busy_cursor();
XVTDLL void    xvt_scr_set_focus_vobj(WINDOW win);

XVTDLL BOOLEAN    xvt_slist_add_at_elt(SLIST x, SLIST_ELT e, const char *sx, long data);
XVTDLL int        xvt_slist_count(SLIST x);
XVTDLL SLIST      xvt_slist_create();
XVTDLL void       xvt_slist_destroy(SLIST list);
XVTDLL char*      xvt_slist_get(SLIST x, SLIST_ELT e, long *datap);
XVTDLL long*      xvt_slist_get_data(SLIST_ELT elt);
XVTDLL SLIST_ELT  xvt_slist_get_first(SLIST list);
XVTDLL SLIST_ELT  xvt_slist_get_next(SLIST list, SLIST_ELT item);

XVTDLL int     xvt_str_compare_ignoring_case (const char* s1, const char* s2);
XVTDLL char*   xvt_str_duplicate(const char* str);
XVTDLL BOOLEAN xvt_str_match(const char* str, const char* pat, BOOLEAN case_sensitive);

// System calls by XVAGA
XVTDLL void     xvt_sys_beep(int severity);
XVTDLL long     xvt_sys_execute(const char* cmdline, BOOLEAN sync, BOOLEAN iconizetask);
XVTDLL long     xvt_sys_execute_in_window(const char* cmdline, WINDOW win);
XVTDLL BOOLEAN  xvt_sys_get_host_name(char* name, int maxlen);
XVTDLL BOOLEAN  xvt_sys_get_user_name(char* name, int maxlen);
XVTDLL BOOLEAN  xvt_sys_goto_url(const char* url, const char* action);
XVTDLL BOOLEAN  xvt_sys_dongle_server_is_running();
XVTDLL BOOLEAN  xvt_sys_find_editor(const char* file, char* editor);
XVTDLL int      xvt_sys_get_profile_string(const char* file, const char* paragraph, const char* name,
                                   const char* defval,  char* value, int maxsize);
XVTDLL BOOLEAN  xvt_sys_set_profile_string(const char* file, const char* paragraph, const char* name,
                                   const char* value);
XVTDLL unsigned long  xvt_sys_get_free_memory();
XVTDLL unsigned long  xvt_sys_get_free_memory_kb();
XVTDLL int          xvt_sys_get_os_version();
XVTDLL unsigned int xvt_sys_load_icon(const char* file);
XVTDLL void         xvt_sys_sleep(unsigned long msec);
XVTDLL BOOLEAN      xvt_sys_test_network_version();
XVTDLL void         xvt_sys_searchenv(const char *filename, const char *varname, char *pathname);
XVTDLL void         xvt_sys_sorry_box(const char * file, int line);
XVTDLL long         xvt_timer_create(WINDOW win, long interval);
XVTDLL void         xvt_timer_destroy(long id);
XVTDLL void         xvt_vobj_destroy(WINDOW win);
XVTDLL long         xvt_vobj_get_attr(WINDOW win, long data);
XVTDLL RCT*         xvt_vobj_get_client_rect(WINDOW win, RCT *rctp);
XVTDLL long         xvt_vobj_get_data(WINDOW win);
XVTDLL RCT*         xvt_vobj_get_outer_rect(WINDOW win, RCT *rctp);
XVTDLL XVT_PALETTE  xvt_vobj_get_palet(WINDOW win);
XVTDLL WINDOW       xvt_vobj_get_parent(WINDOW win);
XVTDLL char*        xvt_vobj_get_title(WINDOW win, char *title, int sz_title);
XVTDLL WIN_TYPE     xvt_vobj_get_type(WINDOW win);
XVTDLL void         xvt_vobj_maximize(WINDOW win); // Added by XVAGA
XVTDLL void         xvt_vobj_minimize(WINDOW win); // Added by XVAGA
XVTDLL void         xvt_vobj_move(WINDOW win, RCT *rctp);
XVTDLL void         xvt_vobj_raise(WINDOW win);
XVTDLL void         xvt_vobj_set_attr(WINDOW win, long data, long value);
XVTDLL void         xvt_vobj_set_data(WINDOW win, long AppData);
XVTDLL void         xvt_vobj_set_enabled(WINDOW win, BOOLEAN enabled);
XVTDLL void         xvt_vobj_set_palet(WINDOW win, XVT_PALETTE palet);
XVTDLL void         xvt_vobj_set_title(WINDOW win, char *title);
XVTDLL void         xvt_vobj_set_visible(WINDOW win, BOOLEAN show);
XVTDLL void         xvt_vobj_translate_points(WINDOW from_win, WINDOW to_win, PNT *pntp, int npnts);
XVTDLL WINDOW  xvt_win_create(WIN_TYPE wtype, RCT *rct_p, char *title, int menu_rid, WINDOW parent_win, long win_flags, EVENT_MASK mask, EVENT_HANDLER eh, long app_data);
XVTDLL long    xvt_win_dispatch_event(WINDOW win, EVENT* event_p);
XVTDLL void    xvt_win_post_event(WINDOW win, EVENT* event_p); // Added by XVAGA
XVTDLL void    xvt_win_release_pointer(void);
XVTDLL void    xvt_win_set_caret_size(WINDOW win, int width, int height);
XVTDLL void    xvt_win_set_caret_pos(WINDOW win, PNT p);
XVTDLL void    xvt_win_set_caret_visible(WINDOW win, BOOLEAN on);
XVTDLL void    xvt_win_set_cursor(WINDOW win, CURSOR Cursor);
XVTDLL void    xvt_win_set_handler(WINDOW win, EVENT_HANDLER eh);
XVTDLL void    xvt_win_trap_pointer(WINDOW win);

#ifdef __cplusplus
}
#endif

#define SORRY_BOX() xvt_sys_sorry_box(__FILE__, __LINE__)

#endif
